#
# $Header: /scratch/fock/cvsroot/xd/xdgraph/tcl/autoContour.tcl,v 1.1 1996/05/23 07:42:10 richter Rel $
#
#
# $Log: autoContour.tcl,v $
# Revision 1.1  1996/05/23 07:42:10  richter
# Initial revision
#
#

proc auto {what args} {
    regsub {^-} $what "" what
    auto_$what $args
}

proc parseargs {store over args} {
    upvar $store a
    upvar $over o

    if {[llength $args] == 1} {
	set args [lindex $args 0]
    }

    while {[llength $args] >= 2} {
	set pair [lrange $args 0 1]
	set args [lreplace $args 0 1]
	set a([lindex $pair 0]) [lindex $pair 1]
    }

    if {[llength $args] == 1} {
	set o $args
    }
}
    
proc auto_contour {args} {
    global active

    if {[llength $args] == 1} {
	set args [lindex $args 0]
    }

    set store() {}
    set over ""
    parseargs store over $args
    set flags [array names store]

    if {[lsearch $flags -file] != -1 && [set file $store(-file)] != ""} {
	set sc 0
	set dl [dataset]
	set set [format ":set%2.2d" $sc]
	while {[lsearch -exact $dl $set] != -1} {
	    incr sc
	    set set [format ":set%2.2d" $sc]
	}
	doCommand "dataset $set -load $file"
	set active(dataset) $set
	doCommand "$set connect -auto"
    } else {
	set set $active(dataset)
    }
    set plus $set:plus
    set zero $set:zero
    set minus $set:minus

    if {[lsearch $flags -mode] != -1} {
	set mode $store(-mode)
    } else {
	set mode ""
    }
    switch -- $mode {
	bw -
	color -
	colorbg -
	ramp {
	}
	coloramp -
	colorramp {
	    set mode ramp
	}
	"" {
	    set mode bw
	}
	default {
	    printmsg wrn "unknown mode \"$mode\" -- reset to \"bw\"\n"
	    set mode bw
	}
    }

    set prop [string tolower [$set property]]
    switch -- $prop {
	d2rho {
	    set pval [generate aim]
	    set mval [generate maim]
	    unset zero
	}
	def {
	    switch -- $mode {
		bw -
		color {
		    set pval [generate lin 20 0.1 0.1]
		    set zval 0.
		    set mval [generate lin 10 -0.1 -0.1]
		}
		colorbg {
		    set pval [generate lin 10 0.2 0.2]
		    set zval 0.
		    set mval [generate lin 50 -0.2 -0.2]
		}
		ramp {
		    set pval [generate lin 10 0.1 0.1]
		    set zval 0.
		    set mval [generate lin 10 -0.1 -0.1]
		}
	    }
	}
    }

    switch -- $mode {
	bw {
	    if {[info exists plus]} {
		contour $plus -values $pval -style solid
	    }
	    if {[info exists zero]} {
		contour $zero -values $zval -style dash
	    }
	    if {[info exists minus]} {
		contour $minus -values $mval -style dot
	    }
	}
	color {
	    if {[info exists plus]} {
		contour $plus -values $pval -foreground blue
	    }
	    if {[info exists zero]} {
		contour $zero -values $zval -foreground black
	    }
	    if {[info exists minus]} {
		contour $minus -values $mval -foreground red
	    }
	}
	colorbg {
	    if {[info exists plus]} {
		contour $plus -values $pval -style solid
	    }
	    if {[info exists zero]} {
		contour $zero -values $zval -style dash
	    }
	    if {[info exists minus]} {
		contour $minus -values $mval -style dot
	    }
	    # colorbg?
	}
	ramp {
	    if {[info exists plus]} {
		contour $plus -values $pval \
		    -foreground {{YUV 0.6 0.6 0.1} {YUV 0.1 0.7 0.44}}
	    }
	    if {[info exists zero]} {
		contour $zero -values $zval -style dash
	    }
	    if {[info exists minus]} {
		contour $minus -values $mval \
		    -foreground {{YUV 0.9 0.2 0.5} {YUV 0.2 0.4 0.88}}
	    }
	}
    }

    doCommand "molecule $set:mol"
    plot C$set
}

proc auto_molecule {args} {
    global active

    if {[llength $args] == 1} {
	set args [lindex $args 0]
    }

    set store() {}
    set over ""
    parseargs store over $args
    set flags [array names store]

    if {[lsearch $flags -file] != -1 && [set file $store(-file)] != ""} {
	set sc 0
	set dl [dataset]
	set set [format ":set%2.2d" $sc]
	while {[lsearch -exact $dl $set] != -1} {
	    incr sc
	    set set [format ":set%2.2d" $sc]
	}
	doCommand "dataset $set -load $file"
	set active(dataset) $set
	doCommand "$set connect -auto"
    } else {
	set set $active(dataset)
    }
    doCommand "molecule $set:mol"
    doCommand "$set:mol plot"
}
