#
# Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.
#
# $Header: /scratch/fock/cvsroot/xd/xdgraph/tcl/cli.tcl,v 2.4 1999/06/03 12:36:26 richter Exp $
#
#
# $Log: cli.tcl,v $
# Revision 2.4  1999/06/03 12:36:26  richter
# first throw at removable views
#
# Revision 2.3  1996/10/11 12:11:01  richter
# removed ps procedure
#
# Revision 2.2  1996/09/14  09:39:56  richter
# doCommand may have to print commands as info text (new parameter)
# plot and clear use new 'show_as_inf' option of doCommand
#
# Revision 2.1  1996/09/06  12:39:31  richter
# new proc config2info
#
# Revision 2.0  1995/05/12  11:54:37  richter
# First semi-public beta release
#
#
##source $xd_library/autoContour.tcl

proc generate {type {nstep 1} {start 0} {incr 1}} {
    set v $start
    set l {}

    switch -exact $type {
	lin -
	linear {
	    for {set i 0} {$i < $nstep} {incr i} {
		lappend l $v
		set v [expr $v + $incr]
	    }
	}
	geo -
	geom -
	log {
	    for {set i 0} {$i < $nstep} {incr i} {
		lappend l $v
		set v [expr $v * $incr]
	    }
	}
	aim {
	    set l { .001 .002 .004 .008 .02 .04 .08 .2 .4 .8 2. 4. 8. 20. 40. 80. 200. 400. 800.}
	    if {$nstep > 1} {
		set l [lrange $l $start [expr $nstep + $start - 1]]
	    }
	}
	maim {
	    set l { -.001 -.002 -.004 -.008 -.02 -.04 -.08 -.2 -.4 -.8 -2. -4. -8. -20. -40. -80. -200. -400. -800.}
	    if {$nstep > 1} {
		set l [lrange $l $start [expr $nstep + $start - 1]]
	    }
	}
    }
    set l
}

proc objects_from_args {{a ""}} {
    global active

    if {$a == {}} {
	if {$active(view) != ""} {
	    set objects [$active(view) objects]
	}
    } elseif {$a == {all} } {
	foreach view [view] {
	    lappend objects [$view objects]
	}
    } else {
	set objects [$a objects]
    }
}

proc plot {args} {
    if {[lindex args 0] == "-force"} {
	set switch "-force"
	set args [lreplace args 0 0]
    } else {
	set switch ""
    }

    foreach pair [objects_from_args $args] {
	set obj [lindex $pair 0]
	set code [catch "doCommand \"$obj plot\" 1 1" result]
	if {$code == 101} {
	    return
	}
    }
}

proc clear {args} {
    foreach pair [objects_from_args $args] {
	set obj [lindex $pair 0]
	set code [catch "doCommand \"$obj clear\" 1 1" result]
	if {$code == 101} {
	    return
	}
    }
    return
}

proc remove {args} {
    global active
    if {$args != {} && $args != {all} && [info commands $args] == {}} {
	return
    }
    foreach pair [objects_from_args $args] {
	set obj [lindex $pair 0]
	set code [catch [list doCommand [list $obj remove] 1 1] result]
	if {$code == 101} {
	    return
	}
    }
    if {$args == {}} {
	if {$active(view) != ""} {
	    set views $active(view)
	}
    } elseif {$args == {all} } {
	set views [view]
    } else {
	set views $args
    }
    foreach v $views {
	$v remove
    }
    return
}


proc view2display {view} {
    return [view2widget $view].display
}

proc view2widget {view} {
    regsub -all {[\.:]} $view "_" v
    return ".v:$v"
}

proc doCommand {{command ""} {return_all_codes 0} {show_as_inf 0}} {
    global next_event
    global in 

    if {[string compare $command ""] == 0} {
	set command $in
    }
    if {$show_as_inf} {
	printmsg inf "$command\n"
    } else {
	printmsg cmd "$command\n"
    }
    history add $command
    set next_event [history nextid]
    set code [catch [list uplevel #0 $command] result]
    if {$code == 0 || $code == 101} {
	if {"$result" != ""} {
	    printmsg inf "$result\n"
	}
	set in ""
	if {!$return_all_codes} {
	    set code 0
	}
    } else {
	printmsg err "$result\n"
	set in $command
    }
    return -code $code
}

proc pi {{string ""}} {
    printmsg inf $string
}

proc config2info {data fmtName} {
    upvar $fmtName fmt

    foreach pair $data {
	set key [lindex $pair 0]
	if {[string index $key 0] == "_"} {
	    if {$key != "_SECTION"} {
		set fmt($key) [lrange $pair 1 end]
	    }
	} else {
	    set value [lrange [lindex $pair 3] 0 end]
	    if {$value == ""} {
		set value [lrange [lindex $pair 2] 0 end]
	    }
	    if {$value != ""} {
		set fmt($key) $value
	    }
	}
    }
}

proc setinfo {o} {
    set fmt(_) ""
    config2info [$o configure] fmt

    pi [format "Dataset '%s'  Flags 0x%2.2x " \
	$o $fmt(_flags)]
    unset fmt(_flags)
    pi [format "Compound '%s'  Property '%s'\n" $fmt(_compid) $fmt(_prop)]
    unset fmt(_compid) fmt(_prop)
    pi "\t$fmt(_title)\n"
    unset fmt(_title)
    foreach key [array names fmt] {
	pi "\t$key: $fmt($key)\n"
    }
}

proc ctrinfo {o} {
    set fmt(_) ""
    config2info [$o configure] fmt

    pi [format "Contour '%s'  Flags 0x%2.2x  View '%s'  Object '%s'\n" \
	$o $fmt(_flags) $fmt(view) $fmt(_object)]
    unset fmt(_flags)
    unset fmt(view)
    unset fmt(_object)
    set values $fmt(values)
    unset fmt(values)
    foreach key [array names fmt] {
	pi "\t$key: $fmt($key)\n"
    }
    pi [format "\tLevels %d %s\n" [llength $values] $values]
}

proc pthinfo {o} {
    set data [$o configure]

    foreach pair $data {
	set fmt([lindex $pair 0]) [lrange $pair 1 end]
    }

    pi [format "Path '%s'  Flags 0x%2.2x  View '%s'  Object '%s'\n" \
	$o 0 $fmt(view) $fmt(_set)]
    unset fmt(view)
    unset fmt(_set)
    foreach key [array names fmt] {
	pi "\t$key: $fmt($key)\n"
    }
}

proc heiinfo {o} {
    set data [$o configure]

    foreach pair $data {
	set fmt([lindex $pair 0]) [lrange $pair 1 end]
    }

    pi [format "Heightfield '%s'  Flags 0x%2.2x  View '%s'  Object '%s'\n" \
	$o $fmt(_flags) $fmt(view) $fmt(_object)]
    unset fmt(_flags)
    unset fmt(view)
    unset fmt(_object)
    foreach key [array names fmt] {
	pi "\t$key: $fmt($key)\n"
    }
}

