#
# Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.
#
# $Header: /scratch/fock/cvsroot/xd/xdgraph/tcl/xd.tcl,v 2.8 1999/05/13 10:37:29 richter Rel $
#
# $Log: xd.tcl,v $
# Revision 2.8  1999/05/13 10:37:29  richter
# allow Tcl/Tk 7.0 upwards, including but not limited to 8.0
# set xd_use_fsbox according to Tcl release and env. variables
#
# Revision 2.7  1997/04/02 13:46:17  richter
# try to recognize Unix and non-Unix (i.e. VMS) platforms
#
# Revision 2.6  1997/03/30  14:20:44  richter
# try to find out if we are on a unix platform
#
# Revision 2.5  1996/11/07  17:11:43  richter
# removed the last change to proc exists
#
# Revision 2.4  1996/11/07  10:50:01  richter
# new compatibilty procedure last
# exists now uses 7.4 facilities if available
#
# Revision 2.3  1996/09/06  12:53:17  richter
# initialize array active
# use apwd when available as /usr/local/bin/apwd
#
# Revision 2.2  1996/06/24  11:06:02  richter
# proc exists
# no more auto loading
#
# Revision 2.0  1995/05/12  11:54:39  richter
# First semi-public beta release
#
#

# Insist on running with compatible versions of Tcl

scan [info tclversion] "%d.%d" xd_tcl_major xd_tcl_minor
if {$xd_tcl_major < 7} {
    error "wrong version of Tcl loaded ([info tclversion]): need 7.x or later"
}
set xd_tcl_version "$xd_tcl_major.$xd_tcl_minor"
# maybe necessary on VMS platforms:
# set xd_library /dka200/xray/xd/lib/tcl

#
# compatibility routines
#

proc exists {varName key} {
    upvar $varName var

    if {[lsearch [array names var] $key] == -1} {
	return 0
    } else {
	return 1
    }
}

proc last {list} {
    global xd_tcl_version

    if {$xd_tcl_version >= 7.4} {
	return [lindex $list end]
    } else {
	set last [expr [llength $list] - 1]
	return [lindex $list $last]
    }
}


if {$xd_tcl_major >= 8 && !([exists env XD_USE_PRIVATE_FSBOX])} {
    set xd_use_fsbox "tcl"
} elseif {$xd_tcl_version == 7.4 && [exists env XD_USE_STD_FSBOX]} {
    set xd_use_fsbox "tcl"
} else {
    set xd_use_fsbox "xd"
}

if {[catch {file exists /usr} xd_unix]} {
    set xd_unix 0
}

#Cljf - Force OS to be defined as "Unix" regardless
    set xd_unix 1
#Cljf

if {$xd_unix == 0} {
    set xd_pwd {}
} elseif {[file isfile /usr/local/bin/apwd]} {
    set xd_pwd apwd
} else {
    set xd_pwd pwd
}

set active(display) ""
set active(view) ""
set active(contour) ""
set active(heightfield) ""
set active(isosurface) ""
set active(molecule) ""
set active(path) ""
set active(relief) ""
set active(dataset) ""
set active(view) ""
set active(xydiagram) ""

proc quit {} { exit }
