#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/ctrCommand.c,v 2.14 1997/03/29 18:01:44 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif

/*
 *======================================================================
 *
 * Module: ctrCommand
 *
 * Abstract: Defines the 'contour' command and creates contour object
 * commands.  Called from the tcl interpreter.
 *
 * $Log: ctrCommand.c,v $
 * Revision 2.14  1997/03/29 18:01:44  richter
 * preset style
 *
 * Revision 2.13  1996/11/14  12:57:58  richter
 * preset foreground colour
 *
 * Revision 2.12  1996/11/03  16:29:50  richter
 * styGetConfig expects a style group mask
 *
 * Revision 2.11  1996/09/05  15:04:51  richter
 * many changes to use the new display list interface
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.3  1996/07/25  05:58:37  richter
 * use Tcl_SetVar2 to set active(contour)
 * handle 'contour <contour>' correctly
 * cmdSetView takes an array of possible view types
 * new option -vertac
 *
 * Revision 2.9.1.2  1996/07/03  14:28:53  richter
 * refined view scaling
 * configure output
 *
 * Revision 2.9.1.1  1996/05/19  16:04:18  richter
 * lots of changes
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:04  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "ctrProto.h"
#include "cmdProto.h"
#include "vieProto.h"
#include "pltProto.h"
#include "setProto.h"
#include "strProto.h"
#include "styProto.h"
#include "vieProto.h"
#include "dlProto.h"

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static int configureContour(Tcl_Interp *interp, ContourP cPtr,
			    int argc,char **argv, ViewOptT *vOpt, int append);
static int showcfgContour(Tcl_Interp *interp, ContourP cPtr);
static int infoContour(Tcl_Interp *interp, ContourP cPtr,
		       int argc,char **argv);
static int infoLevel(Tcl_Interp *interp, LevelP lPtr);
static int plotContour(Tcl_Interp *interp, ViewP view, ContourP cPtr,
		       int argc,char **argv);
static int clearContour(Tcl_Interp *interp, ViewP view, ContourP cPtr);
/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */

int
ContourCmd(ClientData dummy,		/* Not used			*/
	   Tcl_Interp *interp,		/* Current interpreter		*/
	   int argc,			/* Number of arguments		*/
	   char **argv)			/* Argument strings		*/
{
    DatasetP dPtr;
    ContourP cPtr;
    ViewOptT vOpt;
    char *objName;
#define CTR_VIEW_TYPES 2
    ViewE vieType[CTR_VIEW_TYPES] = {VIEW_CONTOUR, VIEW_HEIGHT};
    int i;
    
    if (argc == 1 ||
        (argc == 2 && strEQ(argv[1], "configure", 4))) {
        return showcfgContour(interp, NULL);
    }

    if (cmdAnalyzeName(argv[1], &objName, &dPtr) != TCL_OK) {
        return TCL_ERROR;
    }

    if (dPtr->type != DSET_GRID && dPtr->type != DSET_SLICE) {
	Tcl_AppendResult(interp,
                         "contour is only valid for GRID or SLICE datasets\n",
                         (char *) NULL);
	return TCL_ERROR;
    }
        
    cPtr = ctrAlloc(argv[1]);
    if (!cPtr) {
	Tcl_AppendResult(interp, "couldn't create contour", (char *) NULL);
	return TCL_ERROR;
    }
    cPtr->object = strDupl(objName);
    for (i=0; i<=1; ++i) {
        styFillFG(interp, &cPtr->style[i], "black");
        styFillLinewidth(interp, &cPtr->style[i], 1.);
        styFillPattern(interp, &cPtr->style[i], "solid");
    }
    vOpt.doscale = 1;
    vOpt.doorig  = 0;
    vOpt.dorot   = 0;
    vOpt.dotrans = 0;
    vOpt.scale = -1.;
    vOpt.width = -1.;
    vOpt.height = -1.;
    if (configureContour(interp, cPtr, argc-2, argv+2, &vOpt,
                         FALSE) != TCL_OK ) {
	goto Error;
    }

    vOpt.vrange[0] = dPtr->xsize;
    vOpt.vrange[1] = dPtr->ysize;
    if (cmdSetView(interp, &cPtr->view, CTR_VIEW_TYPES, vieType, objName,
                   argv[1], "contour", &vOpt) != TCL_OK) {
        Tcl_AppendResult(interp, "view could not be set\n", NULL);
        goto Error;
    }
    Tcl_CreateCommand(interp, argv[1], ContourObjectCmd, (ClientData) cPtr,
		      (Tcl_CmdDeleteProc *) NULL );
    Tcl_SetVar2(interp, "active", "contour", argv[1], TCL_GLOBAL_ONLY);
    return TCL_OK;

Error:
    ctrFree(&cPtr);
    return TCL_ERROR;
}

int
ContourObjectCmd(ClientData data,		/* Contour group	*/
		 Tcl_Interp *interp,		/* Current interpreter. */
		 int argc,			/* Number of arguments. */
		 char **argv)			/* Argument strings.	*/
{
    ContourP cPtr;
    ViewP view;
    ViewOptT vOpt;

    cPtr = (ContourP) data;
    view = viewFindByName(cPtr->view, NULL);
    
    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " option ?args?\"", NIL);
        return TCL_ERROR;
    }

    if (strEQ(argv[1],"configure",2)) {
        if (argc <= 2) return showcfgContour(interp, cPtr);
        /*
        vOpt.doorig = 0;
        vOpt.dorot = 0;
        vOpt.dotrans = 0;
        */
        vOpt.doscale = 0;
	if (configureContour(interp, cPtr, argc-2, argv+2, &vOpt, FALSE)
            == TCL_OK) {
            if (vOpt.doscale) {
                view->opt.doscale = vOpt.doscale;
                view->opt.scale = vOpt.scale;
            }
          return TCL_OK;
        } else {
            return TCL_ERROR;
        }
    } else if (strEQ(NULL,"append",1)) {
	return configureContour(interp, cPtr, argc-2, argv+2, &vOpt, TRUE);
    } else if (strEQ(NULL,"clear",2)) {
        clearContour(interp, view, cPtr);
        /*	lvlClearPlotFlag(cPtr->level); */
    } else if (strEQ(NULL,"info",1)) {
	return infoContour(interp, cPtr, argc-2, argv+2);
    } else if (strEQ(NULL,"plot",1)) {
        clearContour(interp, view, cPtr);
	return plotContour(interp, view, cPtr, argc-2, argv+2);
    } else if (strEQ(NULL,"remove",1)) {
        clearContour(interp, view, cPtr);
        viewUnregisterObject(view, cPtr->name);
	Tcl_DeleteCommand(interp, cPtr->name);
	ctrFree(&cPtr);
    } else {
	Tcl_AppendResult(interp, "unknown option \"", argv[1],
			 "\" must be configure, append, plot or remove\n",
			 (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;

}
    
static int
configureContour(Tcl_Interp *interp,
		 ContourP cPtr,
		 int argc,
		 char **argv,
                 ViewOptT *vOpt,
		 int append)
{
    DatasetP dPtr;
    GridP gPtr;
    LevelP *lpPtr;
    LevelP lp;
    int i, k, ival;
    double value;
    int recalc, has2styles, ac;
    int code;
    int elemc;
    char **elemv = NULL;
    ViewP view;
    
    static int iplane = 0;              /* remember selected plane between
					   calls */
    recalc = FALSE;

    dPtr = setFindByName(cPtr->object);
    if (dPtr->type == DSET_GRID) {
        gPtr = &dPtr->data.dg->grid;
    } else {
        gPtr = &dPtr->data.ds->grid;
    }

    code = viewParseArgs(NULL, &argc, argv, vOpt);
    if (code != TCL_OK) return code;

    ac = argc;
    code = styParseArgs(&argc, argv, &cPtr->style[0], &cPtr->style[1],
                        &has2styles);
    if (code != TCL_OK) return code;
    
    if (ac != argc) {
        lvlClearPlotFlag(cPtr->level);
    }
    if (has2styles) {
        cPtr->flags |= M_CTR_HAS_COLORRAMP;
    } else {
        cPtr->flags &= ~ M_CTR_HAS_COLORRAMP;
    }
    
    for (i=0; i<argc; ++i) {
	if (strEQ(argv[i], "-plane", 2)) {
	    i++;
	    if (i >= argc) {
		Tcl_AppendResult(interp, "plane no. is missing",
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetInt(interp, argv[i], &ival) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    if (ival > gPtr->nz || ival < 1) {
		Tcl_AppendResult(interp, "plane no. outside valid range",
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    iplane = ival-1;
	    lvlClearCalcFlag(cPtr->level);
	    recalc = TRUE;
	} else if (strEQ(NULL,"-values",2)) {
	    i++;
	    if ( i >= argc ) {
		Tcl_AppendResult(interp, "no values given",
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    if ( Tcl_SplitList( interp, argv[i], &elemc, &elemv ) != TCL_OK ) {
		return TCL_ERROR;
	    }
            view = viewFindByName(cPtr->view, NULL);
            if (view) clearContour(interp, view, cPtr);
	    lpPtr = &cPtr->level;
	    lp = cPtr->level;
	    if ( lp ) {
		if ( append ) {
		    while (lp->next) {
			lp = lp->next;
		    }
		    lpPtr = &lp->next;
		} else {
		    lvlFree(&lp);
		    cPtr->level = NULL;
		    cPtr->nl =    0;
		}
	    }
	    for (k=0; k<elemc; ++k) {
		if ( Tcl_GetDouble(interp, elemv[k], &value) != TCL_OK ) {
                    goto Error;
		}
		if ( ! (lp = (*lpPtr) = lvlAlloc()) ) {
		    Tcl_AppendResult(interp, "couldn't create contour level",
				     "(at ", elemv[k], ")", (char *) NULL);
                    goto Error;
		}
		lp->value = value;
		lpPtr = &lp->next;
		cPtr->nl += 1;
	    }
	    ckfree((char *) elemv);
	    recalc = TRUE;
        } else if (strEQ(NULL,"-vertac",3)) {
            if (++i >= argc) {
                Tcl_AppendResult(interp, "value for \"-vertac\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            if ( Tcl_GetDouble(interp, argv[i], &cPtr->vertac) != TCL_OK  ) {
                return TCL_ERROR;
            }
        } else if (strEQ(NULL,"-view",3)) {
            i++;
            if (i >= argc) {
                Tcl_AppendResult(interp, "value for \"-view\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            cPtr->view = strDupl(argv[i]);
	} else {
	    Tcl_AppendResult(interp, "unknown option \"",
			     argv[i], "\"", (char *) NULL);
	    return TCL_ERROR;
	}
    }

    if (recalc) {
	gPtr->pl = gPtr->va + gPtr->nx*gPtr->ny*iplane;
	if (ctrEngine(gPtr, cPtr) != TCL_OK){
	    Tcl_AppendResult(interp, "contour plot failed",(char *) NULL);
	    return TCL_ERROR;
	}
    }
    return TCL_OK;

Error:
    ckfree((char *) elemv);
    elemv = NULL;
    return TCL_ERROR;
}

static int
showcfgContour(Tcl_Interp *interp,
               ContourP cPtr)
{
    LevelP lp;
    Tcl_DString cfg;

    Tcl_DStringInit(&cfg);
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "view");
    Tcl_DStringAppendElement(&cfg, "string::create");
    Tcl_DStringAppendElement(&cfg, "View");
    Tcl_DStringAppendElement(&cfg, "");
    if (cPtr) {
        Tcl_DStringAppendElement(&cfg, cPtr->view);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);
        
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "values");
    Tcl_DStringAppendElement(&cfg, "list::required");
    Tcl_DStringAppendElement(&cfg, "Contour Values");
    Tcl_DStringAppendElement(&cfg, "");
    if (cPtr) {
        Tcl_DStringStartSublist(&cfg);
        for(lp=cPtr->level; lp; lp=lp->next) {
            sprintf(buf, "%g", lp->value);
            Tcl_DStringAppendElement(&cfg, buf);
        }
        Tcl_DStringEndSublist(&cfg);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);

    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "vertac");
    Tcl_DStringAppendElement(&cfg, "float");
/* Cljf    Tcl_DStringAppendElement(&cfg, "Vertical Acceleration"); */ 
    Tcl_DStringAppendElement(&cfg, "Vertical Accentuation"); 
    Tcl_DStringAppendElement(&cfg, "");
    if (cPtr) {
        sprintf(buf, "%g", cPtr->vertac);
        Tcl_DStringAppendElement(&cfg, buf);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);
        
    if (cPtr) {
        ViewP view = viewFindByName(cPtr->view, NULL);
        styGetConfig(&cfg, &cPtr->style[0], STY_FG_ALL|STY_LINE_ALL,
                     1, (cPtr->flags & M_CTR_HAS_COLORRAMP), &cPtr->style[1]);
        if (view) {
            viewGetConfig(&cfg, &view->opt, 0);
        } else {
            viewGetConfig(&cfg, NULL, 0);
        }
    } else {
        styGetConfig(&cfg, NULL, STY_FG_ALL|STY_LINE_ALL, 1, 0, NULL);
        viewGetConfig(&cfg, NULL, 0);
    }
        
    if (cPtr) {
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_flags");
        sprintf(buf, "%d", cPtr->flags);
        Tcl_DStringAppendElement(&cfg, buf);
        Tcl_DStringEndSublist(&cfg);
        
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_object");
        Tcl_DStringAppendElement(&cfg, cPtr->object);
        Tcl_DStringEndSublist(&cfg);
    }
        
    Tcl_DStringResult(interp, &cfg);
    return TCL_OK;
}

static int
infoContour(Tcl_Interp *interp,
	    ContourP cPtr,
	    int argc,
	    char **argv )
{
    Tcl_DString cmd;

    Tcl_DStringInit(&cmd);
    Tcl_DStringAppendElement(&cmd, "ctrinfo");
    Tcl_DStringAppendElement(&cmd, cPtr->name);
    if (Tcl_Eval(interp, Tcl_DStringValue(&cmd)) != TCL_OK) {
        return TCL_ERROR;
    }

    return infoLevel(interp, cPtr->level);
}

static int
infoLevel(Tcl_Interp *interp,
          LevelP lPtr)
{
    if (lPtr == NULL) {
	return TCL_OK;
    }
    sprintf(buf, "\t %#.3g 0x%2.2x %d\n", lPtr->value, lPtr->flags, lPtr->ns);
    Tcl_AppendResult(interp, buf, NIL);
    return infoLevel(interp, lPtr->next);
}

static int
plotContour(Tcl_Interp *interp,
            ViewP view,
	    ContourP cPtr,
	    int argc,
	    char **argv )
{
    int force;
    int i;
    size_t length;
    LevelP lPtr;
    StyleT style;
    StyleP sPtr;

    if (!view) return TCL_ERROR;

    if (argc > 1 ) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			 " plot ?-force?\"", NIL);
	return TCL_ERROR;
    } else if (argc == 1 && (length = strlen(argv[0])) > 0) {
	if ( strncmp(argv[0],"-force",length) == 0 ) {
	    force = TRUE;
	} else {
	    Tcl_AppendResult(interp, "bad option \"", argv[0],
			     "\": must be -force", NIL);
	    return TCL_ERROR;
	}
    } else {
	force = FALSE;
    }

    for (i=0, lPtr = cPtr->level; lPtr; ++i, lPtr = lPtr->next) {
        if (lPtr->tag) continue;
	if ( !force && (lPtr->flags & M_LVL_PLOT_DONE)) {
	    continue;
	}
	if (cPtr->flags & M_CTR_HAS_COLORRAMP) {
            styClean(&style, FALSE);
	    styFillInterpol(interp, &style, i, cPtr->nl,
                            &cPtr->style[0], &cPtr->style[1]);
            sPtr = &style;
	} else {
            sPtr = &cPtr->style[0];
	}
#ifdef DEBUG
        pltMessage("inf", "level %8.3f  ", lPtr->value);
#endif
        lPtr->tag = dlAlloc(view, DL_SEGMENT, DL_MAIN_PLANE);
        dlMkSegment(lPtr->tag, lPtr->segments, sPtr);
	lPtr->flags |= M_LVL_PLOT_DONE;
    }
    pltUpdateView(view);
    return TCL_OK;

}

static int
clearContour(Tcl_Interp *interp,
             ViewP view,
             ContourP cPtr)
{
    LevelP lPtr;
        
    for (lPtr = cPtr->level; lPtr; lPtr = lPtr->next) {
        lPtr->flags &= ~ M_LVL_PLOT_DONE;
        if (lPtr->tag != 0) {
            dlRemove(view, lPtr->tag);
            lPtr->tag = 0;
        }
    }
    return TCL_OK;
}
