#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/ctrEngine.c,v 2.10 1996/07/25 09:14:52 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif


/*
 * $Log: ctrEngine.c,v $
 * Revision 2.10  1996/07/25 09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.2  1996/07/25  05:54:26  richter
 * handle vertac
 *
 * Revision 2.9.1.1  1996/05/19  15:55:48  richter
 * enhanced segments
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:05  richter
 * First semi-public beta release
 *
 *
 */

/*
 *======================================================================
 *
 * Module: conEngine
 *
 *======================================================================
 */
/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration			*/
#include "xdgraph.h"		/* type definitions and array sizes	*/
#include "ctrProto.h"
#include "segProto.h"
#include "pltProto.h"

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static void contour(void);
static int curve(int ixp, int iyp, int isp);
static void exit1(char *amess);
static int interpol(int ix, int iy, int isp );
/*
 *----------------------------------------------------------------------
 * variables with internal linkage
 *----------------------------------------------------------------------
 */

static SegmentT *curseg;/* pointer to the currently active segment 	*/

static GridT   *grid;	/* pointer to the raw data 		*/

static struct {
    double xmin, xmax, ymin, ymax, xincr, yincr;
} scldat;			/* */

static struct {
    double cl, vertac;
    int kx, mx, ky, my, iz, nr, ixo, iyo, iso;
} crptpm;			/* */

static struct {
    int pc, lopc;
} logic;			/* */

static int record[recordSize];  /* */

int
ctrEngine( GridT *g, ContourP c )
{
    LevelP  cont;

    grid  = g;
    cont = c->level;

    crptpm.iz = grid->nx >= grid->ny ? grid->nx : grid->ny;
    crptpm.kx = 0;
    crptpm.ky = 0;
    crptpm.mx = grid->nx - 1;
    crptpm.my = grid->ny - 1;

    scldat.xmin = (double) crptpm.kx * grid->xran / (double) (grid->nx - 1)
	- grid->xran/2.;
    scldat.xmax = (double) crptpm.mx * grid->xran / (double) (grid->nx - 1)
	- grid->xran/2.;
    scldat.ymin = (double) crptpm.ky * grid->yran / (double) (grid->ny - 1)
	- grid->yran/2.;
    scldat.ymax = (double) crptpm.my * grid->yran / (double) (grid->ny - 1)
	- grid->yran/2.;
    scldat.xincr = (scldat.xmax - scldat.xmin) / (crptpm.mx - crptpm.kx);
    scldat.yincr = (scldat.ymax - scldat.ymin) / (crptpm.my - crptpm.ky);

    /*
     * Call contour for each contour
     */
    while (cont) {
	if ( ! (cont->flags & M_LVL_CALC_DONE) ) {
	    if ( cont->segments ) {
		segFree( &cont->segments );
	    }
	    if ( ! (curseg = cont->segments = segAlloc(0, 0)) ){
		Tcl_AppendResult(interp, "couldn't allocate segment",
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    curseg->p.point[0].x = 0.;
	    curseg->p.point[0].y = 0.;
	    crptpm.cl = cont->value;
            crptpm.vertac = c->vertac;
	    crptpm.nr = 0;
	    contour();
	    cont->flags |= M_LVL_CALC_DONE;
	}
	cont = cont->next;
    }
    return TCL_OK;
}

/*
 * The three routines, contour, curve and interpol form a general
 * contour mapping package.
 *
 * This program is based on M.O. Dayhoff's paper
 *     'A Contour Map Program For X-Ray Crystallography'
 *     Oct. 1963 Communications of the ACM.
 * First programmed by Bruce Langdon, plasma physics lab,
 * Princeton university, nov. 1966.
 */

static void
contour(void)
{
    int i, j;
    int id, com;
    int docurve;

    /*
     * scan right edge, bottom to top
     */
    for (j=crptpm.ky+1; j <= crptpm.my; ++j) {
	if (grid->pl[crptpm.mx + (j-1)*grid->nx] < crptpm.cl && 
	    grid->pl[crptpm.mx +     j*grid->nx] >= crptpm.cl) {
	    if (curve( crptpm.mx, j, 7)) return;
	}
    }

    /*
     * scan top edge, right to left
     */
    for (i = crptpm.mx-1; i >= crptpm.kx; --i) {
	if (grid->pl[i+1 + crptpm.my*grid->nx] < crptpm.cl && 
	    grid->pl[  i + crptpm.my*grid->nx] >= crptpm.cl) {
	    if (curve( i, crptpm.my, 5)) return;
	}
    }

    /*
     * scan left edge top to bottom
     */
    for (j = crptpm.my-1; j >= crptpm.ky; --j) {
	if (grid->pl[crptpm.kx + (j+1)*grid->nx] < crptpm.cl &&
	    grid->pl[crptpm.kx +     j*grid->nx] >= crptpm.cl) {
	    if (curve( crptpm.kx, j, 3)) return;
	}
    }

    /*
     * scan bottom edge and interior points
     */
    for (j = crptpm.ky; j <= crptpm.my-1 ; ++j) {
	for (i = crptpm.kx+1; i <= crptpm.mx; ++i) {
	    if (grid->pl[i-1 + j*grid->nx] >= crptpm.cl ||
		grid->pl[  i + j*grid->nx] < crptpm.cl) {
		continue;
	    }
	    docurve = !crptpm.nr;
	    if (!docurve) {
		docurve = TRUE;
		com = crptpm.iz * i + j;
		for (id = 0; id < crptpm.nr; ++id) {
		    if (record[id] == com) {
			docurve = FALSE;
			break;
		    }
		}
	    }
	    if (docurve) {
		if (curve( i, j, 1)) return;
	    }
	}
    }
    return;
}

/*
 *
 */
static int
curve(int ixp,
      int iyp,
      int isp)
{
    static int inx[8] = { -1,-1,0,1,1,1,0,-1 };
    static int iny[8] = { 0,1,1,1,0,-1,-1,-1 };

    /* Local variables */
    static int iret;
    static int dx, dy, is, ix, iy;
    static int lpc;

    is = isp;
    ix = ixp;
    iy = iyp;
    crptpm.ixo = ix;
    crptpm.iyo = iy;
    crptpm.iso = is;
    logic.lopc = FALSE;
    iret = interpol(ix, iy, is );
    if (iret == 1 || iret == 2) {
	exit1(" Bad return from interpol.");
    }
    if (iret == 3) {
	goto L3;
    }

/*    ROTATE */

L1:
    is = is % 8 + 1;
    dx = inx[is - 1];
    dy = iny[is - 1];

/*    FIND PLOT POINT */

L2:
    iret = interpol(ix, iy, is );
    if (iret == 1) {

/*    DIAG FAIL */

	is = is % 8 + 1;
	logic.lopc = TRUE;
	iret = interpol(ix + inx[is - 1], iy + iny[is - 1], is - 3 );
	if (iret == 1 || iret == 2) {
	    exit1(" Bad return from interpol/");
	}
	if (iret == 3) {
	    goto L10;
	}
	ix += dx;
	iy += dy;
	is = (is + 3) % 8 + 1;
	goto L2;
    } else if (iret == 2) {

/*    NON-DIAG FAIL */

	ix += dx;
	iy += dy;
	is = (is + 4) % 8 + 1;
	lpc = logic.pc;
	iret = interpol(ix, iy, is );
	if (iret == 2) {
	    exit1(" Bad return from interpol.");
	}
	if (iret == 3) {
	    goto L10;
	}
	if (! (lpc && logic.pc)) {
	    goto L1;
	}

/*   PLOT POINT SWITCH */

	segSwitchSegment(&curseg);
	goto L1;
    }
    if (iret == 3) {
	goto L10;
    }
    if (is != 1) {
	goto L1;
    }

/*   RECORD A 'CENTER' */

L3:
    if (crptpm.nr >= recordSize) {
	pltMessage("ERR", "Too many points in ctrEngine. Program must be re-dimensioned\n");
	segFinishSegment(&curseg);
	return 1;
    } else {
	record[crptpm.nr] = crptpm.iz * ix + iy;
	++crptpm.nr;
	goto L1;
    }
L10:
    segFinishSegment(&curseg);
    return 0;
} /* curve */

static void
exit1(char *amess)
{

/*    TERMINATES PLOTTING IN CASE OF AN ABORT. */
    pltMessage("err", "%s\n%s\n", amess, "ctrEngine stops.");
    exit(0);
}

/*
 *
 */
static int
interpol(int ix,
       int iy,
       int isp )
{
    /* Initialized data */

    static int inx[8] = { -1,-1,0,1,1,1,0,-1 };
    static int iny[8] = { 0,1,1,1,0,-1,-1,-1 };

    /* Local variables */
    static double v, cp;
    static int dx, dy, is;
    static int ix1, iy1;
    static double fdx, fdy, fix, fiy;

    is = isp;
    if (is < 1) {
	is += 8;
    }
    dx = inx[is - 1];
    dy = iny[is - 1];
    fdx = (double) dx * scldat.xincr;
    fdy = (double) dy * scldat.yincr;
    fix = (double) (ix - crptpm.kx) * scldat.xincr + scldat.xmin;
    fiy = (double) (iy - crptpm.ky) * scldat.yincr + scldat.ymin;
    ix1 = ix + dx;
    iy1 = iy + dy;
    if (ix1 < crptpm.kx || ix1 > crptpm.mx ||
	iy1 < crptpm.ky || iy1 > crptpm.my) {
	return 3;
    }
    if (dx * dy == 0) {
	/*
	 * non-diagonal case, check for fail
	 */
	if (grid->pl[ix1 + iy1 * grid->nx] >= crptpm.cl) {
	    return 2;
	}
	v = grid->pl[ix + iy * grid->nx];
	if (dx == 0) {
	    segAddSegment(&curseg, fix,
                          fiy + fdy * ((v - crptpm.cl)
                                       / (v - grid->pl[ix + iy1 * grid->nx])),
                          crptpm.cl*crptpm.vertac);
	} else {
	    segAddSegment(&curseg,
                          fix + fdx * ((v - crptpm.cl)
                                       / (v - grid->pl[ix1 + iy * grid->nx])),
                          fiy, crptpm.cl*crptpm.vertac);
	}
	logic.pc = FALSE;
    } else {
	/*
	 * diagonal case
	 */
	cp = (grid->pl[ix + iy * grid->nx] + grid->pl[ix1 + iy * grid->nx]
	      + grid->pl[ix + iy1 * grid->nx] + grid->pl[ix1 + iy1 * grid->nx])
	  / (double)4.;
	if (cp >= crptpm.cl || logic.lopc) {
	    /*
	     * contour passes on far side of center point
	     */
	    if (grid->pl[ix1 + iy1 * grid->nx] >= crptpm.cl &&
		cp >= crptpm.cl) {
		return 1;
	    }
	    v = (grid->pl[ix1 + iy1 * grid->nx] - crptpm.cl) /
	      (grid->pl[ix1 + iy1 * grid->nx] - cp) / (double)2.;
	    segAddSegment(&curseg, ((double)1. - v) * fdx + fix,
                          ((double)1. - v) * fdy + fiy,
                          crptpm.cl*crptpm.vertac);
	    logic.pc = TRUE;
	    logic.lopc = FALSE;
	} else {
	    /*
	     * contour passes on near side of center point
	     */
	    v = (grid->pl[ix + iy * grid->nx] - crptpm.cl) /
	      (grid->pl[ix + iy * grid->nx] - cp) / (double)2.;
	    segAddSegment(&curseg, v * fdx + fix, v * fdy + fiy,
                          crptpm.cl*crptpm.vertac);
	    logic.pc = FALSE;
	}
    }
    if (ix == crptpm.ixo && iy == crptpm.iyo && is == crptpm.iso) {
	return 3;
    }
    return 0;
}
