#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/drvGTRedraw.c,v 2.17 2000/01/20 09:40:06 richter Exp $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: drvGTRedraw.c,v $
 * Revision 2.17  2000/01/20 09:40:06  richter
 * work around some problems with text/line drawing using Mesa 3.1
 *
 * Revision 2.16  1996/12/19 11:02:28  richter
 * added the CLEAN_VIEW_STYLE flag
 *
 * Revision 2.15  1996/11/07  17:08:49  richter
 * MESA seems to have some problems with glPushAttrib(GL_ENABLE_BIT)
 * (gtDrawHeightfield): don't use display lists for the moment because
 *                      dlRmHeight doesn't clean up as it should
 * added some debug output available with DEBUG_LEVEL
 *
 * Revision 2.14  1996/11/06  10:55:38  richter
 * disable depthcueing for xy diagrams
 * set border only for xy diagrams
 * draw black border for xydiagrams
 *
 * Revision 2.13  1996/11/03  16:43:40  richter
 * recognize polygon mode
 *
 * Revision 2.12  1996/09/23  12:34:02  richter
 * dont preset drawin color
 *
 * Revision 2.11  1996/09/05  15:10:07  richter
 * use the new generic display list
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.3  1996/07/25  06:09:53  richter
 * new flag DONT_USE_DISPLAY_LIST
 * new options: fill, culling, black_bg
 * better performance analysis (count polygons)
 * don't use display lists for dumps
 * apply high and low clipping planes for heightfields
 * allow color maps for height fields (optimizations needed!)
 *
 * Revision 2.9.1.2  1996/07/03  10:55:10  richter
 * fixed stipple
 *
 * Revision 2.9.1.1  1996/05/20  11:20:33  richter
 * lots of changes
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:06  richter
 * First semi-public beta release
 *
 *
 */

/*
 *======================================================================
 *
 * Module: drvGTRedraw
 *
 *======================================================================
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#ifdef WIN32                    /* Cljf for Windows compilation */
#include <windows.h>
#endif
#include "conf.h"		/* global configuration			*/
#include "xdgraph.h"		/* type definitions and array sizes	*/
#include "drvProto.h"
#include "styProto.h"
#include <GL/gl.h> 
#include <GL/glu.h>
#ifdef XD_USE_PERFORMANCE
#  define _BSD_COMPAT
#  include <sys/time.h>
#endif

#define REDRAW_PENDING          1
#define EXPOSE_PENDING          2
#define CONFIGURE_PENDING       4
#define ROTATE_PENDING          8
#define DONT_USE_DISPLAY_LIST  16
#define CLEAN_VIEW_STYLE       32

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static void gtScanDisplayList(ViewP view, DLobjP);
static void gtDrawDebugSpheres(void);
static void gtDrawSegment(DLSegment *p);
static void gtDrawHeightfield(DLHeight *h);
static void gtDrawRectangle(DLRectangle *r);
static void gtDrawSphere(DLSphere *s);
static void gtDrawCylinder(DLCylinder *s);
static void gtDrawText(DLText *t);
static void gtDrawTriangle(DLTriangle *tri);
static void gtDrawColor(GtColor *c);
static void gtDrawLinewidth(GtLinewidth *l);
static void gtDrawLinestipple(GtLinestipple *l);
static void gtApplyStyle(int mask, StyleP style);

/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */
extern char buf[1000];
extern struct {
    int twoside ;
    int antialias ;
    int flat;
    int depthcue;
    int performance;
    int culling;
    int black_bg;
    int configure_pending;
} gtdriver_options;

/*
 *----------------------------------------------------------------------
 * variables with internal linkage
 *----------------------------------------------------------------------
 */
static int npolygon = 0;
static int use_dl = 0;

void
gtRedraw(ViewP view,
         int flags)
{
    DrvgtP ddp;
#ifdef XD_USE_PERFORMANCE
    struct timeval von, bis;
#endif
    static double depth = 10.;
    static double near1 = 10.;
    double mat[16];
    int i, is_xyview, use_ortho;

    if (!view) return;
    is_xyview = (view->type == VIEW_XYDIAG);
    use_ortho = is_xyview || (view->type == VIEW_CONTOUR) ||
        (view->type == VIEW_RELIEF);
             
    ddp = (DrvgtP) view->drdat;

#ifdef XD_USE_PERFORMANCE
    if (gtdriver_options.performance) {
        npolygon = 0;
        gettimeofday(&von, 0);
    }
#endif

    use_dl = !(flags & DONT_USE_DISPLAY_LIST);
    if ((flags & EXPOSE_PENDING) || (flags & CONFIGURE_PENDING) ||
        gtdriver_options.configure_pending) {
        double w, h, aw, ah;
        
        gtdriver_options.configure_pending = 0;
        /*
         * viewport is the whole window
         */
        glViewport(0, 0, ddp->pixw, ddp->pixh);
        /*
         * setup projection transformation which retains the correct
         * aspect ratio
         */
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        aw = (double) ddp->pixw / (double) ddp->pixh;
        ah = 1.;
        if (aw < 1.) {
            ah = 1./aw;
            aw = 1.;
        }
        if (use_ortho) {
            if (is_xyview) {
                w = aw * (view->opt.width+2.)/2.;
                h = ah * (view->opt.height+2.)/2.;
            } else {
                w = aw * view->opt.width/2.;
                h = ah * view->opt.height/2.;
            }
            depth = 2.*sqrt(w*w+h*h);
            near1 = 3.*depth;
            glOrtho(-w, w, -h, h, near1, depth+near1);
        } else { 
            w = aw * view->opt.width/2.;
            h = ah * view->opt.height/2.;
            depth = 2.*sqrt(w*w+h*h);
            near1 = 3.*depth;
            glFrustum(-w, w, -h, h, near1, depth+near1);
        }
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        glEnable(GL_NORMALIZE);
        glEnable(GL_DEPTH_TEST);
        glEnable(GL_LIGHTING);
        glEnable(GL_COLOR_MATERIAL); /* */

        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
        if (gtdriver_options.flat) {
            glShadeModel(GL_FLAT);
        } else {
            glShadeModel(GL_SMOOTH);
        }
        
        
        if (gtdriver_options.antialias) {
            glEnable(GL_LINE_SMOOTH);
            glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
            glEnable(GL_POINT_SMOOTH);
            glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
            /* glLineWidth((GLfloat) 2.); /* */
            /* glEnable(GL_POLYGON_SMOOTH); /* */
            /* glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); /* */
            /* glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST); /* */
        } else {
            glDisable(GL_LINE_SMOOTH);
        }
        
        for (i=0; i<ddp->lightnum; i++) {
            glEnable(GL_LIGHT0+i);
            glLightfv(GL_LIGHT0+i, GL_POSITION, ddp->light[i].position);
            glLightfv(GL_LIGHT0+i, GL_DIFFUSE, ddp->light[i].diffuse);
        }
        for (i=ddp->lightnum; i<5; i++) {
            glDisable(GL_LIGHT0+i);
        }

        if (gtdriver_options.depthcue && !is_xyview) {
            GLfloat fogColor[4] = {0., 0., 0., 1.0}; 
            glEnable(GL_FOG);
            glHint(GL_FOG_HINT, GL_DONT_CARE);
            glFogi(GL_FOG_MODE, GL_LINEAR);
            glFogfv(GL_FOG_COLOR, fogColor);
            glFogf(GL_FOG_START, near1);
            glFogf(GL_FOG_END, 2.*(depth+near1));
        } else {
            glDisable(GL_FOG);
        }
        
        if (gtdriver_options.twoside) {
            GLfloat back_diff[4] = {0.1, 0.3, 0.2, 1.};
            glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
            glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
            glMaterialfv(GL_BACK, GL_AMBIENT_AND_DIFFUSE, back_diff);
        } else {
            glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 0);
            glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
        }

        if (gtdriver_options.culling) {
            glEnable(GL_CULL_FACE);
        } else {
            glDisable(GL_CULL_FACE);
        }
    }

    if (!is_xyview && flags & ROTATE_PENDING) {
         glPushMatrix();
         /* printf("Rotate: %f %f\n", ddp->xang, ddp->yang); /* */
         glLoadIdentity();
         glRotatef(ddp->xang, 1.0, 0.0, 0.0);
         glRotatef(ddp->yang, 0.0, 1.0, 0.0);
         glMultMatrixf(ddp->rotmat);
         glGetFloatv(GL_MODELVIEW_MATRIX, ddp->rotmat);
         ddp->xang = 0.;
         ddp->yang = 0.;
         glPopMatrix();
    }
         

    /*
     * gtRedraw only gets called when REDRAW_PENDING was set,
     * so there's no need to check for that.
     *
     * setup viewing transformation
     */
    if (gtdriver_options.black_bg) {
        glClearColor(0., 0., 0., 0.);
    } else {
        glClearColor(1., 1., 1., 0.);
    }
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
    
    glPushMatrix();
    glLoadIdentity();
    glTranslated(0., 0., -(depth/2. + near1)); /* */
    glMultMatrixf(ddp->rotmat);

    /* gtDrawDebugSpheres(); /* */
    
    glPushMatrix();

    if (view->opt.doscale) {
        glScaled(view->opt.scale, view->opt.scale, view->opt.scale);
    }
    if (view->opt.dotrans) {
        glTranslated(view->opt.trans[0], view->opt.trans[1], view->opt.trans[2]);
    }
    if (view->opt.dorot) {
        int i, k;
        
        for (i=0; i<3; ++i) {
            for (k=0; k<3; ++k) mat[i*4+k] = view->opt.rotm[k][i];
            mat[i*4+3] = 0.;
        }
        for (k=0; k<3; ++k) mat[12+k] = 0.;
        mat[15] = 1.;
        glMultMatrixd(mat);
    }
    if (view->opt.doorig) {
        glTranslated(-view->opt.orig[0], -view->opt.orig[1], -view->opt.orig[2]);
    }
    if (is_xyview) {
        glBegin(GL_LINE_LOOP);
        glColor3d(0., 0., 0.);
        glVertex2d(0., 0.);
        glVertex2d(0., 1.);
        glVertex2d(1., 1.);
        glVertex2d(1., 0.);
        glEnd();
    }
    if (flags & CLEAN_VIEW_STYLE) {
        styClean(&view->style, TRUE);
    }
    gtScanDisplayList(view, view->dl_head);
    glPopMatrix();
    glPopMatrix();
    glFlush ();
#ifdef XD_USE_PERFORMANCE
    if (gtdriver_options.performance) {
        double sec;
        gettimeofday(&bis, 0);
        sec = bis.tv_sec - von.tv_sec;
        sec += (double) (bis.tv_usec - von.tv_usec)/1000000.;
        printf(" %.2f frames/sec, %d polygons %.0f polygons/sec\n",
               1./sec, npolygon, ((double)npolygon)/sec);
    }
#endif
}

static void
gtDrawDebugSpheres(void)
{
    static GLUquadricObj *s1 = NULL;
    static GLUquadricObj *s2 = NULL;
    static GLUquadricObj *s3 = NULL;

    if (!s1) s1 = gluNewQuadric();
    if (!s2) s2 = gluNewQuadric();
    if (!s3) s3 = gluNewQuadric();

    glPushMatrix();
    glTranslated(0., 0., 2.);
    glColor3d(1., 1., 1.);
    gluSphere(s1, 1., 16, 16);
    glPopMatrix();

    glPushMatrix();
    glTranslated(5., 0., 2.);
    glColor3d(1., 0., 0.);
    gluSphere(s2, 1., 16, 16);
    glPopMatrix();
    
    glPushMatrix();
    glTranslated(0., 5., 2.);
    glColor3d(0., 1., 1.);
    gluSphere(s3, 1., 16, 16);
    glPopMatrix();
}

static void
gtScanDisplayList(ViewP view,
                  DLobjP o)
{
    int mask;
    StyleT style = {
        {NULL, UNKNOWNMODEL, {0., 0., 0.}},
        {0., 0, NULL},
        {0., NULL, NULL}
    };
    
    for (; o; o=o->next) {
        styClean(&style, TRUE);
        mask = 0;
#if DEBUG_LEVEL > 2
        printf("dl type: %d\n", o->type);
#endif
        switch (o->type) {
        case DL_SEGMENT:
            styApply(&style, &mask, &view->style, &o->style,
                     STY_FG_ALL|STY_LINE_ALL);
            if (mask) gtApplyStyle(mask, &style);
            gtDrawSegment(o->object.Segment);
            break;
        case DL_HEIGHTFIELD:
            styApply(&style, &mask, &view->style, &o->style,
                     STY_FG_ALL|STY_POLYGON_MODE);
            if (mask) gtApplyStyle(mask, &style);
            gtDrawHeightfield(o->object.Height);
            break;
            /*
        case DL_RECTANGLE: 
            styApply(&style, &mask, &view->style, &o->style,
                     STY_FG_ALL);
            if (mask) gtApplyStyle(mask, &style);
            gtDrawRectangle(o->object.Rectangle);
            break;
            */
        case DL_TRIANGLE: 
            styApply(&style, &mask, &view->style, &o->style,
                     STY_FG_ALL|STY_POLYGON_MODE);
            if (mask) gtApplyStyle(mask, &style);
            gtDrawTriangle(o->object.Triangle);
            break;
        case DL_SPHERE: 
            styApply(&style, &mask, &view->style, &o->style,
                     STY_FG_ALL);
            if (mask) gtApplyStyle(mask, &style);
            gtDrawSphere(o->object.Sphere);
            break;
        case DL_CYLINDER: 
            styApply(&style, &mask, &view->style, &o->style,
                     STY_FG_ALL);
            if (mask) gtApplyStyle(mask, &style);
            gtDrawCylinder(o->object.Cylinder);
            break;
        case DL_TEXT: 
            styApply(&style, &mask, &view->style, &o->style,
                     STY_FG_ALL|STY_FONT_ALL);
            if (mask) gtApplyStyle(mask, &style);
            gtDrawText(o->object.Text);
            break;
        }
    }
}

static void
gtDrawSegment(DLSegment *p)
{
    SegmentP s;
    c3dT *v;
    int n, open;

    if (!p) {
        printf("nil seg.pointer\n");
        return;
    }
    glPushAttrib(GL_ENABLE_BIT);
    glDisable(GL_LIGHTING);
#if !defined(MESA) || defined(MESA_MAJOR_VERSION)
    glDisable(GL_COLOR_MATERIAL); /* don't disable for Mesa 2.x */
#endif
    
    open = 0;
    for (s=p->v; s; s=s->next) {
        if ( s->type == INVALIDSEG || s->np <= 0 ) {
            continue;
        }
        switch (s->type) {
        case MOVESEG: {
            if (open) glEnd();
            glBegin(GL_LINE_STRIP);
            open = 1;
        }; break;
        case DRAWSEG: {
            if (!open) {
                glBegin(GL_LINE_STRIP);
                open = 1;
            }
        }; break;
        case POINTSET: {
            if (!open) {
                glBegin(GL_POINTS);
                open = 1;
            }
        }; break;
        case LINESET: {
            if (!open) {
                glBegin(GL_LINES);
                open = 1;
            }
        }; break;
        }
        if (s->type == LINESET) {
            for (n=0; n < s->np; n++) {
                v = &(s->p.line.a[n]);
                glVertex3d(v->x, v->y, v->z);
                v = &(s->p.line.b[n]);
                glVertex3d(v->x, v->y, v->z);
            }
        } else {
            for (n=0; n < s->np; n++) {
                v = &(s->p.point[n]);
                glVertex3d(v->x, v->y, v->z);
            }
        }
    }
    if (open) glEnd();
    glPopAttrib();
}

static void
gtDrawHeightfield(DLHeight *h)
{
    double lowcut[] = {0., 0., 1., 1.};
    double hihcut[] = {0., 0., -1., 1.};
    int i, k;
 
    if (gtdriver_options.performance) npolygon += (h->nx-1)*(h->ny-1);
    
    if (use_dl && 0) {
        if (h->listid != 0) {
            glCallList(h->listid);
            return;
        }
    
        h->listid = glGenLists(1);
        glNewList(h->listid, GL_COMPILE_AND_EXECUTE);
    }
    
    glPushAttrib(GL_POLYGON_BIT|GL_ENABLE_BIT|GL_CURRENT_BIT);
    if (h->hcut != h->lcut) {
        glEnable(GL_CLIP_PLANE4);
        glEnable(GL_CLIP_PLANE5);
        hihcut[3] = h->hcut*h->vertac;
        lowcut[3] = fabs(h->lcut*h->vertac);
        glClipPlane(GL_CLIP_PLANE4, hihcut);
        glClipPlane(GL_CLIP_PLANE5, lowcut);
    }
    glFrontFace(GL_CW); /* */
    glPushMatrix();
    glTranslated(0., 0., -0.01);

    for (k=0; k < (h->ny-1); ++k) {
        glBegin(GL_QUAD_STRIP); /* */
        for (i=0; i<h->nx; ++i) {
            int me;
            c3dT *v, *n;
            me = i + k*h->nx;
            n = &h->normal[me];
            v = &h->vertex[me];
            if (h->color) glColor3fv(h->color[me].rgb);
            glNormal3d(n->x, n->y, n->z);
            glVertex3d(v->x, v->y, v->z);
            me = i + (k+1)*h->nx;
            n = &h->normal[me];
            v = &h->vertex[me];
            if (h->color) glColor3fv(h->color[me].rgb);
            glNormal3d(n->x, n->y, n->z);
            glVertex3d(v->x, v->y, v->z);
        }
        glEnd();
    }
    glPopMatrix();
    glPopAttrib();
    if (use_dl && h->listid) glEndList();
}

/*
static void
gtEnableDepth(void)
{
    glEnable(GL_DEPTH_TEST);
}

static void
gtDisableDepth(void)
{
    glDisable(GL_DEPTH_TEST);
}
*/

static void
gtDrawRectangle(DLRectangle *r)
{
    glBegin(GL_POLYGON);
    glVertex3d(r->vertex[0].x, r->vertex[0].y, r->vertex[0].z);
    glEnd();
}

static void
gtDrawSphere(DLSphere *s)
{
    static GLUquadricObj *squad = NULL;

    glPushAttrib(GL_POLYGON_BIT);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    if (!squad) squad = gluNewQuadric();
    glPushMatrix();
    glTranslated(s->vertex.x, s->vertex.y, s->vertex.z);
    gluSphere(squad, 0.1, 16, 16);
    glPopMatrix();
    glPopAttrib();
}

static void
gtDrawCylinder(DLCylinder *s)
{
    static GLUquadricObj *squad = NULL;

    glPushAttrib(GL_POLYGON_BIT);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    if (!squad) squad = gluNewQuadric();
    glPushMatrix();
    glTranslated(s->origin.x, s->origin.y, s->origin.z);
    glRotated(s->angle, s->axis.x, s->axis.y, s->axis.z);/* */
    gluCylinder(squad, s->radius, s->radius, s->height, 8, 8);
    glPopMatrix();
    glPopAttrib();
}

static void
gtDrawText(DLText *t)
{
    if (!t) return;
    glPushAttrib(GL_ENABLE_BIT|GL_CURRENT_BIT);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_FOG);
    glDisable(GL_LIGHTING);
#if !defined(MESA) || defined(MESA_MAJOR_VERSION)
    glDisable(GL_COLOR_MATERIAL); /* don't disable for Mesa 2.x */
#endif

    if (gtdriver_options.black_bg) {
        glColor3d(1., 1., 1.);
    } else {
        glColor3d(0., 0., 0.);
    }

    glRasterPos3d(t->vertex.x, t->vertex.y, t->vertex.z);
    glListBase(1000);
    glCallLists(strlen(t->string), GL_BYTE, t->string);

#if defined(MESA_MAJOR_VERSION) && MESA_MAJOR_VERSION==3 && MESA_MINOR_VERSION==1
    /* hack around a bug (?) in Mesa 3.1*/
    glBegin(GL_TRIANGLES);
    if (gtdriver_options.black_bg) {
        glColor3d(0., 0., 0.);
    } else {
        glColor3d(1., 1., 1.);
    }
    glNormal3d(0., 0., 1.);
    glVertex3d(0., 0., 0.);
    glVertex3d(0., .001, 0.);
    glVertex3d(.001, 0., 0.);
    glEnd();
#endif
    glPopAttrib();
}

static void
gtDrawTriangle(DLTriangle *tri)
{
    TriangleP t;

    /* glPushMatrix(); /* */
    int n = 0;
    for (t = tri->head; t; t=t->next) {
        glBegin(GL_POLYGON);
        glNormal3d(t->n1.x, t->n1.y, t->n1.z); /* */
        glVertex3d(t->v1.x, t->v1.y, t->v1.z);
        glNormal3d(t->n2.x, t->n2.y, t->n2.z); /* */
        glVertex3d(t->v2.x, t->v2.y, t->v2.z);
        glNormal3d(t->n3.x, t->n3.y, t->n3.z); /* */
        glVertex3d(t->v3.x, t->v3.y, t->v3.z);
        glEnd();
        if (gtdriver_options.performance) ++npolygon;
    }
    /* glPopMatrix(); /* */
}

static void
gtApplyStyle(int mask,
             StyleP style)
{
    if (mask&STY_FG_ALL) {
        GtColor c;
        ColorP cPtr = &style->fg;

        switch (cPtr->model) {
        case RGBMODEL:
            c.rgb[0] = cPtr->val[0];
            c.rgb[1] = cPtr->val[1];
            c.rgb[2] = cPtr->val[2];
            break;
        case HSVMODEL:
            styCvthsv2rgb((float) cPtr->val[0]/360.,
                          cPtr->val[1], cPtr->val[2],
                          &c.rgb[0], &c.rgb[1], &c.rgb[2]);
            break;
        case YUVMODEL:
            styCvtyuv2rgb(cPtr->val[0], cPtr->val[1], cPtr->val[2],
                          &c.rgb[0], &c.rgb[1], &c.rgb[2]);
            break;
        case MAPMODEL: 
            {
                if (!styCvtindex2rgb(0, (int) cPtr->val[1], &c.rgb[0],
                                     &c.rgb[1], &c.rgb[2])) {
                    c.rgb[0] = 0.77;
                    c.rgb[1] = 0.66;
                    c.rgb[2] = 0.55;
                }
            }
            break;
        }
        gtDrawColor(&c);
    }

    if (mask&STY_LINE_WIDTH) {
        GtLinewidth l;

        l.width = style->line.width;
        gtDrawLinewidth(&l);
    }

    if (mask&STY_LINE_PATTERN) {
        GtLinestipple *l;

        l = gtSetLinestipple(style->line.plength, style->line.pattern);
        gtDrawLinestipple(l);
        ckfree((char *)l);
    }

    if (mask&STY_POLYGON_MODE) {
        switch (style->pg.mode) {
        case PG_POINT:
            glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
            break;
        case PG_LINE:
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            break;
        default:
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
            break;
        }
    }
}

static void
gtDrawColor(GtColor *c)
{
#if DEBUG_LEVEL > 2
    printf("COLOR: %f %f %f\n", c->rgb[0], c->rgb[1], c->rgb[2]); /* */
#endif
    glColor3fv(c->rgb);
}

static void
gtDrawLinewidth(GtLinewidth *l)
{
#if DEBUG_LEVEL > 2
    printf("LINEWIDTH: %f\n", l->width);
#endif
    glLineWidth(l->width);
}

static void
gtDrawLinestipple(GtLinestipple *l)
{
#if DEBUG_LEVEL > 2
    printf("LINESTIPPLE %d 0x%x\n", l->factor, l->mask);
#endif
    if (l->factor > 0) {
        glEnable(GL_LINE_STIPPLE);
        glLineStipple(l->factor, l->mask);
    } else {
        glDisable(GL_LINE_STIPPLE);
    }
}

