#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/drvGTWidget.c,v 2.19 1999/08/22 12:31:39 richter Exp $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: drvGTWidget.c,v $
 * Revision 2.19  1999/08/22 12:31:39  richter
 * a couple of casts
 *
 * Revision 2.18  1999/06/03 12:41:08  richter
 * use prototypes
 *
 * Revision 2.17  1998/02/09 15:26:19  richter
 * error messages to stderr
 *
 * Revision 2.16  1996/12/19  11:03:14  richter
 * added the CLEAN_VIEW_STYLE flag
 *
 * Revision 2.15  1996/12/18  15:30:39  richter
 * (DisplayGlxwin): prototype wath declared with static, actual function was not
 *
 * Revision 2.14  1996/12/04  15:38:01  richter
 * fixed many problems with pixmap dumps (MESA, strange visuals)
 *
 * Revision 2.13  1996/11/29  14:30:57  richter
 * added some XD_TK_VERSION #ifs
 *
 * Revision 2.12  1996/10/11  12:07:24  richter
 * dump non-square pixmaps correctly
 * use two different font sizes for different pixmap sizes
 * (NOTE: font handling needs much more work!)
 *
 * Revision 2.11  1996/09/06  13:48:40  richter
 * dump pixmaps from non-24bit visuals correctly
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.3  1996/07/25  06:08:15  richter
 * new flag DONT_USE_DISPLAY_LIST
 * allow simple visuals for dumps
 * handle cases where not matching visual was found
 *
 * Revision 2.9.1.2  1996/07/03  10:54:02  richter
 * load differnet fonts depending on resolution
 * dump ppm file
 *
 * Revision 2.9.1.1  1996/05/20  11:19:33  richter
 * changes for Tk 4.1
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:06  richter
 * First semi-public beta release
 *
 *
 */

/*
 *======================================================================
 *
 * Module: drvGTWidget
 *
 *======================================================================
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#ifdef WIN32                  /* Cljf for Windows compilation        */
#include <windows.h>
#endif
#include "conf.h"
#include "xdgraph.h"
#include <tk.h>
#include "strProto.h"
#include "vieProto.h"
#include <X11/Xlib.h>
#include <GL/glx.h>
#include <GL/gl.h>
extern void gtRedraw(void *handle, int flags);
extern void gtSetWindowSize(void *handle, int width, int height);
extern void gtSetRotation(void *handle, int vx, int vy);
extern void gtSetOrigin(void *handle, int vx, int vy);
extern void gtSetRotx(void *handle, int vx);
extern void gtSetRoty(void *handle, int vy);

/*
 * A data structure of the following type is kept for each glxwin
 * widget managed by this file:
 */

typedef struct {
    Tk_Window tkwin;		/* Window that embodies the glxwin.  NULL
				 * means window has been deleted but
				 * widget record hasn't been cleaned up yet. */
    Display *display;		/* X's token for the window's display. */
    Tcl_Interp *interp;		/* Interpreter associated with widget. */
#if XD_TK_VERSION >= 40
    Tcl_Command widgetCmd;
#endif
    GLXContext glxcontext; 
    XVisualInfo *visual;
    Colormap colormap;
    void *handle;
    int toplevel;               /* Flag */
    int width;			/* Width to request for window.  <= 0 means
				 * don't request any size. */
    int height;			/* Height to request for window.  <= 0 means
				 * don't request any size. */
    int flags;			/* Various flags;  see below for
				 * definitions. */
    int mask;
    char *className, *handleName, *screenName;
    int initialized;
} Glxwin;
/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static int ConfigureGlxwin(Tcl_Interp *interp, Glxwin *glxwinPtr,
                           int argc, char **argv, int flags);
#if XD_TK_VERSION >= 41
static void DestroyGlxwin(char *clientData);
#else
static void DestroyGlxwin(ClientData clientData);
#endif
static void DisplayGlxwin(ClientData clientData);
static void GlxwinCmdDeletedProc(ClientData clientData);
static void GlxwinEventProc(ClientData clientData, XEvent *eventPtr);
static int  GlxwinWidgetCmd(ClientData clientData,
			    Tcl_Interp *, int argc, char **argv);
static void MapGlxwin(ClientData clientData);
static int GlxwinDump(Tcl_Interp *interp, Glxwin *glxwinPtr,
                      int argc, char **argv);
static void LoadFontGlxwin(Glxwin *glxwinPtr, GLXDrawable drawable,
                           GLXContext glxcontext, int size);
static void dumpPixmap(Display *dpy, XVisualInfo *vi,
                       Colormap cmap, Pixmap pixmap, char *file);
/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */

/*
 *----------------------------------------------------------------------
 * variables with internal linkage
 *----------------------------------------------------------------------
 */
/*
 * Currently there is no way to dynamically configure the requirements
 * of the GLX window. You have to change the following static variables
 * to get the set of OpenGL window attributes you want.
 */
static int attributeList[] = {GLX_RGBA,
                              GLX_DOUBLEBUFFER,
                              GLX_DEPTH_SIZE, 8,
                              GLX_RED_SIZE,   1, /*3,*/
                              GLX_GREEN_SIZE, 1, /*3,*/
                              GLX_BLUE_SIZE,  1, /*2,*/
                              None };

static int has_glx = -1;


#define REDRAW_PENDING          1
#define EXPOSE_PENDING          2
#define CONFIGURE_PENDING       4
#define ROTATE_PENDING          8
#define DONT_USE_DISPLAY_LIST  16
#define CLEAN_VIEW_STYLE       32

#define DUMP_FLAGS EXPOSE_PENDING|DONT_USE_DISPLAY_LIST|CLEAN_VIEW_STYLE

#define TOPLEVEL           TK_CONFIG_USER_BIT

/*
 * Information used for argv parsing.
 */

static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_STRING, "-class", "class", "Class",
        "Glxwin", Tk_Offset(Glxwin, className), 0},

    {TK_CONFIG_STRING, "-handle", "handle", "Handle",
        "", Tk_Offset(Glxwin, handleName), 0},

    {TK_CONFIG_PIXELS, "-height", "height", "Height",
       0, Tk_Offset(Glxwin, height), 0},

    {TK_CONFIG_STRING, "-screen", "screen", "Screen",
        "", Tk_Offset(Glxwin, screenName),
        TOPLEVEL|TK_CONFIG_NULL_OK},
  
    {TK_CONFIG_PIXELS, "-width", "width", "Width",
       0, Tk_Offset(Glxwin, width), 0},

    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
       (char *) NULL, 0, 0}
};

/*
 *--------------------------------------------------------------
 *
 * GlxwinCmd --
 *
 *	This procedure is invoked to process the "glxwin" Tcl
 *	command.  It creates a new "glxwin" widget.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	A new widget is created and configured.
 *
 *--------------------------------------------------------------
 */

int
GlxwinCmd(ClientData clientData, /* Main window associated with interpreter. */
          Tcl_Interp *interp,    /* Current interpreter. */
          int argc,              /* Number of arguments. */
          char **argv)           /* Argument strings. */
{
    Tk_Window main = (Tk_Window) clientData;
    Glxwin *glxwinPtr;
    Tk_Window new = NULL;
    Display *dpy;
    int i, length, toplevel;
    GLXContext cx;
    XVisualInfo *vi;
    Colormap cm;
    char *className, *screenName, *arg, c;
    
    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
		argv[0], " pathName ?options?\"", (char *) NULL);
	return TCL_ERROR;
    }

    toplevel = (argv[0][0] == 't');

    className = screenName = NULL;
    for (i = 2; i < argc; i += 2) {
        arg = argv[i];
        length = strlen(arg);
        if (length < 2) {
            continue;
        }
        c = arg[1];
        if ((c == 'c') && (strncmp(arg, "-class", strlen(arg)) == 0)
                && (length >= 3)) {
            className = argv[i+1];
        } else if ((c == 's') && toplevel
                && (strncmp(arg, "-screen", strlen(arg)) == 0)) {
            screenName = argv[i+1];
        }
    }

    if (screenName == NULL) {
        screenName = (toplevel) ? "" : NULL;
    }
    
    new = Tk_CreateWindowFromPath(interp, main, argv[1], screenName);
    if (new == NULL) {
	return TCL_ERROR;
    }

    if (className == NULL) {
        className = Tk_GetOption(new, "class", "Class");
        if (className == NULL) {
            className = "GLxwin";
        }
    }
    Tk_SetClass(new, className);

    dpy = Tk_Display(new);
    if (has_glx < 0) {
        has_glx = glXQueryExtension(dpy, NULL, NULL);
    }
    if (!has_glx) {
        Tcl_AppendResult(interp, "GLX extension is not available",
                         (char *) NULL);
	goto error;
    }

    vi = glXChooseVisual(dpy, DefaultScreen(dpy), attributeList);
    if (! vi) {
	Tcl_AppendResult(interp, "No visual", (char *) NULL);
	goto error;
    }
#ifdef DEBUG
    printf("vi id=0x%x class=%d\n", vi->visualid, vi->class);
#endif
    cx = glXCreateContext(dpy, vi, 0, GL_TRUE); /* */
    /* cx = glXCreateContext(dpy, vi, 0, GL_FALSE); /* */
    cm = XCreateColormap(dpy, RootWindow(dpy, vi->screen), vi->visual,
                         AllocNone);

    if ( ! Tk_SetWindowVisual(new, vi->visual, vi->depth, cm) ) {
	fprintf(stderr, "Tk could not set the requested visual (0x%x %d)\n",
	vi->visual, vi->depth);
    }
    /*
     * ? Tk_SetWindowColormap(new, cm);
     */
    /*
     * Allocate and initialize the widget record.
     */

    glxwinPtr = (Glxwin *) ckalloc(sizeof(Glxwin));
    glxwinPtr->tkwin = new;
    glxwinPtr->display = dpy;
    glxwinPtr->interp = interp;
    glxwinPtr->width  = 0;
    glxwinPtr->height = 0;
#if XD_TK_VERSION >= 40
    glxwinPtr->widgetCmd =
#endif
        Tcl_CreateCommand(interp, Tk_PathName(glxwinPtr->tkwin),
                          GlxwinWidgetCmd, (ClientData) glxwinPtr,
                          GlxwinCmdDeletedProc);
                          /* (void (*)()) NULL); */
    glxwinPtr->glxcontext = cx;
    glxwinPtr->visual = vi;
    glxwinPtr->colormap = cm;
    glxwinPtr->className = NULL;
    glxwinPtr->mask = (toplevel) ? TOPLEVEL : 0;
    glxwinPtr->screenName = NULL;
    glxwinPtr->handleName = NULL;
    glxwinPtr->flags = 0;

    Tk_CreateEventHandler(glxwinPtr->tkwin,
                          ExposureMask|StructureNotifyMask|FocusChangeMask,
                          GlxwinEventProc, (ClientData) glxwinPtr);
    
    if (ConfigureGlxwin(interp, glxwinPtr, argc-2, argv+2, 0) != TCL_OK) {
        goto error;
    }

    glxwinPtr->handle = (void *) viewFindByName(glxwinPtr->handleName, NULL);
    if (!glxwinPtr->handle) {
        goto error;
    }
    
    if (toplevel) {
#if XD_TK_VERSION >= 41
        Tcl_DoWhenIdle(MapGlxwin, (ClientData) glxwinPtr);
#else
        Tk_DoWhenIdle(MapGlxwin, (ClientData) glxwinPtr);
#endif
    }

    interp->result = Tk_PathName(glxwinPtr->tkwin);
    return TCL_OK;

error:
    if (new != NULL) {
        Tk_DestroyWindow(new);
    }
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * GlxwinWidgetCmd --
 *
 *	This procedure is invoked to process the Tcl command
 *	that corresponds to a widget managed by this module.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int
GlxwinWidgetCmd(ClientData clientData, /* Information about glxwin widget. */
                Tcl_Interp *interp,    /* Current interpreter. */
                int argc,              /* Number of arguments. */
                char **argv)           /* Argument strings. */
{
    Glxwin *glxwinPtr = (Glxwin *) clientData;
    int result = TCL_OK;
    int length;
    char c;


    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    Tk_Preserve((ClientData) glxwinPtr);

    c = argv[1][0];
    length = strlen(argv[1]);
#if XD_TK_VERSION >= 40
    if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)
        && (length >= 2)) {
        if (argc != 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                             argv[0], " cget option\"",
                             (char *) NULL);
            result = TCL_ERROR;
            goto done;
        }
        result = Tk_ConfigureValue(interp, glxwinPtr->tkwin, configSpecs,
                (char *) glxwinPtr, argv[2], glxwinPtr->mask);
    } else
#endif
        if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)) {
	if (argc == 2) {
	    result = Tk_ConfigureInfo(interp, glxwinPtr->tkwin, configSpecs,
		    (char *) glxwinPtr, (char *) NULL, glxwinPtr->mask);
	} else if (argc == 3) {
	    result = Tk_ConfigureInfo(interp, glxwinPtr->tkwin, configSpecs,
		    (char *) glxwinPtr, argv[2], glxwinPtr->mask);
	} else {
            /*
             * Don't allow the options -class or -screen to be changed.
             */
            int i;
            for (i = 2; i < argc; i++) {
                length = strlen(argv[i]);
                if (length < 2) {
                    continue;
                }
                c = argv[i][1];
                if (((c == 'c') && (strncmp(argv[i], "-class", length) == 0))
                    || ((c == 's') && (glxwinPtr->mask == TOPLEVEL)
                        && (strncmp(argv[i], "-screen", length) == 0))) {
                    Tcl_AppendResult(interp, "can't modify ", argv[i],
                                     " option after widget is created",
                                     (char *) NULL);
                    result = TCL_ERROR;
                    goto done;
                }
            }
	    result = ConfigureGlxwin(interp, glxwinPtr, argc-2, argv+2,
		    TK_CONFIG_ARGV_ONLY);
	}
    } else if ((c == 'd') && (strncmp(argv[1], "dump", length) == 0)) {
        result = GlxwinDump(interp, glxwinPtr, argc, argv);
    } else if ((c == 'r') && (strncmp(argv[1], "rotate", length) == 0)) {
        int vx, vy;
        if (argc != 4) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                             argv[0], " rotate x y\"",
                             (char *) NULL);
            result = TCL_ERROR;
            goto done;
        }
        if (Tcl_GetInt(interp, argv[2], &vx) != TCL_OK) {
            result = TCL_ERROR;
            goto done;
        }
        if (Tcl_GetInt(interp, argv[3], &vy) != TCL_OK) {
            result = TCL_ERROR;
            goto done;
        }
        gtSetRotation(glxwinPtr->handle, vx, vy);
        if (!(glxwinPtr->flags & REDRAW_PENDING)) {
            Tk_DoWhenIdle(DisplayGlxwin, (ClientData) glxwinPtr);
        }
        glxwinPtr->flags |= ROTATE_PENDING|REDRAW_PENDING;
    } else if ((c == 'r') && (strncmp(argv[1], "rotx", length) == 0)) {
        int vx;
        if (argc != 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                             argv[0], " rotx angle\"",
                             (char *) NULL);
            result = TCL_ERROR;
            goto done;
        }
        if (Tcl_GetInt(interp, argv[2], &vx) != TCL_OK) {
            result = TCL_ERROR;
            goto done;
        }
        gtSetRotx(glxwinPtr->handle, vx);
        if (!(glxwinPtr->flags & REDRAW_PENDING)) {
            Tk_DoWhenIdle(DisplayGlxwin, (ClientData) glxwinPtr);
        }
        glxwinPtr->flags |= ROTATE_PENDING|REDRAW_PENDING;
    } else if ((c == 'r') && (strncmp(argv[1], "roty", length) == 0)) {
        int vy;
        if (argc != 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                             argv[0], " roty angle\"",
                             (char *) NULL);
            result = TCL_ERROR;
            goto done;
        }
        if (Tcl_GetInt(interp, argv[2], &vy) != TCL_OK) {
            result = TCL_ERROR;
            goto done;
        }
        gtSetRoty(glxwinPtr->handle, vy);
        if (!(glxwinPtr->flags & REDRAW_PENDING)) {
            Tk_DoWhenIdle(DisplayGlxwin, (ClientData) glxwinPtr);
        }
        glxwinPtr->flags |= ROTATE_PENDING|REDRAW_PENDING;
    } else if ((c == 'u') && (strncmp(argv[1], "update", length) == 0)) {
        if (!(glxwinPtr->flags & REDRAW_PENDING)) {
            Tk_DoWhenIdle(DisplayGlxwin, (ClientData) glxwinPtr);
        }
        glxwinPtr->flags |= ROTATE_PENDING|REDRAW_PENDING;
    } else if ((c == 'o') && (strncmp(argv[1], "origin", length) == 0)) {
        int vx, vy;
        if (argc != 4) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                             argv[0], " origin x y\"",
                             (char *) NULL);
            result = TCL_ERROR;
            goto done;
        }
        if (Tcl_GetInt(interp, argv[2], &vx) != TCL_OK) {
            result = TCL_ERROR;
            goto done;
        }
        if (Tcl_GetInt(interp, argv[3], &vy) != TCL_OK) {
            result = TCL_ERROR;
            goto done;
        }
        gtSetOrigin(glxwinPtr->handle, vx, vy);
    } else if ((c == 'l') && (strncmp(argv[1], "loadfont", length) == 0)) {
        LoadFontGlxwin(glxwinPtr, Tk_WindowId(glxwinPtr->tkwin),
                       glxwinPtr->glxcontext, 120);
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
		"\":  must be configure, position, or size", (char *) NULL);
        result = TCL_ERROR;
    }

done:
    Tk_Release((ClientData) glxwinPtr);
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * DestroyGlxwin --
 *
 *	This procedure is invoked by Tk_EventuallyFree or Tk_Release
 *	to clean up the internal structure of a glxwin at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the glxwin is freed up.
 *
 *----------------------------------------------------------------------
 */

#if XD_TK_VERSION >= 41
static void
DestroyGlxwin(char *clientData)
#else
static void
DestroyGlxwin(ClientData clientData)
#endif
{
    Glxwin *glxwinPtr = (Glxwin *) clientData;

    Tk_FreeOptions(configSpecs, (char *) glxwinPtr, glxwinPtr->display,
                   glxwinPtr->mask);
    /*
     * what about colormap and glxcontext?
     */
    ckfree((char *) glxwinPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * ConfigureGlxwin --
 *
 *	This procedure is called to process an argv/argc list in
 *	conjunction with the Tk option database to configure (or
 *	reconfigure) a glxwin widget.
 *
 * Results:
 *	The return value is a standard Tcl result.  If TCL_ERROR is
 *	returned, then interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information, such as width and height
 *	get set for glxwinPtr;  old resources get freed,
 *	if there were any.
 *
 *----------------------------------------------------------------------
 */

static int
ConfigureGlxwin(Tcl_Interp *interp,  /* Used for error reporting. */
                Glxwin *glxwinPtr,   /* Information about widget. */
                int argc,            /* Number of valid entries in argv. */
                char **argv,         /* Arguments. */
                int flags)           /* Flags to pass to Tk_ConfigureWidget. */
{
    if (Tk_ConfigureWidget(interp, glxwinPtr->tkwin, configSpecs,
                           argc, argv, (char *) glxwinPtr,
                           flags | glxwinPtr->mask ) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Register the desired geometry for the window.  Then arrange for
     * the window to be redisplayed.
     */
    if ((glxwinPtr->width > 0) || (glxwinPtr->height > 0)) {
	Tk_GeometryRequest(glxwinPtr->tkwin, glxwinPtr->width,
			   glxwinPtr->height);
    }
    
    if (Tk_IsMapped(glxwinPtr->tkwin)) {
        if (!(glxwinPtr->flags & REDRAW_PENDING)) {
            Tk_DoWhenIdle(DisplayGlxwin, (ClientData) glxwinPtr);
        }
        glxwinPtr->flags |= REDRAW_PENDING;
    }
    return TCL_OK;
}

static void
DisplayGlxwin(ClientData clientdata)
{
    Glxwin *glxwinPtr = (Glxwin *) clientdata;

    glxwinPtr->flags &= ~REDRAW_PENDING;
    if ((glxwinPtr->tkwin == NULL) || !Tk_IsMapped(glxwinPtr->tkwin)) {
        return;
    }
    
    if(!glXMakeCurrent(glxwinPtr->display, Tk_WindowId(glxwinPtr->tkwin),
                       glxwinPtr->glxcontext)) {
       fprintf(stderr, "glXMakeCurrent failed.\n");
    }
    gtRedraw(glxwinPtr->handle, glxwinPtr->flags);
    glXSwapBuffers(glxwinPtr->display, Tk_WindowId(glxwinPtr->tkwin));
    glxwinPtr->flags = 0;
}

/*
 *--------------------------------------------------------------
 *
 * GlxwinEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher for various
 *	events on glxwins.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

static void
GlxwinEventProc(ClientData clientData,  /* Information about window. */
                XEvent *eventPtr)       /* Information about event. */
{
    Glxwin *glxwinPtr = (Glxwin *) clientData;

    /* printf("Event!\n"); /* */
    
    if ((eventPtr->type == Expose) && (eventPtr->xexpose.count == 0)) {
        if ((glxwinPtr->tkwin != NULL) &&
            !(glxwinPtr->flags & REDRAW_PENDING))
            Tk_DoWhenIdle(DisplayGlxwin, (ClientData) glxwinPtr);
        glxwinPtr->flags |= EXPOSE_PENDING|REDRAW_PENDING;
    } else if (eventPtr->type == ConfigureNotify) {
        gtSetWindowSize(glxwinPtr->handle, Tk_Width(glxwinPtr->tkwin),
                        Tk_Height(glxwinPtr->tkwin));
        if ((glxwinPtr->tkwin != NULL) &&
            !(glxwinPtr->flags & REDRAW_PENDING))
            Tk_DoWhenIdle(DisplayGlxwin, (ClientData) glxwinPtr);
        glxwinPtr->flags |= CONFIGURE_PENDING|REDRAW_PENDING;
    } else if (eventPtr->type == DestroyNotify) {
#if XD_TK_VERSION >= 40
        if (glxwinPtr->tkwin != NULL) {
            glxwinPtr->tkwin = NULL;
            Tcl_DeleteCommand(glxwinPtr->interp,
                              Tcl_GetCommandName(glxwinPtr->interp,
                                                 glxwinPtr->widgetCmd));
        }
#else
        Tcl_DeleteCommand(glxwinPtr->interp, Tk_PathName(glxwinPtr->tkwin));
        glxwinPtr->tkwin = NULL;
#endif
        if (glxwinPtr->flags & REDRAW_PENDING) {
#if XD_TK_VERSION >= 41
            Tcl_CancelIdleCall(DisplayGlxwin, (ClientData) glxwinPtr);
#else
            Tk_CancelIdleCall(DisplayGlxwin, (ClientData) glxwinPtr);
#endif
        }
#if XD_TK_VERSION >= 41
        Tcl_CancelIdleCall(MapGlxwin, (ClientData) glxwinPtr);
#else
        Tk_CancelIdleCall(MapGlxwin, (ClientData) glxwinPtr);
#endif
#if XD_TK_VERSION >= 41
        Tcl_EventuallyFree((ClientData) glxwinPtr, DestroyGlxwin);
#else
        Tk_EventuallyFree((ClientData) glxwinPtr, DestroyGlxwin);
#endif
    }
    
    return;
}

/*
 *----------------------------------------------------------------------
 *
 * GlxwinCmdDeletedProc --
 *
 *      This procedure is invoked when a widget command is deleted.  If
 *      the widget isn't already in the process of being destroyed,
 *      this command destroys it.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The widget is destroyed.
 *
 *----------------------------------------------------------------------
 */

static void
GlxwinCmdDeletedProc(ClientData clientData) /* Pointer to widget record */
{
    Glxwin *glxwinPtr = (Glxwin *) clientData;
    Tk_Window tkwin = glxwinPtr->tkwin;

    /*
     * This procedure could be invoked either because the window was
     * destroyed and the command was then deleted (in which case tkwin
     * is NULL) or because the command was deleted, and then this procedure
     * destroys the widget.
     */

    if (tkwin != NULL) {
        glxwinPtr->tkwin = NULL;
        Tk_DestroyWindow(tkwin);
    }
}


static void
MapGlxwin(ClientData clientData)
{
    Glxwin *glxwinPtr = (Glxwin *) clientData;

    /*
     * Wait for all other background events to be processed before
     * mapping window.  This ensures that the window's correct geometry
     * will have been determined before it is first mapped, so that the
     * window manager doesn't get a false idea of its desired geometry.
     */

#if XD_TK_VERSION >= 41
    Tcl_Preserve((ClientData) glxwinPtr);
#else
    Tk_Preserve((ClientData) glxwinPtr);
#endif
    while (1) {
#if XD_TK_VERSION >= 41
        if (Tcl_DoOneEvent(TK_IDLE_EVENTS) == 0) {
#else
        if (Tk_DoOneEvent(TK_IDLE_EVENTS) == 0) {
#endif
            break;
        }

        /*
         * After each event, make sure that the window still exists
         * and quit if the window has been destroyed.
         */

        if (glxwinPtr->tkwin == NULL) {
#if XD_TK_VERSION >= 41
            Tcl_Release((ClientData) glxwinPtr);
#else
            Tk_Release((ClientData) glxwinPtr);
#endif
            return;
        }
    }
    Tk_MapWindow(glxwinPtr->tkwin);
#if XD_TK_VERSION >= 41
    Tcl_Release((ClientData) glxwinPtr);
#else
    Tk_Release((ClientData) glxwinPtr);
#endif
    
}

static void
LoadFontGlxwin(Glxwin *glxwinPtr,
               GLXDrawable drawable,
               GLXContext glxcontext,
               int size)
{
    Font font;
    if (size <= 120) {
      font = XLoadFont(glxwinPtr->display,
                         "-Adobe-Helvetica-Bold-R-Normal--*-120-*-*-*-*-*-*"); 
    } else {
     font = XLoadFont(glxwinPtr->display,
                         "-Adobe-Helvetica-Bold-R-Normal--*-240-*-*-*-*-*-*"); 
    }
    glXMakeCurrent(glxwinPtr->display, drawable, glxcontext);
    glXUseXFont(font, 0, 256, 1000);
}

/*----------------------------------------------------------------------
 * TkGLx_Init()
 *
 * TkGlx Initialization routine: Create the command that manipulates
 * OpenGL widgets.
 *
 *----------------------------------------------------------------------
 */

int TkGLx_Init(Tcl_Interp *interp, Tk_Window  topLevel)
{
    Tcl_CreateCommand(interp, "GLxwin", GlxwinCmd, topLevel, NULL);

    return TCL_OK;
}

static int
GlxwinDump(Tcl_Interp *interp,
           Glxwin *glxwinPtr,
           int argc,
           char **argv)
{
    Pixmap pixmap;
    GLXPixmap glxpixmap;
    XVisualInfo *vi;
    Display *dpy;
    GLXContext cx;
    XWindowAttributes attr;
    Colormap cmap;
    GLenum err;
    int i, width, height;
    char *file;
    static char *defaultFile = "xdgraph.ppm";
    int attributeList[] = {GLX_RGBA,
                           GLX_DEPTH_SIZE, 8,
                           GLX_RED_SIZE,   3,
                           GLX_GREEN_SIZE, 3,
                           GLX_BLUE_SIZE,  2,
                           None};

    width = height = 500;
    file = NULL;
    for (i=2; i<argc; ++i) {
        if (strEQ(argv[i], "-width", 3)) {
            if (++i < argc) {
                if (Tcl_GetInt(interp, argv[i], &width) != TCL_OK) {
                    return TCL_ERROR;
                }
            }
        } else if (strEQ(NULL, "-height", 3)) {
            if (++i < argc) {
                if (Tcl_GetInt(interp, argv[i], &height) != TCL_OK) {
                    return TCL_ERROR;
                }
            }
        } else if (strEQ(NULL, "-file", 3)) {
            if (++i < argc) {
                file = strDuplOver(file, argv[i]);
            }
        }
    }
    if (!file) {
        file = strDupl(defaultFile);
    }

    dpy = glxwinPtr->display;
    vi = glXChooseVisual(dpy, DefaultScreen(dpy), attributeList);
    if (! vi) {
	Tcl_AppendResult(interp, "No visual for dump", (char *) NULL);
	return TCL_ERROR;
    }
#ifdef DEBUG
    printf("Using visual id=0x%x\n", vi->visualid);
#endif
    printf("Creating a pixmap (depth: %d)\n", vi->depth);
    pixmap = (Pixmap) NULL;
    pixmap = XCreatePixmap(dpy, Tk_WindowId(glxwinPtr->tkwin),
                           width, height, vi->depth);

    if (!pixmap) {
        fprintf(stderr,"Could not create the X Window System pixmap\n");
        return TCL_ERROR;
    }

    /*
     * IMPORTANT:
     *   Use the glXCreateGLXPixmapMESA funtion when using Mesa because
     *   Mesa needs to know the colormap associated with a pixmap in order
     *   to render correctly.  This is because Mesa allows RGB rendering
     *   into any kind of visual, not just TrueColor or DirectColor.
     */
    glxpixmap = (GLXPixmap) NULL;
    XGetWindowAttributes(dpy, Tk_WindowId(glxwinPtr->tkwin), &attr);
    cmap = attr.colormap;

#ifdef GLX_MESA_pixmap_colormap
    /* CLJF the call to glXCreateGLXPixmapMESA seems to cause hard copy */
    /* to fail in recent Linux installations - revert to alternative */
    /* glxpixmap = glXCreateGLXPixmapMESA(dpy, vi, pixmap, cmap); */

    glxpixmap = glXCreateGLXPixmap(dpy, vi, pixmap);
#else
    /* This will work with Mesa too if the visual is TrueColor or DirectColor */
    glxpixmap = glXCreateGLXPixmap(dpy, vi, pixmap);
#endif

    if (glxpixmap) {
        cx = NULL;
        cx = glXCreateContext(dpy, vi, 0, GL_FALSE);
        if (cx) {
            if (glXMakeCurrent(dpy, glxpixmap, cx)) {
                if (width > 1280) {
                    LoadFontGlxwin(glxwinPtr, glxpixmap, cx, 160);
                } else {
                    LoadFontGlxwin(glxwinPtr, glxpixmap, cx, 120);
                }
                gtSetWindowSize(glxwinPtr->handle, width, height);
                printf("Rendering ...\n");
                gtRedraw(glxwinPtr->handle, glxwinPtr->flags|DUMP_FLAGS);
                gtSetWindowSize(glxwinPtr->handle, Tk_Width(glxwinPtr->tkwin),
                                Tk_Height(glxwinPtr->tkwin));
                glFlush();
                printf("Writing to ppm file %s\n", file);
                dumpPixmap(dpy, vi, cmap, pixmap, file);
            } else {
                fprintf(stderr,"Could not attach the current context to the GLX pixmap\n");
            }
            glXDestroyContext(dpy, cx);
        } else {
            fprintf(stderr,"Could not create new context\n");
        }
        XFreePixmap(dpy, pixmap);  
        glXDestroyGLXPixmap(dpy, glxpixmap);
    } else {
        fprintf(stderr,"Could not create the GLX pixmap\n");
        err = glGetError();
        fprintf(stderr,"Reason: %d %s\n", err, gluErrorString(err));
    }
    ckfree(file);
    printf("Done.\n");
    return TCL_OK;
    
}

static void
dumpPixmap(Display *dpy,
           XVisualInfo *vi,
           Colormap cmap,
           Pixmap pixmap,
           char *file)
{
    XImage *ximage;
    Window win;
    XColor *color;
    FILE *f;
    int dummy;
    struct rgbstruct {
        double scale;
        unsigned long mask;
        int shift;
        unsigned short max;
        GLubyte value;
    } rgb[3];
    unsigned int dum;
    unsigned long pixel;
    unsigned int width, height, i, k, c;
    
    switch (vi->class) {
    case TrueColor:
    case DirectColor:
        rgb[0].mask = vi->red_mask;
        rgb[1].mask = vi->green_mask;
        rgb[2].mask = vi->blue_mask;
        for (c=0; c<3; ++c) {
            rgb[c].scale = 1.;
            rgb[c].shift = 0;
            rgb[c].value = 0;
            while ((rgb[c].mask >> rgb[c].shift) % 2 != 1)
                ++rgb[c].shift;
            rgb[c].max = (rgb[c].mask >> rgb[c].shift);
            if (rgb[c].max != 255) rgb[c].scale = 255./((double) rgb[c].max);
        }
#ifdef DEBUG
        printf("rmask,rshift 0x%8.8x, %d\n", rgb[0].mask, rgb[0].shift);
        printf("gmask,gshift 0x%8.8x, %d\n", rgb[1].mask, rgb[1].shift);
        printf("bmask,bshift 0x%8.8x, %d\n", rgb[2].mask, rgb[2].shift);
        printf("scaling: %f %f %f\n",rgb[0].scale, rgb[1].scale, rgb[2].scale);
#endif
                cmap = (Colormap) NULL;
        break;
        
    default:
#ifdef DEBUG
        printf("colormap_size: %d\n", vi->colormap_size);
        printf("bits_per_rgb: %d\n", vi->bits_per_rgb);
#endif
        color = (XColor *) ckalloc(vi->colormap_size * sizeof(XColor));
        for (i=0; i<vi->colormap_size; ++i) color[i].pixel = i;
        XQueryColors(dpy, cmap, color, vi->colormap_size);
        for (i=0; i<vi->colormap_size; ++i) {
            color[i].red /= 257.;
            color[i].green /= 257.;
            color[i].blue /= 257.;
        }
        break;
    }

    XGetGeometry(dpy, pixmap, &win, &dummy, &dummy,
                 &width, &height, &dum, &dum);
    ximage = XGetImage(dpy, pixmap, 0, 0, width, height,
                       AllPlanes, ZPixmap);

    f = fopen(file, "w" );
    if (!f) {
        perror("PPM file");
        return;
    }
    fprintf(f,"P6\n");
    fprintf(f,"# ppm-file created by xdgraph\n");
    fprintf(f,"%i %i\n", width, height);
    fprintf(f,"255\n");

    for (i=0; i<height; ++i) {
        for (k=0; k<width; ++k) {
            pixel = XGetPixel(ximage, k, i);
            if (cmap) {
                rgb[0].value = color[pixel].red;
                rgb[1].value = color[pixel].green;
                rgb[2].value = color[pixel].blue;
            } else {
                for (c=0; c<3; ++c) {
                    rgb[c].value = (GLubyte)
                        ((double) ((pixel & rgb[c].mask) >> rgb[c].shift)
                         * rgb[c].scale);
                }
            }
            fputc(rgb[0].value, f);
            fputc(rgb[1].value, f);
            fputc(rgb[2].value, f);
        }
    }
    fclose(f);
    return;
}
