#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/grdManager.c,v 2.10 1996/07/25 09:14:52 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: grdManager.c,v $
 * Revision 2.10  1996/07/25 09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.1  1996/07/03  14:24:48  richter
 * slicing works again
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:07  richter
 * First semi-public beta release
 *
 *
 */
/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "grdProto.h"
#include "pltProto.h"

void
grdClean(GridP grid,
	 int old)		/* was in use ==> free malloc'ed objects  */
{
    if ( ! grid ) return;

    if ( old ) {
	if (grid->va) ckfree((char *)grid->va);
    }

    grid->nx =   0;
    grid->ny =   0;
    grid->nz =   0;
    grid->xran = 0.;
    grid->yran = 0.;
    grid->zran = 0.;
    grid->xinc = 0.;
    grid->yinc = 0.;
    grid->zinc = 0.;
    grid->ox   = 0.;
    grid->oy   = 0.;
    grid->oz   = 0.;
    grid->vmin = 0.;
    grid->vmax = 0.;
    grid->va =   NULL;
    grid->pl =   NULL;
    return;
}

int
grdSetRange(Tcl_Interp *interp,
	    GridP gPtr)
{
    int i, nv;

    if (gPtr->va) {
	gPtr->vmin=gPtr->vmax=gPtr->va[0];
	nv = 1;
	if (gPtr->nx > 1) nv *= gPtr->nx;
	if (gPtr->ny > 1) nv *= gPtr->ny;
	if (gPtr->nz > 1) nv *= gPtr->nz;
	for (i=1; i<nv; ++i) {
	    if (gPtr->va[i] > gPtr->vmax) {
		gPtr->vmax = gPtr->va[i];
	    } else if (gPtr->va[i] < gPtr->vmin) {
		gPtr->vmin = gPtr->va[i];
	    }
	}
    } else {
	gPtr->vmin=gPtr->vmax = 0.;
    }
    return TCL_OK;
}

int
grdInfo(Tcl_Interp *interp,
	GridP gPtr)
{
    int hist[10];
    int i, nv, slot;
    double range;

    for (i=0; i<10; ++i) {
	hist[i] = 0;
    }

    nv = 1;
    if (gPtr->nx > 1) nv *= gPtr->nx;
    if (gPtr->ny > 1) nv *= gPtr->ny;
    if (gPtr->nz > 1) nv *= gPtr->nz;

    if (gPtr->va) {
	range = (gPtr->vmax - gPtr->vmin);
	for (i=0; i<nv; ++i) {
	    slot = (int) (10. * (gPtr->va[i] - gPtr->vmin) / range);
	    if ( slot == 10 ) slot = 9;
	    hist[slot]++;
	}
    } else {
	gPtr->vmin=gPtr->vmax = 0.;
    }


    sprintf(buf, "\tNo. of values: %10d %10d %10d\n",
	    gPtr->nx, gPtr->ny, gPtr->nz );
    Tcl_AppendResult(interp, buf, NIL);
    sprintf(buf, "\tReal size [A]: %10.5f %10.5f %10.5f\n",
	    gPtr->xran, gPtr->yran, gPtr->zran );
    Tcl_AppendResult(interp, buf, NIL);
    sprintf(buf, "\tIncrement [A]: %10.5f %10.5f %10.5f\n",
	    gPtr->xinc, gPtr->yinc, gPtr->zinc );
    Tcl_AppendResult(interp, buf, NIL);
    sprintf(buf, "\tOrigin:        %10.5f %10.5f %10.5f\n",
	    gPtr->ox, gPtr->oy, gPtr->oz );
    Tcl_AppendResult(interp, buf, NIL);

    sprintf(buf, "\tValues range [%f..%f]\n", gPtr->vmin, gPtr->vmax );
    Tcl_AppendResult(interp, buf, "\tHistogram:  ", NIL);
    for (i=0; i<10; ++i) {
	sprintf(buf, " %d", hist[i] );
	Tcl_AppendResult(interp, buf, NIL);
    }
    Tcl_AppendResult(interp, "\n", NIL);

    return TCL_OK; 
}

double
grdLinInterpol(GridP gPtr,
	       double p[3])
{
    int gx[2], gy[2], gz[2];
    int i, j, k, n;
    double sx[2], sy[2], sz[2];
    double v, px, py, pz;

    px = p[0] + gPtr->xran/2.;
    py = p[1] + gPtr->yran/2.;
    pz = p[2] + gPtr->zran/2.;
    
    gx[0] = floor(px/gPtr->xinc);
    gx[1] = gx[0] + 1;
    if (gx[1] >= gPtr->nx) {
	gx[0] = gx[1] = gPtr->nx - 1;
    } else if (gx[0] < 0) {
        gx[0] = gx[1] = 0;
    }
    /*    printf("  %d %d", gx[0], gx[1]); /* */
    sx[1] = px/gPtr->xinc - (double)gx[0];
    sx[0] = 1. - sx[1];

    gy[0] = floor(py/gPtr->yinc);
    gy[1] = gy[0] + 1;
    if (gy[1] >= gPtr->ny) {
        gy[0] = gy[1] = gPtr->ny - 1;
    } else if (gy[0] < 0) {
        gy[0] = gy[1] = 0;
    }
    /*     printf("  %d %d", gy[0], gy[1]); /* */
    sy[1] = py/gPtr->yinc - (double)gy[0];
    sy[0] = 1. - sy[1];

    gz[0] = floor(pz/gPtr->zinc);
    gz[1] = gz[0] + 1;
    if (gz[1] >= gPtr->nz) {
        gz[0] = gz[1] = gPtr->nz - 1;
    } else if (gz[0] < 0) {
        gz[0] = gz[1] = 0;
    }
    /*    printf("  %d %d ", gz[0], gz[1]); /* */
    sz[1] = pz/gPtr->zinc - (double)gz[0];
    sz[0] = 1. - sz[1];

    v = 0;
    for (i=0; i<2; ++i) {
	for (j=0; j<2; ++j) {
	    for (k=0; k<2; ++k) {
		n = gx[i] + gy[j]*gPtr->nx + gz[k]*gPtr->nx*gPtr->ny;
		v += gPtr->va[n] * sx[i] * sy[j] * sz[k];
	    }
	}
    }

    return v;
}
