#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/isoManager.c,v 2.11 1996/09/06 11:21:34 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: isoManager.c,v $
 * Revision 2.11  1996/09/06 11:21:34  richter
 * new 'tag' field (for the display list interface)
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.1  1996/07/03  11:00:14  richter
 * handle negative iso surface values correctly
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:05  richter
 * First semi-public beta release
 *
 *
 */
/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "isoProto.h"
#include "namProto.h"
#include "pltProto.h"
#include "strProto.h"
#include "setProto.h"
#include "styProto.h"
#include "linProto.h"

static void srfFreeTriangle(TriangleP *t);

IsosurfaceP
isoAlloc(char *name)
{
    IsosurfaceP iPtr;


    iPtr = (IsosurfaceP) ckalloc( sizeof(IsosurfaceT) );
    if (! iPtr) {
	Tcl_AppendResult(interp, "memory allocation error", NIL);
	return NULL;
    }

    isoClean(iPtr, FALSE);
    iPtr->name = strDupl(name);
    if (namAlloc("ISO", name, (void *)iPtr) != TCL_OK) {
        isoFree(&iPtr);
        iPtr = NULL;
    }

    return(iPtr);
}

void
isoClean(IsosurfaceP iPtr,
	 int old)
{
    if (!iPtr) return;

    styClean(&iPtr->style, old);

    if (old) {
	if (iPtr->name)    ckfree(iPtr->name);
	if (iPtr->surface) srfFree(&iPtr->surface);
	if (iPtr->view)    ckfree(iPtr->view);
	if (iPtr->object)  ckfree(iPtr->object);
    }

    iPtr->name =   NULL;
    iPtr->flags =  0;
    iPtr->ns =     0;
    iPtr->surface =  NULL;
    iPtr->view =    NULL;
    iPtr->object = NULL;
    return;
}

void
isoFree(IsosurfaceP *iso)
{
    IsosurfaceP iPtr;

    iPtr = *iso;
    if (!iPtr) return;

    namFree("ISO", iPtr->name);
    styClean(&iPtr->style, TRUE);
    if (iPtr->surface) srfFree(&iPtr->surface);
    if (iPtr->name)    ckfree(iPtr->name);
    if (iPtr->view)    ckfree(iPtr->view);
    if (iPtr->object)  ckfree(iPtr->object);
    ckfree( (char *)iPtr );
    *iso = (IsosurfaceP) NULL;
    return;
}

int
isoListByName(char *match)
{
    return namListByName("ISO", match);
}

IsosurfaceP
isoFindByName(char *name)
{
    return (IsosurfaceP) namFindByName("ISO", name);
}

/*
 *----------------------------------------------------------------------
 * Surface (srf)
 *----------------------------------------------------------------------
 */
SurfaceP
srfAlloc( void )
{
    SurfaceP srf;

    srf = (SurfaceP) ckalloc( sizeof(SurfaceT) );
    srfClean(srf, FALSE);
    return(srf);
}

void
srfClean(SurfaceP srf,
	 int old)
{
    if (!srf) return;

    if (old) {
	if (srf->triangles) srfFreeTriangle(&srf->triangles);
    }
    srf->nt        = 0;
    srf->flags     = 0;
    srf->tag       = 0;
    srf->value     = 0.;
    srf->triangles = NULL;
    srf->tritail   = NULL;
    srf->next      = NULL;
}

static void
srfFreeTriangle(TriangleP *t)
{
    TriangleP n;
    
    while (*t) {
        n = (*t)->next;
        ckfree((char *)*t);
        *t = n;
    }
}
void
srfAddTriangle(void *clientdata,
               double *v1, double *v2, double *v3,
               double *n1, double *n2, double *n3)
/*
 * invert normal for negative surface values
 * re-order vertices to have a consistent CW orientation
 * add triangle to isosurface
 */
{
    TriangleP t;
    SurfaceP sPtr = (SurfaceP)clientdata;
    int i;
    double norm1[3], norm2[3], norm3[3];
    double x[3], y[3], q[3], p[3], dot;
    
    t = (TriangleP) ckalloc( sizeof(TriangleT) );
    t->next = NULL;

    if (sPtr->value < 0.) {
        for (i=0; i<3; ++i) {
            norm1[i] = -n1[i];
            norm2[i] = -n2[i];
            norm3[i] = -n3[i];
        }
    } else {
        for (i=0; i<3; ++i) {
            norm1[i] = n1[i];
            norm2[i] = n2[i];
            norm3[i] = n3[i];
        }
    }
    for (i=0; i<3; ++i) {
        x[i] = v2[i] - v1[i];
        y[i] = v3[i] - v2[i];
        q[i] = (norm1[i] + norm2[i] + norm3[i]) / 3.;
    }
    linCrosuv(x, y, p);
    dot = p[0]*q[0] + p[1]*q[1] + p[2]*q[2];
    if (dot > 0.) {
        t->v1.x = v1[0];
        t->v1.y = v1[1];
        t->v1.z = v1[2];
        t->v2.x = v2[0];
        t->v2.y = v2[1];
        t->v2.z = v2[2];
        t->v3.x = v3[0];
        t->v3.y = v3[1];
        t->v3.z = v3[2];
        t->n1.x = norm1[0];
        t->n1.y = norm1[1];
        t->n1.z = norm1[2];
        t->n2.x = norm2[0];
        t->n2.y = norm2[1];
        t->n2.z = norm2[2];
        t->n3.x = norm3[0];
        t->n3.y = norm3[1];
        t->n3.z = norm3[2];
    } else {
        t->v1.x = v1[0];
        t->v1.y = v1[1];
        t->v1.z = v1[2];
        t->v2.x = v3[0];
        t->v2.y = v3[1];
        t->v2.z = v3[2];
        t->v3.x = v2[0];
        t->v3.y = v2[1];
        t->v3.z = v2[2];
        t->n1.x = norm1[0];
        t->n1.y = norm1[1];
        t->n1.z = norm1[2];
        t->n2.x = norm3[0];
        t->n2.y = norm3[1];
        t->n2.z = norm3[2];
        t->n3.x = norm2[0];
        t->n3.y = norm2[1];
        t->n3.z = norm2[2];
    }        
    if (sPtr->tritail) {
        sPtr->tritail->next = t;
    }
    sPtr->tritail = t;
    if (sPtr->triangles == NULL) sPtr->triangles = t;

}

void
srfFree(SurfaceP *srf)
{
    SurfaceP l;

    l = *srf;
    if (!l) return;
    if (l->triangles) srfFreeTriangle(&l->triangles);
    if (l->next) srfFree(&l->next);
    ckfree( (char *)l );
    *srf = (SurfaceP) NULL;
    return;
}

void
srfClearPlotFlag(SurfaceP srf)
{
    while (srf) {
	srf->flags &= ~ M_SRF_PLOT_DONE;
	srf = srf->next;
    }
    return;
}

void
srfClearCalcFlag(SurfaceP srf)
{
    while (srf) {
	srf->flags &= ~ M_SRF_CALC_DONE;
	srf = srf->next;
    }
    return;
}
