#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/linManager.c,v 2.10 1996/07/25 09:14:52 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: linManager.c,v $
 * Revision 2.10  1996/07/25 09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.1  1996/07/03  10:00:14  richter
 * *** empty log message ***
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:08  richter
 * First semi-public beta release
 *
 *
 */
/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"               /* type definitions and array sizes     */
#include "linProto.h"
#include "pltProto.h"

void
linCoord(double axes[3][3],
	 double vx[3],
	 double vy[3])
{
    double fnorm=0., vz[3];

    /*
     * calculate Z-axis
     */
    linCrosuv(vx, vy, vz);
    /*
     * normalize X- and Z-axis
     */
    linUnit(vx, axes, fnorm, 0);
    linUnit(vz, axes, fnorm, 2);
    /*
     * re-calculate Y-axis
     */
    linCrosuv(axes[2], axes[0], vy);
    /*
     * normalize Y-axis
     */
    linUnit(vy, axes, fnorm, 1);
}
    
void
linUnit(double a[3],
	double basis[3][3],
	double fnorm,
	int jdel)
{
    fnorm = sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2]);

    basis[jdel][0] = a[0] / fnorm;
    basis[jdel][1] = a[1] / fnorm;
    basis[jdel][2] = a[2] / fnorm;

    return;
}

void
linCrosuv(double uu[3],
       double vv[3],
       double uv[3])
{
    uv[0] = uu[1]*vv[2] - uu[2]*vv[1];
    uv[1] = uu[2]*vv[0] - uu[0]*vv[2];
    uv[2] = uu[0]*vv[1] - uu[1]*vv[0];
    return;
}

void
linRotate(double basis[3][3],
	  double angle,
	  int irotax,
	  int iax1,
	  int iax2)
{
/*
 * rotate the vectors iax1 and iax2 <angle> degrees in their plane
 * the positive direction is from iax1 towards iax2
 */
    int i;
    double v1[3], v2[3], rad;
    static const double conv = 3.1415926535897932384626433 / 180.;

    rad = angle*conv;
    for (i=0; i<3; ++i) {
	v1[i] = basis[iax1][i];
	v2[i] = basis[iax2][i];
    }
    for (i=0; i<3; ++i) {
	basis[iax1][i] =   v1[i]*cos(rad)  + v2[i]*sin(rad);
	basis[iax2][i] = (-v1[i]*sin(rad)) + v2[i]*cos(rad);
    }
    return;
}

void
linInvert(double a[3][3], double b[3][3])
{
    double c1, c2, c3, c4, c5, c6, c7, c8, c9;
    double det, rdet;

/*
1 0,0
2 0,1
3 0,2
4 1,0
5 1,1
6 1,2
7 2,0
8 2,1
9 2,2
*/
    c1 = a[1][1]*a[2][2] - a[2][1]*a[1][2];
    c2 = a[2][1]*a[0][2] - a[0][1]*a[2][2];
    c3 = a[0][1]*a[1][2] - a[1][1]*a[0][2];
    c4 = a[0][0]*a[2][2] - a[2][0]*a[0][2];
    c5 = a[1][0]*a[0][2] - a[0][0]*a[1][2];
    c6 = a[0][0]*a[1][1] - a[1][0]*a[0][1];
    c7 = a[1][2]*a[2][0] - a[1][0]*a[2][2];
    c8 = a[1][0]*a[2][1] - a[1][1]*a[2][0];
    c9 = a[0][1]*a[2][0] - a[0][0]*a[2][1];

    det = a[0][0]*c1 + a[1][0]*c2 + a[2][0]*c3;

    if ( det < 0.0001) {
	fprintf(stderr, "singular matrix %f", det);
	exit(1);
    }
    rdet = 1./det;
    b[0][0] = c1 * rdet;
    b[0][1] = c2 * rdet;
    b[0][2] = c3 * rdet;
    b[1][1] = c4 * rdet;
    b[1][2] = c5 * rdet;
    b[2][2] = c6 * rdet;
    b[1][0] = c7 * rdet;
    b[2][0] = c8 * rdet;
    b[2][1] = c9 * rdet;

    return;
}
