#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/molCommand.c,v 2.18 1999/06/04 09:55:04 richter Exp $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif

/*
 * $Log: molCommand.c,v $
 * Revision 2.18  1999/06/04 09:55:04  richter
 * fixed core dump when label where plotted without labels or vice-versa
 *
 * Revision 2.17  1999/06/03 12:40:00  richter
 * we have to scale the zlimit to match the transformed z values
 *
 * Revision 2.16  1997/06/24 10:58:25  richter
 * *** empty log message ***
 *
 * Revision 2.16  1997/06/23 19:16:07  richter
 * added options to suppress atom, bond or label plot
 *
 * Revision 2.15  1997/03/29  18:08:25  richter
 * preset style
 *
 * Revision 2.14  1996/12/19  11:01:35  richter
 * initialize the default styles
 *
 * Revision 2.13  1996/11/03  16:30:06  richter
 * styGetConfig expects a style group mask
 *
 * Revision 2.12  1996/09/06  11:22:04  richter
 * many changes to use the new display list interface
 *
 * Revision 2.11  1996/07/28  11:46:43  richter
 * changed to use display lists
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.3  1996/07/25  06:18:54  richter
 * use Tcl_SetVar2 to set active(molecule)
 * cmdSetView takes an array of possible view types
 * handle 'molecule <molecule>' correctly
 * draw atoms (labels!) after bonds
 *
 * Revision 2.9.1.2  1996/07/03  10:59:22  richter
 * view scaling
 * zlimit is used
 *
 * Revision 2.9.1.1  1996/05/20  11:23:17  richter
 * lots of changes
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:12  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "molProto.h"
#include "cmdProto.h"
#include "vieProto.h"
#include "pltProto.h"
#include "setProto.h"
#include "styProto.h"
#include "strProto.h"
#include "objProto.h"
#include "segProto.h"
#include "dlProto.h"

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static int configureMolecule(Tcl_Interp *interp, MoleculeP mPtr,
                             int argc, char **argv, ViewOptT *vOpt);
static int showcfgMolecule(Tcl_Interp *interp, MoleculeP mPtr);
static int plotMolecule(Tcl_Interp *interp, ViewP view, MoleculeP mPtr);
static int plotAtoms(Tcl_Interp *interp, ViewP view, MoleculeP mPtr);
static int plotBonds(Tcl_Interp *interp, ViewP view, MoleculeP mPtr);
static int clearMolecule(Tcl_Interp *interp, ViewP view, MoleculeP mPtr);

/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */

int
MoleculeCmd(ClientData dummy,		/* Not used. */
            Tcl_Interp *interp,		/* Current interpreter. */
            int argc,			/* Number of arguments. */
            char **argv)			/* Argument strings. */
{
    MoleculeP mPtr;
    DatasetP dPtr;
    ObjectP oPtr;
    ConnP cPtr;
    ViewOptT vOpt;
    ViewP view;
    char *setName;
#define MOL_VIEW_TYPES 4
    ViewE vieType[MOL_VIEW_TYPES] =
    {VIEW_CONTOUR, VIEW_PATH, VIEW_HEIGHT, VIEW_SURFACE};

    if (argc == 1 ||
        (argc == 2 && strEQ(argv[1], "configure", 4))) {
        return showcfgMolecule(interp, NULL);
    }
    
    if (cmdAnalyzeName(argv[1], &setName, &dPtr) != TCL_OK) {
        return TCL_ERROR;
    }
    switch (dPtr->type) {
    case DSET_GRID: {
        oPtr = &dPtr->data.dg->objects;
        cPtr = &dPtr->data.dg->connex;
    }; break;
    case DSET_SLICE: {
        oPtr = &dPtr->data.ds->objects;
        cPtr = &dPtr->data.ds->connex;
    }; break;
    case DSET_PATH: {
        oPtr = &dPtr->data.dp->objects;
        cPtr = &dPtr->data.dp->connex;
    }; break;
    default: {
	Tcl_AppendResult(interp, "molecule is only valid for GRID, SLICE or PATH datasets\n",
                         (char *) NULL);
	return TCL_ERROR;
    }
    }

    mPtr = molAlloc(argv[1]);
    if (!mPtr) {
        Tcl_AppendResult(interp, "Molecule creation failed", (char *) NULL);
        return TCL_ERROR;
    }
    mPtr->set = strDupl(setName);
    mPtr->atoms = oPtr;
    mPtr->bonds  = cPtr;

    styFillLinewidth(interp, &mPtr->def_atom_style, 1.);
    styFillLinewidth(interp, &mPtr->def_bond_style, 1.);
    styFillPattern(interp, &mPtr->def_bond_style, "solid");
    styFillFG(interp, &mPtr->def_label_style, "black");
    styFillAnchor(interp, &mPtr->def_label_style, "c");
    
    vOpt.doscale = 1;
    vOpt.doorig  = 0;
    vOpt.dorot   = 0;
    vOpt.dotrans = 0;
    vOpt.scale = -1.;
    vOpt.width = -1.;
    vOpt.height = -1.;
    if (configureMolecule(interp, mPtr, argc-2, argv+2, &vOpt) != TCL_OK) {
        goto Error;
    }

    vOpt.vrange[0] = dPtr->xsize;
    vOpt.vrange[1] = dPtr->ysize;
    if (cmdSetView(interp, &mPtr->view, MOL_VIEW_TYPES, vieType, setName,
                   argv[1], "molecule", &vOpt) != TCL_OK) {
        Tcl_AppendResult(interp, "view could not be set\n", NULL);
        goto Error;
    }

    view = viewFindByName(mPtr->view, NULL);
    switch (view->type) {
    case VIEW_SURFACE: {
        styFillFG(interp, &mPtr->def_atom_style, "RGB 0.66 0.66 0.66");
        styFillPolygon(interp, &mPtr->def_atom_style, "fill");
        styFillFG(interp, &mPtr->def_bond_style, "RGB 0.66 0.33 0.11");
    }; break;
    case VIEW_HEIGHT: {
        styFillFG(interp, &mPtr->def_atom_style, "RGB 0.66 0.66 0.66");
        styFillPolygon(interp, &mPtr->def_atom_style, "fill");
        styFillFG(interp, &mPtr->def_bond_style, "black");
    }; break;
    default: {
        styFillFG(interp, &mPtr->def_atom_style, "black");
        styFillPattern(interp, &mPtr->def_atom_style, "solid");
        styFillFG(interp, &mPtr->def_bond_style, "black");
    }; break;
    }
        
    Tcl_CreateCommand(interp, mPtr->name, MoleculeObjectCmd, (ClientData) mPtr,
                      (Tcl_CmdDeleteProc *) NULL );
    Tcl_SetVar2(interp, "active", "molecule", mPtr->name, TCL_GLOBAL_ONLY);
    return TCL_OK;

Error:
    molFree(&mPtr);
    return TCL_ERROR;
}

int
MoleculeObjectCmd(ClientData data,              /* Molecule		*/
                  Tcl_Interp *interp,           /* Current interpreter. */
                  int argc,                     /* Number of arguments. */
                  char **argv)                  /* Argument strings.	*/
{
    MoleculeP mPtr;
    ViewP view;
    ViewOptT vOpt;

    mPtr = (MoleculeP) data;
    view = viewFindByName(mPtr->view, NULL);

    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " option ?args?\"", (char *) NULL);
        return TCL_ERROR;
    }

    if (strEQ(argv[1], "configure", 2)) {
        if (argc <= 2) {
            return showcfgMolecule(interp, mPtr);
        } else {
            return configureMolecule(interp, mPtr, argc-2, argv+2, &vOpt);
        }
    } else if (strEQ(NULL, "clear", 2)) {
        clearMolecule(interp, view, mPtr);
    } else if (strEQ(NULL, "info", 1)) {
        return TCL_OK;
        /*	return infoMolecule(interp, mptr); /* */
    } else if (strEQ(NULL, "plot", 2)) {
        clearMolecule(interp, view, mPtr);
	return plotMolecule(interp, view, mPtr);
    } else if (strEQ(NULL, "remove", 3)) {
        clearMolecule(interp, view, mPtr);
        viewUnregisterObject(view, mPtr->name);
        Tcl_DeleteCommand(interp, mPtr->name);
        molFree(&mPtr);
    } else {
	Tcl_AppendResult(interp, "unknown option \"", argv[1],
			 "\" must be configure, clear, info, plot ",
			 " or remove\n", (char *) NULL);
	return TCL_ERROR;
    }

    return TCL_OK;
}

static int
configureMolecule(Tcl_Interp *interp,
                  MoleculeP mPtr,
                  int argc,
                  char **argv,
                  ViewOptT *vOpt)
{
    int i;
    
    viewParseArgs(NULL, &argc, argv, vOpt);
    
    for (i=0; i<argc; ++i) {
        if (strEQ(argv[i], "-zlimit", 3)) {
	    if (++i >= argc || Tcl_GetDouble(interp, argv[i],
					     &mPtr->zlimit) != TCL_OK ) {
                Tcl_AppendResult(interp, "float value for \"-zlimit\" missing or invalid", (char *) NULL);
		return TCL_ERROR;
            }
        } else if (strEQ(NULL, "-atoms", 3)) {
	    if (++i >= argc || (Tcl_GetBoolean(interp, argv[i], 
					     &mPtr->do_atoms) != TCL_OK)) {
                Tcl_AppendResult(interp, "boolean value for \"-atoms\" missing or invalid", (char *) NULL);
		return TCL_ERROR;
            }
        } else if (strEQ(NULL, "-bonds", 3)) {
	    if (++i >= argc || (Tcl_GetBoolean(interp, argv[i], 
					     &mPtr->do_bonds) != TCL_OK)) {
                Tcl_AppendResult(interp, "boolean value for \"-bonds\" missing or invalid", (char *) NULL);
		return TCL_ERROR;
            }
        } else if (strEQ(NULL, "-labels", 3)) {
	    if (++i >= argc || (Tcl_GetBoolean(interp, argv[i], 
					     &mPtr->do_labels) != TCL_OK)) {
                Tcl_AppendResult(interp, "boolean value for \"-labels\" missing or invalid", (char *) NULL);
		return TCL_ERROR;
            }
        } else if (strEQ(NULL, "-view", 3)) {
            i++;
            if (i >= argc) {
                Tcl_AppendResult(interp, "string value for \"-view\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            mPtr->view = strDupl(argv[i]);
        } else {
            Tcl_AppendResult(interp, "unknown option \"",
                             argv[i], "\"", (char *) NULL);
            return TCL_ERROR;
        }
    }
    return TCL_OK;
}


static int
showcfgMolecule(Tcl_Interp *interp,
                MoleculeP mPtr)
{
    Tcl_DString cfg;

    Tcl_DStringInit(&cfg);
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "view");
    Tcl_DStringAppendElement(&cfg, "string::create");
    Tcl_DStringAppendElement(&cfg, "View");
    Tcl_DStringAppendElement(&cfg, "");
    if (mPtr) {
        Tcl_DStringAppendElement(&cfg, mPtr->view);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);

    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "zlimit");
    Tcl_DStringAppendElement(&cfg, "float");
    Tcl_DStringAppendElement(&cfg, "Exclude Atoms by Zlimit");
    Tcl_DStringAppendElement(&cfg, "");
    if (mPtr) {
        sprintf(buf, "%g", mPtr->zlimit);
        Tcl_DStringAppendElement(&cfg, buf);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }       
    Tcl_DStringEndSublist(&cfg);

    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "atoms");
    Tcl_DStringAppendElement(&cfg, "flag");
    Tcl_DStringAppendElement(&cfg, "Draw Atoms");
    Tcl_DStringAppendElement(&cfg, "on");
    if (mPtr) {
	if (mPtr->do_atoms) {
	    Tcl_DStringAppendElement(&cfg, "on");
	} else {
	    Tcl_DStringAppendElement(&cfg, "off");
	}
    } else {
	Tcl_DStringAppendElement(&cfg, "on");
    }
    Tcl_DStringEndSublist(&cfg);

    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "bonds");
    Tcl_DStringAppendElement(&cfg, "flag");
    Tcl_DStringAppendElement(&cfg, "Draw Bonds");
    Tcl_DStringAppendElement(&cfg, "on");
    if (mPtr) {
	if (mPtr->do_bonds) {
	    Tcl_DStringAppendElement(&cfg, "on");
	} else {
	    Tcl_DStringAppendElement(&cfg, "off");
	}
    } else {
	Tcl_DStringAppendElement(&cfg, "on");
    }
    Tcl_DStringEndSublist(&cfg);

    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "labels");
    Tcl_DStringAppendElement(&cfg, "flag");
    Tcl_DStringAppendElement(&cfg, "Draw Labels");
    Tcl_DStringAppendElement(&cfg, "on");
    if (mPtr) {
	if (mPtr->do_labels) {
	    Tcl_DStringAppendElement(&cfg, "on");
	} else {
	    Tcl_DStringAppendElement(&cfg, "off");
	}
    } else {
	Tcl_DStringAppendElement(&cfg, "on");
    }
    Tcl_DStringEndSublist(&cfg);
    if (mPtr) {
        ViewP view = viewFindByName(mPtr->view, NULL);
        /*        styGetConfig(&cfg, &mPtr->style[0], STY_FG_ALL|STY_LINE_ALL|STY_FONT_ALL, 0, 0, NULL); /* */
        if (view) {
            viewGetConfig(&cfg, &view->opt, 0);
        } else {
            viewGetConfig(&cfg, NULL, 0);
        }
    } else {
        /* styGetConfig(&cfg, NULL, STY_FG_ALL|STY_LINE_ALL|STY_FONT_ALL, 0, 0, NULL); /* */
        viewGetConfig(&cfg, NULL, 0);
    }

        
    /*        Tcl_DStringStartSublist(&cfg);
              Tcl_DStringAppendElement(&cfg, "_flags");
              sprintf(buf, "%d", mPtr->flags);
              Tcl_DStringAppendElement(&cfg, buf);
              Tcl_DStringEndSublist(&cfg);
              */

    if (mPtr) {
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_set");
        Tcl_DStringAppendElement(&cfg, mPtr->set);
        Tcl_DStringEndSublist(&cfg);
    }
        
    Tcl_DStringResult(interp, &cfg);
    return TCL_OK;
}

static int
plotMolecule(Tcl_Interp *interp,
             ViewP view,
             MoleculeP mPtr)
{
    if (!view) return TCL_ERROR;

    mPtr->num_atom_tags = 0;
    mPtr->num_label_tags = 0;
    if (mPtr->atoms->na > 0 &&
	(mPtr->do_atoms || mPtr->do_labels)) plotAtoms(interp, view, mPtr);

    mPtr->num_bond_tags = 0;
    if (mPtr->bonds->nc > 0 && mPtr->do_bonds) plotBonds(interp, view, mPtr);

    pltUpdateView(view);
    return TCL_OK;
}

static int
plotAtoms(Tcl_Interp *interp,
	  ViewP view,
	  MoleculeP mPtr)
{
    int n;

    if (mPtr->do_atoms) {
	mPtr->num_atom_tags = mPtr->atoms->na;
	mPtr->atom_tag = (int *) ckalloc(mPtr->num_atom_tags * sizeof(int *));
    }

    if (mPtr->do_labels) {
	mPtr->num_label_tags = mPtr->atoms->na;
	mPtr->label_tag = (int *) ckalloc(mPtr->num_label_tags* sizeof(int *));
    }

    for (n=0; n<mPtr->atoms->na; n++) {
	double radius, zlimit;
	int marker;
	c3dT crd, cxf;
	StyleT style;
	char *key;
            
	crd.x = mPtr->atoms->coord[n].x;
	crd.y = mPtr->atoms->coord[n].y;
	crd.z = mPtr->atoms->coord[n].z;
	pltTransform(view, &crd, &cxf);
        zlimit = pltScale(view, mPtr->zlimit);
        
	if (view->type == VIEW_SURFACE || fabs(cxf.z) < zlimit) {
	    key = getElement(mPtr->atoms->name[n]);
	    styClean(&style, FALSE);
	    styFillFG(interp, &style, styGetElementFG(key));
	    radius = styGetElementRadius(key)/10.;
	    if (mPtr->do_atoms) {
		if (view->type == VIEW_HEIGHT ||
		    view->type == VIEW_SURFACE) {
		    mPtr->atom_tag[n] = dlAlloc(view, DL_SPHERE,
						DL_MAIN_PLANE);
		    dlMkSphere(mPtr->atom_tag[n], &crd, radius, &style);
		} else {
		    mPtr->atom_tag[n] = dlAlloc(view, DL_MARKER,
						DL_MAIN_PLANE);
		    if (fabs(cxf.z) < mPtr->zlimit/10.) {
			marker = 0;
		    } else if ( cxf.z > 0. ) {
			marker = 1;
		    } else {
			marker = 2;
		    }
		    dlMkMarker(mPtr->atom_tag[n], &crd, marker,
			       &mPtr->def_atom_style);
		}
	    }
	    if (mPtr->do_labels) {
		mPtr->label_tag[n] = dlAlloc(view, DL_TEXT, DL_FG_PLANE);
		dlMkText(mPtr->label_tag[n], &crd,
			 mPtr->atoms->name[n], &mPtr->def_label_style);
	    }
	} else {
            if (mPtr->do_atoms) {
                mPtr->atom_tag[n] = 0;
            }
            if (mPtr->do_labels) {
                mPtr->label_tag[n] = 0;
            }
	}
    }
    return TCL_OK;
}

static int
plotBonds(Tcl_Interp *interp,
	  ViewP view,
	  MoleculeP mPtr)
{
    int n;

    mPtr->num_bond_tags = mPtr->bonds->nc;
    mPtr->bond_tag = (int *) ckalloc(mPtr->num_bond_tags * sizeof(int *));
    for (n=0; n < mPtr->bonds->nc; n++) {
	int i, k;
	c3dT c1, c2, cx1, cx2;

	i = mPtr->bonds->obj1[n];
	k = mPtr->bonds->obj2[n];
	c1.x = mPtr->atoms->coord[i].x;
	c1.y = mPtr->atoms->coord[i].y;
	c1.z = mPtr->atoms->coord[i].z;
	c2.x = mPtr->atoms->coord[k].x;
	c2.y = mPtr->atoms->coord[k].y;
	c2.z = mPtr->atoms->coord[k].z;
	pltTransform(view, &c1, &cx1);
	pltTransform(view, &c2, &cx2);

	if (view->type == VIEW_SURFACE ||
	    (fabs(cx1.z) < mPtr->zlimit && fabs(cx2.z) < mPtr->zlimit)) {
	    if (view->type == VIEW_SURFACE) {
		mPtr->bond_tag[n] = dlAlloc(view, DL_CYLINDER,
					    DL_MAIN_PLANE);
		dlMkCylinder(mPtr->bond_tag[n],
			     &c1, &c2, 0.03, &mPtr->def_bond_style);
	    } else {
		SegmentP s;
		mPtr->bond_tag[n] = dlAlloc(view, DL_SEGMENT,
					    DL_MAIN_PLANE);
		s = segAlloc(2, MOVESEG);
		s->type = MOVESEG;
		s->np = 2;
		s->p.point[0].x = c1.x;
		s->p.point[0].y = c1.y;
		s->p.point[0].z = c1.z;
		s->p.point[1].x = c2.x;
		s->p.point[1].y = c2.y;
		s->p.point[1].z = c2.z;
		dlMkSegment(mPtr->bond_tag[n], s, &mPtr->def_bond_style);
	    }
	} else {
	    mPtr->bond_tag[n] = 0;
        }
    }
    return TCL_OK;
}

static int
clearMolecule(Tcl_Interp *interp,
              ViewP view,
              MoleculeP mPtr)
{
    int n;
    
    for (n=0; n<mPtr->num_atom_tags; ++n) {
        dlRemove(view, mPtr->atom_tag[n]);
    }
    if (mPtr->atom_tag)  ckfree((char *)mPtr->atom_tag);
    mPtr->atom_tag = NULL;
    mPtr->num_atom_tags = 0;

    for (n=0; n<mPtr->num_label_tags; ++n) {
        dlRemove(view, mPtr->label_tag[n]);
    }
    if (mPtr->label_tag) ckfree((char *)mPtr->label_tag);
    mPtr->label_tag = NULL;
    mPtr->num_label_tags = 0;

    for (n=0; n<mPtr->num_bond_tags; ++n) {
        dlRemoveData(view, mPtr->bond_tag[n]);
        dlRemove(view, mPtr->bond_tag[n]);
    }
    if (mPtr->bond_tag) ckfree((char *)mPtr->bond_tag);
    mPtr->bond_tag = NULL;
    mPtr->num_bond_tags = 0;

    return TCL_OK;
}
