#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/objManager.c,v 2.12 1999/01/04 14:51:56 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: objManager.c,v $
 * Revision 2.12  1999/01/04 14:51:56  richter
 * apply standard capitalization to element symbols in getElement()
 *
 * Revision 2.11  1996/07/25 12:49:40  richter
 * defstyle moved to MoleculeT, arg 'all' is no longer needed.
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.2  1996/07/03  14:25:15  richter
 * removed obj->s
 *
 * Revision 2.9.1.1  1996/07/03  10:00:43  richter
 * testElement and getElement
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:08  richter
 * First semi-public beta release
 *
 *
 */
/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "objProto.h"
#include "styProto.h"
#include "strProto.h"

void
objClean(ObjectP obj,
	 int old)		/* was in use ==> free malloc'ed objects  */
{
    if (!obj) return;

    if (old) {
	if (obj->coord) ckfree((char *)obj->coord);
	if (obj->name)  ckfree((char *)obj->name);
	if (obj->type)  ckfree((char *)obj->type);
	if (obj->style) ckfree((char *)obj->style);
    }

    obj->na =    0;
    obj->coord = NULL;
    obj->name =  NULL;
    obj->type =  NULL;
    obj->style = NULL;
    return;
}

int
objFindByName(ObjectP oPtr,
	      char *name)
{
    int i;

    if (!oPtr) return -1;
    if (!name) return -1;

    for (i=0; i<oPtr->na; ++i) {
	if (strcmp(name, oPtr->name[i]) == 0) {
	    return i;
	}
    }

    return -1;
}

int
objGetCoord(Tcl_Interp *interp,
	    char *arg, 
	    ObjectP oPtr,
	    double c[3])
{
    int k, n;
    double value;
    int    elemc;
    char **elemv;

    if (Tcl_SplitList(interp, arg, &elemc, &elemv) != TCL_OK) {
	return TCL_ERROR;
    }
    if (elemc > 3) {
	Tcl_AppendResult(interp, "too many values", NIL);
	goto Error;
    }
    if (elemc == 1) {
	if (elemv[0][0] == '#') {
	    n = sscanf(elemv[0], "#%d", &k);
	    if (n == 0) {
		Tcl_AppendResult(interp, "input conversion error", NIL);
		goto Error;
	    }
	} else {	    
	    k = objFindByName(oPtr, elemv[0]);
	}
	if (k < 0 || k > oPtr->na) {
	    Tcl_AppendResult(interp, "atom \"", elemv[0], "\" not found", NIL);
	    goto Error;
	}
	c[0] = oPtr->coord[k].x;
	c[1] = oPtr->coord[k].y;
	c[2] = oPtr->coord[k].z;
    } else if (elemc == 3) {
	for (k=0; k<3; ++k) {
	    if (Tcl_GetDouble(interp, elemv[k], &value) != TCL_OK ) {
		goto Error;
	    }
	    c[k] = value;
	}
    } else {
	Tcl_AppendResult(interp, "wrong number of values", NIL);
	goto Error;
    }

    ckfree((char *) elemv);
    elemv = NULL;
    return TCL_OK;

  Error:
    ckfree((char *) elemv);
    elemv = NULL;
    return TCL_ERROR;

}

/*
int
testElement(char *name,
            char *element)
{
    return (strcasecmp(getElement(name), element) == 0);
}
*/

char *
getElement(char *name)
{
    char *p, *n;
    static char *e = NULL;

    if ( !(name[0] == 'X' && (n = strchr(name, '_')) && (++n)[0] != '\n') ) {
        n = name;
    }
    if (e != NULL) {
        e = strDuplOver(e, n);
    } else {
        e = strDupl(n);
    }
    
    p = strchr(e, '(');
    if (p) {
        *p = '\000';
    } else {
        p = strchr(e, ' ');
        if (p) {
            *p = '\000';
        }
    }
    if (e[0]) e[0] = toupper(e[0]);
    if (e[1]) e[1] = tolower(e[1]);
    return e;

}
