#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/pltViews.c,v 2.12 1999/06/03 12:47:12 richter Exp $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: pltViews.c,v $
 * Revision 2.12  1999/06/03 12:47:12  richter
 * first throw at removable views
 *
 * Revision 2.11  1996/09/06 11:34:17  richter
 * added a call to tkRedraw (for the TK driver) in pltUpdateView
 * added unconditional calls to pltSelectView and pltFlush in pltUpdateView
 * TODO: ccRedraw
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.3  1996/07/25  06:00:27  richter
 * quick DDR (dump driver rotation) hack
 *
 * Revision 2.9.1.2  1996/07/03  10:07:32  richter
 * refined view scaling
 *
 * Revision 2.9.1.1  1996/05/19  16:05:24  richter
 * cleaned up using new view options
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:09  richter
 * First semi-public beta release
 *
 *
 */

/*
 *======================================================================
 *
 * Module: pltViewManager
 *
 *======================================================================
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration			*/
#include "xdgraph.h"		/* type definitions and array sizes	*/
#include "drvProto.h"
#include "vieProto.h"
#include "pltProto.h"
#include "strProto.h"
#include <stdarg.h>

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */

/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */
extern TheDriver driver;

int
pltCreateView(Tcl_Interp *interp,
              ViewP view)
{
    char w[100], h[100];
    int code;
    char *name;

    name = view->name;
    switch (driver.type) {
#ifdef CC_OUTPUT
    case CC: {
    }; break;
#endif
#ifdef TK_OUTPUT
    case TK: {
	tkAllocDriverData( (DrvtkP *)&view->drdat );
	tkSetYoff( (DrvtkP) view->drdat, 0.);
	sprintf(w, " %f", view->opt.width);
	sprintf(h, " %f", view->opt.height);
	code = Tcl_VarEval(interp, "CreateView ", name, w, h, NIL);
	if (code == TCL_ERROR) {
	    Tcl_AddErrorInfo(interp, "(pltCreateView)\n");
	    viewFree(view);
	    return TCL_ERROR;
	}
    }; break;
#endif
#ifdef GT_OUTPUT
    case GT: {
	gtAllocDriverData((DrvgtP *)&view->drdat);
	sprintf( w, " %f", view->opt.width);
	sprintf( h, " %f", view->opt.height);
	code = Tcl_VarEval(interp, "GTCreateView ", name, w, h, NIL);
	if (code == TCL_ERROR) {
	    Tcl_AddErrorInfo(interp, "(pltCreateView)\n");
	    viewFree(view);
	    return TCL_ERROR;
	}
    }; break;
#endif
    }
    return TCL_OK;
    
}

ViewP
pltSelectView(char *name)
{
    ViewP view;
    int code;

    if (!name) return NULL;
    view = viewFindByName(name, NULL);

    if (view) {
        switch (driver.type) {
#ifdef CC_OUTPUT
        case CC: {
        }; break;
#endif
#ifdef TK_OUTPUT
        case TK: {
            code = Tcl_VarEval(interp, "SetActiveView ", name, NULL);
            if (code == TCL_ERROR) {
                Tcl_AddErrorInfo(interp, "(pltSelectView)\n");
                return NULL;
            }
        }; break;
#endif
#ifdef GT_OUTPUT
        case GT: {
        }; break;
#endif
        }
    }

    return view;
}

int
pltUpdateView(ViewP view)
{
    pltSelectView(view->name);
    
    switch (driver.type) {
#ifdef CC_OUTPUT
    case CC:
        break;
#endif
#ifdef TK_OUTPUT
    case TK:
        tkRedraw(view, view->dl_head);
        break;
#endif
#ifdef GT_OUTPUT
    case GT: {
        int code = Tcl_VarEval(interp, "UpdateView ", view->name, NULL);
        if (code != TCL_OK) return code;
    }; break;
#endif
    }
    return pltFlush(view);
    
}

int
pltRemoveView(ViewP view)
{
    pltSelectView("");
    switch (driver.type) {
#ifdef CC_OUTPUT
      case CC:
	  return TCL_OK;
#endif
#ifdef TK_OUTPUT
      case TK:
	  return Tcl_VarEval(interp, "RemoveView ", view->name, NULL);
#endif
#ifdef GT_OUTPUT
      case GT:
	  return Tcl_VarEval(interp, "RemoveView ", view->name, NULL);
#endif
    }
    return TCL_OK;
    
}

#if 0
int
pltClearView(char *name)
{
    ViewP  view;
    
    view = viewFindByName(name, NULL);
    
    switch (driver.type) {
#ifdef CC_OUTPUT
      case CC: {
	  plot(0., 0., -999);
	  return TCL_OK;
      }; break;
#endif
#ifdef TK_OUTPUT
      case TK: {
          return Tcl_VarEval(interp, "ClearView ", name, NIL);
      }; break;
#endif
#ifdef GT_OUTPUT
      case GT: {
	  return gtClearView((DrvgtP)view->drdat);
      }; break;
#endif
    }
    return TCL_OK;
    
}

int
pltRescaleView(ViewP view,
               double scale)
{
    char w[100], h[100];
    int code;

    if (scale <= 0.) {
	scale = 18./view->opt.width;
	if (view->opt.height*scale > 26.) {
	    scale = 26./view->opt.height;
	}
    }


    switch (driver.type) {
#ifdef CC_OUTPUT
      case CC: {
      }; break;
#endif
#ifdef TK_OUTPUT
    case TK: {
	sprintf( w, " %f", view->opt.width*scale );
	sprintf( h, " %f", view->opt.height*scale );
	code = Tcl_VarEval(interp, "RescaleView ", view->name, w, h, NIL);
	if (code == TCL_ERROR) {
	    Tcl_AddErrorInfo(interp, "(pltRescaleView)\n");
	    return TCL_ERROR;
	}
	view->opt.scale = scale;
    }; break;
#endif
#ifdef GT_OUTPUT
      case GT: {
      }; break;
#endif
    }

    return TCL_OK;    
}
#endif

int
pltDDR(ViewP view)
{
#ifdef GT_OUTPUT
    if (driver.type == GT) {
        int i,k;
        DrvgtT *ddp = (DrvgtT *) view->drdat;
        for (i=0; i<4; ++i) {
            for (k=0; k<4; ++k) {
                printf("%f ", ddp->rotmat[i+k*4]);
            }
            printf("\n");
        }
    }
#endif
    return TCL_OK;
}
