#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/pthCommand.c,v 2.14 1997/03/29 18:10:26 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 *======================================================================
 *
 * Module: pthCommand
 *
 * $Log: pthCommand.c,v $
 * Revision 2.14  1997/03/29 18:10:26  richter
 * preset style
 *
 * Revision 2.13  1996/11/03  16:28:38  richter
 * styGetConfig expects a style group mask
 *
 * Revision 2.12  1996/09/11  11:24:43  richter
 * :path configure did not work correctly
 *
 * Revision 2.11  1996/09/06  11:35:05  richter
 * many changes to use the new display list interface
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.4  1996/07/25  06:05:58  richter
 * use Tcl_SetVar2 to set active(path)
 * handle 'path <path>' correctly
 * cmdSetView takes an array of possible view types
 * some fixes concerning transformations (can't remember)
 *
 * Revision 2.9.1.3  1996/07/03  10:14:42  richter
 * refined view scaling
 * configure output
 *
 * Revision 2.9.1.2  1996/05/20  11:18:48  richter
 * lots of changes
 * transform view not path
 *
 * Revision 2.9.1.1  1996/04/29  12:09:19  richter
 * last ci before translate/rotate etc. removal
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:09  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "setProto.h"
#include "cmdProto.h"
#include "segProto.h"
#include "strProto.h"
#include "vieProto.h"
#include "pltProto.h"
#include "linProto.h"
#include "pthProto.h"
#include "objProto.h"
#include "styProto.h"
#include "vieProto.h"
#include "dlProto.h"

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static int configurePath(Tcl_Interp *interp, PathP pPtr,
                         int argc, char **argv, ViewOptT *vOpt);
static int showcfgPath(Tcl_Interp *interp, PathP pPtr);
static int infoPath(Tcl_Interp *interp, PathP pPtr, int argc, char **argv);
static int plotPath(Tcl_Interp *interp, ViewP view, PathP pPtr);
static int clearPath(Tcl_Interp *interp, ViewP view, PathP pPtr);

/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */

/*
 *----------------------------------------------------------------------
 * functions with external and internal linkage
 *----------------------------------------------------------------------
 */

int
PathCmd(ClientData dummy,		/* Not used. */
	   Tcl_Interp *interp,		/* Current interpreter. */
	   int argc,			/* Number of arguments. */
	   char **argv)			/* Argument strings. */
{
    PathP pPtr;
    DatasetP dPtr;
    ViewOptT vOpt;
    char *setName;
    int code;
    ViewE vieType = VIEW_PATH;

    if (argc == 1 ||
        (argc == 2 && strEQ(argv[1], "configure", 4))) {
        return showcfgPath(interp, NULL);
    }

    if (cmdAnalyzeName(argv[1], &setName, &dPtr) != TCL_OK) {
        return TCL_ERROR;
    }
    if (dPtr->type != DSET_PATH) {
	Tcl_AppendResult(interp, "path is only valid for PATH datasets\n",
                         (char *) NULL);
	return TCL_ERROR;
    }

    pPtr = pthAlloc(argv[1]);
    if ( ! pPtr ) {
	Tcl_AppendResult(interp, "couldn't create path", (char *) NULL);
	return TCL_ERROR;
    }
    pPtr->objects = &dPtr->data.dp->objects;
    pPtr->path    = dPtr->data.dp->path;
    pPtr->set     = strDupl(setName);
    styFillFG(interp, &pPtr->style, "red");
    styFillLinewidth(interp, &pPtr->style, 0.2);
    styFillPattern(interp, &pPtr->style, "solid");

    vOpt.doscale = 1;
    vOpt.dorot   = 0;
    vOpt.dotrans = 0;
    vOpt.scale = -1.;
    vOpt.width = -1.;
    vOpt.height = -1.;
    vOpt.doorig = 1;
    vOpt.orig[0] = (dPtr->data.dp->xmax + dPtr->data.dp->xmin) / 2.;
    vOpt.orig[1] = (dPtr->data.dp->ymax + dPtr->data.dp->ymin) / 2.;
    vOpt.orig[2] = (dPtr->data.dp->zmax + dPtr->data.dp->zmin) / 2.;
    code = configurePath(interp, pPtr, argc-2, argv+2, &vOpt);
    if ( code != TCL_OK ) {
	pthFree(&pPtr);
	return TCL_ERROR;
    }

    vOpt.vrange[0] = dPtr->xsize;
    vOpt.vrange[1] = dPtr->ysize;
    if (cmdSetView(interp, &pPtr->view, 1, &vieType, setName,
                   argv[1], "path", &vOpt) != TCL_OK) {
        Tcl_AppendResult(interp, "view could not be set\n", NULL);
        goto Error;
    }

    Tcl_CreateCommand(interp, argv[1], PathObjectCmd, (ClientData) pPtr,
                      (Tcl_CmdDeleteProc *) NULL );
    Tcl_SetVar2(interp, "active", "path", argv[1], TCL_GLOBAL_ONLY);
    return TCL_OK;

  Error:
    pthFree(&pPtr);
    return TCL_ERROR;
}


int
PathObjectCmd(ClientData data,               /* dataset		*/
	      Tcl_Interp *interp,            /* Current interpreter. */
	      int argc,                      /* Number of arguments. */
	      char **argv)                   /* Argument strings.	*/
{
    PathP pPtr;
    ViewP view;
    ViewOptT vOpt;
    int i, k;
    
    pPtr = (PathP) data;
    view = viewFindByName(pPtr->view, NULL);
    
    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " option ?args?\"", (char *) NULL);
        return TCL_ERROR;
    }

    if (strEQ(argv[1],"configure",2)) {
        if (argc <=2 ) {
            return showcfgPath(interp, pPtr);
        } else {
            vOpt.doscale = 0;
            vOpt.doorig = 0;
            vOpt.dorot = 0;
            vOpt.dotrans = 0;
            if (configurePath(interp, pPtr, argc-2, argv+2, &vOpt)
                == TCL_OK) {
                if (vOpt.doscale) {
                    view->opt.doscale = vOpt.doscale;
                    view->opt.scale = vOpt.scale;
                }
                if (vOpt.doorig) {
                    view->opt.doorig = vOpt.doorig;
                    for (i=0; i<3; ++i) {
                        view->opt.orig[i] = vOpt.orig[i];
                    }
                }
                if (vOpt.dorot) {
                    view->opt.dorot = vOpt.dorot;
                    for (i=0; i<3; ++i) {
                        for (k=0; k<3; ++k) {
                            view->opt.rotm[i][k] = vOpt.rotm[i][k];
                        }
                    }
                }
                if (vOpt.dotrans) {
                    view->opt.dotrans = vOpt.dotrans;
                    for (i=0; i<3; ++i) {
                        view->opt.trans[i] = vOpt.trans[i];
                    }
                }
                return TCL_OK;
            }
        }
        return TCL_ERROR;
    } else if ( strEQ(NULL,"clear",2)) {
        clearPath(interp, view, pPtr);
    } else if ( strEQ(NULL,"info",1)) {
	return infoPath(interp, pPtr, argc-2, argv+2 );
    } else if ( strEQ(NULL,"plot",1)) {
        clearPath(interp, view, pPtr);
	return plotPath(interp, view, pPtr);
    } else if ( strEQ(NULL,"remove",2)) {
	Tcl_DeleteCommand(interp, pPtr->name);
	pthFree(&pPtr);
    } else {
	Tcl_AppendResult(interp, "unknown or ambigous option \"", argv[1],
			 "\" must be configure, clear, info, plot,",
			 "rotate or remove", (char *) NULL);
	return TCL_ERROR;
    }

    return TCL_OK;
}

static int
configurePath(Tcl_Interp *interp,
	      PathP pPtr,
	      int argc,
	      char **argv,
              ViewOptT *vOpt)
{
    int i;
    int has2styles;

    viewParseArgs(pPtr->objects, &argc, argv, vOpt);

    styParseArgs(&argc, argv, &pPtr->style, NULL, &has2styles);
    
    for (i=0; i<argc; ++i) {
        if (strEQ(argv[i],"-view",3)) {
            i++;
            if (i >= argc) {
                Tcl_AppendResult(interp, "value for \"-view\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            pPtr->view = strDupl(argv[i]);
	} else {
	    Tcl_AppendResult(interp, "unknown option \"", argv[i], "\"", NIL);
	    return TCL_ERROR;
	}
    }

    return TCL_OK;

}

static int
showcfgPath(Tcl_Interp *interp,
            PathP pPtr)
{
    Tcl_DString cfg;

    Tcl_DStringInit(&cfg);
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "view");
    Tcl_DStringAppendElement(&cfg, "string::create");
    Tcl_DStringAppendElement(&cfg, "View");
    Tcl_DStringAppendElement(&cfg, "");
    if (pPtr) {
        Tcl_DStringAppendElement(&cfg, pPtr->view);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);

    if (pPtr) {
        ViewP view = viewFindByName(pPtr->view, NULL);
        styGetConfig(&cfg, &pPtr->style, STY_FG_ALL|STY_LINE_ALL, 0, 0, NULL);
        if (view) {
            viewGetConfig(&cfg, &view->opt, 1);
        } else {
            viewGetConfig(&cfg, NULL, 1);
        }
    } else {
        styGetConfig(&cfg, NULL, STY_FG_ALL|STY_LINE_ALL, 0, 0, NULL);
        viewGetConfig(&cfg, NULL, 1);
    }

    if (pPtr) {
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_set");
        Tcl_DStringAppendElement(&cfg, pPtr->set);
        Tcl_DStringEndSublist(&cfg);
    }
        
    Tcl_DStringResult(interp, &cfg);
    return TCL_OK;
}

static int
infoPath(Tcl_Interp *interp,
	 PathP pPtr,
         int argc,
         char **argv)
{
    Tcl_DString cmd;

    Tcl_DStringInit(&cmd);
    Tcl_DStringAppendElement(&cmd, "pthinfo");
    Tcl_DStringAppendElement(&cmd, pPtr->name);
    return Tcl_Eval(interp, Tcl_DStringValue(&cmd));
}

static int
plotPath(Tcl_Interp *interp,
         ViewP view,
	 PathP pPtr)
{
    PathdataP data;

    if (!view) return TCL_ERROR;

    for (data=pPtr->path; data; data=data->next) {
        if (data->tag == 0) {
            data->tag = dlAlloc(view, DL_SEGMENT, DL_MAIN_PLANE);
            dlMkSegment(data->tag, data->seg, &pPtr->style);
        }
    }
    pltUpdateView(view);
    return TCL_OK;
}

static int
clearPath(Tcl_Interp *interp,
          ViewP view,
          PathP pPtr)
{
    PathdataP data;

    if (!view) return TCL_ERROR;

    for (data=pPtr->path; data; data=data->next) {
        if (data->tag) {
            dlRemove(view, data->tag);
            data->tag = 0;
        }
    }
    return TCL_OK;
}
