#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/relEngine.c,v 2.10 1996/07/25 09:14:52 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 *======================================================================
 *
 * Module: relEngine
 *
 * $Log: relEngine.c,v $
 * Revision 2.10  1996/07/25 09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.2  1996/07/03  10:03:30  richter
 * some de-linting
 *
 * Revision 2.9.1.1  1996/05/19  15:59:46  richter
 * enhanced segments
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:11  richter
 * First semi-public beta release
 *
 *
 */
/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "relProto.h"
#include "segProto.h"

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static double fun(double x, double y);
static void iplot(double xo, double yo, double xn, double yn);
static void plt3dm(double xlow, double xhigh, double ylow, double yhigh,
		   double a, double b, double c, int n, int m, int idi);

/*
 *----------------------------------------------------------------------
 * variables with internal linkage
 *----------------------------------------------------------------------
 */
static struct {
    double hcut, lcut, vertac, scalex, scaley, xoff, yoff;
} param;

static SegmentP curseg; /* pointer to the currently active segment      */

static GridT   *grid;   /* pointer to the raw data              */

/*
 *----------------------------------------------------------------------
 * functions with external and internal linkage
 *----------------------------------------------------------------------
 */

void
relEngine(GridT *g,
	  ReliefP r )
{

    grid = g;

    if (r->seg) segFree(&r->seg);
    r->seg = curseg = segAlloc(0, 0);

    param.scalex = (grid->xran/1024.);
    param.scaley = (grid->yran/ 780.);
    param.xoff   = grid->xran/2.;
    param.yoff   = grid->yran/2.;
    param.hcut   = r->hcut;
    param.lcut   = r->lcut;
    param.vertac = r->vertac;
    plt3dm(- (double) grid->nx/2, (double) grid->nx/2,
	   - (double) grid->ny/2, (double) grid->ny/2,
	   r->eye.x, r->eye.y, r->eye.z, r->nx, r->ny, r->mode );

/*    segDump(top); */
    return;
}
     
static
double fun(double x,
	   double y)
{
    static int ix, iy;
    static double z;

    ix = (int) (x + (double) (grid->nx/2)) - 1;
    iy = (int) (y + (double) (grid->ny/2)) - 1;
    if (ix < 0) {
	ix = 0;
    } else if ( ix >= grid->nx ) {
	ix = grid->nx - 1;
    }
    if (iy < 0) {
	iy = 0;
    } else if ( iy >= grid->ny ) {
	iy = grid->ny - 1;
    }
    z = grid->va[ix + iy * grid->nx];
    if (z > param.hcut) {
	z = param.hcut;
    } else if (z < param.lcut) {
	z = param.lcut;
    }
    return z * param.vertac;

}

static void
iplot(double xo,
      double yo,
      double xn, 
      double yn)
{
    static double x0 = 0.;
    static double y0 = 0.;
    static double p0 = 100.;
    static double dp = 0.;

    double x1, y1, x2, y2, phi;

    x1 = xo * param.scalex - param.xoff;
    y1 = yo * param.scaley - param.yoff;
    x2 = xn * param.scalex - param.xoff;
    y2 = yn * param.scaley - param.yoff;

    if ( fabs(x0-x1) > 1.e-6  || fabs(y0-y1) > 1.e-6 ) {
	segFinishSegment(&curseg);
	segAddSegment(&curseg, x1, y1, 0.);
	p0 = atan2((y2-y1), (x2-x1));
	dp = 0.;
    } else {
	phi = atan2((y2-y1), (x2-x1));
	dp += p0-phi;
        if (fabs(dp) < 1.e-3) {
	    segReplaceSegment(&curseg, x2, y2, 0.);
	    x0 = x2;
	    y0 = y2;
	    return;
        }
        p0  = phi;
	dp = 0.;
    }

    segAddSegment(&curseg, x2, y2, 0.);

    x0 = x2;
    y0 = y2;

    return;
}

static void
plt3dm(double xlow,
       double xhigh,
       double ylow,
       double yhigh,
       double a, double b, double c, 
       int n,
       int m,
       int idi)
{
    /* Initialized data */

    static double great = 1e20;

    /* System generated locals */
    double d_1, d_2, d_3;

    /* Local variables */
    static int indf, idir, indn, indo, imax, jmax;
#define XRES 1024
    int mask[XRES+1][2];

    static double rmin, rmax;
    static int ixoi, iplt, ixni;
    static double rzzr;
    static int iplt1;
    static double d, e;
    static int i, j, k;
    static double x[2], y[2], z[2],
	    highx, highy;
    static int ifplt, indxn, indxo;
    static double a1, a2, b1, c1, c2;

    static double *icordx, *icordy; /* was [2][403] */

    static int indyo, indyn, iintp;
    static double rintp, rlowx, rlowy;
    static int ihplt1;
    static double dd;
    static int ih, jj, il;
    static double xa, xb, ya, yb, yc, za, zb, hx, hy, xi, yi, xm, ym, 
	    xh, yh, ti, tj, zi;
    static int it, np1, mp2, mp3;
    static double rrmaxx, rrmaxy, rrmaxz, ddd;
    static int ifl;

    static double scx, scy;
    static int ixo, iyo, ixn, iyn;
    static double fxy;

/* THE ROUTINE PLT3DM MAKES A TWO DIMENSIONAL PICTURE OF A */
/* FUNCTION OF TWO ARGUMENTS ,THE HIDDEN PARTS OF THE SURFACE */
/* NOT SHOWN . */

/* AUTHOR &    WAMBECQ A. */
/*             APLIED MATHEMATICS AND PROGRAMMING DIVISION */
/*             CELESTIJNENLAAN 200B */
/*             B-3030  HEVERLEE */
/*             BELGIUM */

/* THE ROUTINE FIRST FINDS THE POINT NEAREST TO THE VIEWER]S */
/* EYE ,THEN DETERMINES THE CENTER OF THE PICTURE AND PROJECTS */
/* THIS TO THE POINT (512,390) OF THE WORKING SPACE OF THE */
/* ROUTINE ,BEING  1024 INT UNITS IN THE X - DIRECTION , */
/* AND  780 INT UNITS IN THE  Y - DIRECTION . */
/* THEN LINES ON THE SURFACE ARE DISPLAYED. */
/* LINES IN THE FOREGROUND ARE ALWAYS PLOTTED FIRST . */
/* IF A LINE LIES OUTSIDE THE VISIBILITY AREA ,THE LINE IS */
/* NOT DRAWN .ON THE EDGE ,LINEAR INTERPOLATION IS USED . */
/* THE EDGES OF THE VISIBILITY AREA ARE STORED IN THE ARRAY */
/* MASK ,AND ARE UPDATED EACH TIME A LINE WAS WRITTEN . */
/* CALLING SEQUENCE & */

/*     CALL PLT3DM (FUN,XLOW,XHIGH,YLOW,YHIGH,A,B,C,N,M,IDI) */

/* ARGUMENTS & */
/* FUN       & FUNCTION OF TWO VARIABLES . */
/*             FUN = FUN(X,Y) */
/*             (MUST BE DECLARED 'EXTERNAL ' IN THE CALLING */
/*              PROGRAM) */
/* XLOW,XHIGH& BOUNDS OF THE  X - INTERVAL */
/* YLOW,YHIGH& BOUNDS OF THE  Y - INTERVAL */
/* A,B,C     & VIEWERS EYE POSITION IN THREE-SPACE */
/* N         & NUMBER OF LINES ON THE SURFACE */
/* M         & NUMBER OF POINTS PER LINE */
/*             (FOR N AND M ,A VALUE OF  60 - 100  WILL BE */
/*              SUFFICIENT . M MUST BE SMALLER THAN 198) */
/* IDI       & = 1 & LINES ARE DRAWN PARALLEL TO THE */
/*                   Y - DIRECTION . */
/*             = 2 & LINES ARE DRAWN PARALLEL TO THE */
/*                   X - DIRECTION . */
/*           = 3 & LINES ARE DRAWN IN BOTH DIRECTIONS */


/*        1) CALLING PROGRAM */
/*        2) FUNCTION FUN(X,Y) */
/*        3) SUBROUTINE PLT3DM */
/*        4) SUBROUTINE IPLOT (IXO,IYO,IXN,IYN) */
/*           THIS ROUTINE LINKS THE (1024 X 780) WORKING SPACE */
/*           OF PLT3DM TO THE PHISICAL SPACE OF THE DEVICE(S)) */
/*           USED . (PLOTTER AND/OR DISPLAY AND/OR LINEPRINTER) */
/*           IPLOT DRAWS A LINE IN THE  1024 X 780 WORKING */
/*           SPACE ,FROM (IXO,IYO) TO (IXN,IYN) */

/*   REMARK & A SCALING IN THE CALLING PROGRAM MUST BE DONE */
/*            TO LINK SOFTWARE TO HARDWARE SAPCE . */
/*            IN CASE OF A PLOTTER ,THE PEN SHOULD BE */
/*            POSITIONED IN THE 'UP' STATE AT THE BOTTEM LEFT */
/*            CORNER OF THE PAPER . */


/*   NOTE: WHEN CHANGING THE DIMENSION OF GRID, ICORD MUST ALSO BE */
/*   CHANGED.  IF GRID(N,N) THEN ICORD MUST HAVE DIMENSIONS OF */
/*   ICORD(N,N+3) */


/* GREAT IS A GREAT (MACHINE DEPENDENT) NUMBER . */
/* STATEMENT FUNCTIONS THAT COMPUTE THE COORDINATES IN TWO */
/* DIMENSIONS .(THIS IS A PROJECTION AFTER TWO ROTATIONS , */
/* SUCH THAT VERTICAL LINES IN THREE-SPACE ARE PROJECTED */
/* AS VERTICAL LINES IN TWO-SPACE ) . */
/* INITIALISATION */


    int maxco;

    maxco = grid->nx > grid->ny ? grid->nx : grid->ny;
    icordx = (double *) malloc( (maxco+3) * sizeof(double) );
    icordy = (double *) malloc( (maxco+3) * sizeof(double) );
    dd = a * a + b * b;
    d = 1. / sqrt(dd);
    e = 1. / sqrt(dd + c * c);
    xa = -(b) * d;
    xb = a * d;
    ya = -(a) * c * d * e;
    yb = -(b) * c * d * e;
    yc = dd * d * e;
    za = a * e;
    zb = b * e;
    ifl = 0;
    ifplt = 1;
    np1 = n + 1;
    mp2 = m + 2;
    mp3 = m + 3;
    il = idi;
    ih = il;
    if (idi == 3) {
	il = 1;
    }
    if (idi == 3) {
	ih = 2;
    }
/* SEARCH THE GREATEST FUNCTION VALUE */
    rmin = great;
    rmax = -rmin;
    hx = (xhigh - xlow) / 19.;
    hy = (yhigh - ylow) / 19.;
    for (i = 1; i <= 20; ++i) {
	xi = xlow + (double) (i - 1) * hx;
	for (j = 1; j <= 20; ++j) {
	    yi = ylow + (double) (j - 1) * hy;
	    fxy = fun(xi, yi);
	    if (fxy < rmin) {
		rmin = fxy;
	    }
	    if (fxy > rmax) {
		rmax = fxy;
	    }
	}
    }

/* DETERMINE THE CENTER OF THE PICTURE */
    x[0] = xlow;
    x[1] = xhigh;
    y[0] = ylow;
    y[1] = yhigh;
    z[0] = rmin - fabs(rmin) * .03;
    z[1] = rmax + fabs(rmax) * .03;
    d_1 = (x[0] + x[1]) / 2;
    d_2 = (y[0] + y[1]) / 2;
    d_3 = (z[0] + z[1]) / 2;
    xm = xa * d_1 + xb * d_2 + d_3 * 0.;
    d_1 = (x[0] + x[1]) / 2;
    d_2 = (y[0] + y[1]) / 2;
    d_3 = (z[0] + z[1]) / 2;
    ym = ya * d_1 + yb * d_2 + yc * d_3;
/* DETERMINE THE POINT NEAREST TO THE VIEWERS EYE . */
    rrmaxx = -great;
    rrmaxy = -great;
    rrmaxz = -great;
    for (i = 1; i <= 2; ++i) {
	for (j = 1; j <= 2; ++j) {
	    rzzr = za * x[i - 1] + zb * y[j - 1];
	    if (rzzr > rrmaxz) {
		imax = i;
		jmax = j;
		rrmaxz = rzzr;
	    }
	    for (k = 1; k <= 2; ++k) {
		d_3 = fabs(xa * x[i - 1] + xb * y[j - 1] - xm);
		rrmaxx = rrmaxx > d_3 ? rrmaxx : d_3;
		d_3 = fabs(ya * x[i - 1] + yb * y[j - 1] + yc * z[k - 1] - ym);
		rrmaxy = rrmaxy > d_3 ? rrmaxy : d_3;
	    }
	}
    }

    highx = x[imax - 1];
    highy = y[jmax - 1];
    imax = 3 - imax;
    jmax = 3 - jmax;
    rlowx = x[imax - 1];
    rlowy = y[jmax - 1];
    for (idir = il; idir <= ih; ++idir) {
	for (i=0; i<XRES+1; ++i) {
	    mask[i][0] = 780;
	    mask[i][1] = -1;
	}
	xh = (rlowx - highx) / (double) (n);
	yh = (rlowy - highy) / (double) (m);
/* DETERMINE THE DESIRED DIRECTION OF THE LINES ON THE SURFACE */
	if (idir == 2) {
	    xh = (rlowx - highx) / (double) (m);
	    yh = (rlowy - highy) / (double) (n);
	}
/* SCALING */
	scx = 511. / rrmaxx;
	scy = 389. / rrmaxy;

/* IF A UNIFORM SCALING IS DESIRED ,REMOVE THE C FROM THE */
/* FOLLOWING TWO CARDS & */

	if (scx > scy) {
	    scx = scy;
	}
	scy = scx;

	for (i = 1; i <= np1; ++i) {
/* COMPUTE A LINE */
	    ti = (double) (i - 1);
	    xi = highx + ti * xh;
	    if (i != 1) {
/* COMPUTE THE BEGIN AND END SEGMENTS OF EACH LINE */
		icordx[0] = icordx[1];
		icordy[0] = icordy[1];
		icordx[mp3-1] = icordx[mp2-1];
		icordy[mp3-1] = icordy[mp2-1];
	    }
	    for (j = 2; j <= mp2; ++j) {
		tj = (double) (j - 2);
		yi = highy + tj * yh;
		if (idir == 2) {
		    yi = highy + ti * yh;
		    xi = highx + tj * xh;
		}
		zi = fun(xi, yi);
		icordx[j-1] = (xa * xi + xb * yi + zi * 0. - xm) * scx + 513.;
		icordy[j-1] = (ya * xi + yb * yi + yc * zi - ym) * scy + 391.;
	    }
	    if (i == 1) {
		icordx[0] = icordx[1];
		icordy[0] = icordy[1];
		icordx[mp3-1] = icordx[mp2-1];
		icordy[mp3-1] = icordy[mp2-1];
	    }
	    indf = 1;
	    if (ifplt < 0) {
		indf = mp3;
	    }

/* UP TO STATEMENT  190  ,A LINE IS DRAWN ,WITH THE HIDDEN */
/* PARTS NOT SHOWN . */
	    ixo = (int) icordx[indf-1];
	    iyo = (int) icordy[indf-1];

	    ixoi = indf;

	    iplt = 0;
	    if (mask[ixo-1][1] < iyo) iplt = 2;
	    if (mask[ixo-1][0] > iyo) iplt = 1;
	    for (jj = 2; jj <= mp3; ++jj) {
		j = jj;
		if (ifplt < 0) {
		    j = mp3 - jj + 1;
		}
		ixn = (int) icordx[j-1];
		iyn = (int) icordy[j-1];

		ixni = j;

		iplt1 = 0;
		if (mask[ixn-1][1] < iyn) iplt1 = 2;
		if (mask[ixn-1][0] > iyn) iplt1 = 1;
/* THE WHOLE SEGMENT IS HIDDEN */
		if (iplt1 == 0 && iplt == 0) {
		    goto L180;
		}
/* THE SEGMENT INTERSECTS WITH ONE EDGE .DETERMINE THE */
/* INTERSECTION POINT BY LINEAR INTERPOLATION . */
		if (iplt1 == 0 || iplt == 0) {
		    goto L160;
		}
/* THE SEGMENT INTERSECTS WITH THE TWO EDGES . */
		if (iplt1 != iplt) {
		    goto L140;
		}
/* THE WHOLE SEGMENT IS VISIBLE . */

L130:
		iplot(icordx[ixoi-1], icordy[ixoi-1], 
		      icordx[ixni-1], icordy[ixni-1]);

		goto L180;
L140:
		if (mask[ixo-1][0] > iyo && mask[ixo-1][1] < iyo) iplt = iplt1;
		if (mask[ixn-1][0] > iyn && mask[ixn-1][1] < iyn) iplt1 = iplt;
		if (iplt == iplt1) {
		    goto L130;
		}
		ifl = 1;
		ihplt1 = iplt1;
		iplt1 = 0;
L160:
		it = iplt + iplt1;
		a1 = (double) (iyo - iyn);
		a2 = (double) (mask[ixo-1][it-1] - mask[ixn-1][it-1]);
		b1 = (double) (ixn - ixo);
		c1 = (double) (ixn * iyo - ixo * iyn);
		c2 = (double) (ixn * mask[ixo-1][it-1] -
				 ixo * mask[ixn-1][it-1]);
		ddd = b1 * (a1 - a2);
		if (ddd == 0.) {
		    goto L130;
		}

		xi = b1 * (c1 - c2) / ddd;
		yi = (a1 * c2 - a2 * c1) / ddd;

		if (iplt == 0) {
/* DRAW THE VISIBLE PART OF THE SEGMENT . */
		    iplot(xi, yi, icordx[ixni-1], icordy[ixni-1]);
		} else {
		    iplot(icordx[ixoi-1], icordy[ixoi-1], xi, yi);
		}

L180:
		iplt = iplt1;
		if (ifl == 1) {
		    iplt1 = ihplt1;
		    ifl = 0;
		    goto L160;
		}
		ixo = ixn;
		iyo = iyn;

		ixoi = ixni;
	    } /* L190: */

/* UPDATE THE ARRAY MASK ,CONTAINING THE EDGES OF THE */
/* VISIBILITY AREA . */

	    ifplt = -ifplt;
	    indxo = (int) icordx[0];
	    indyo = (int) icordy[0];
	    for (j = 2; j <= mp3; ++j) {
		indxn = (int) icordx[j-1];
		indyn = (int) icordy[j-1];
		if (indxo == indxn) {
		    mask[indxn-1][0] = indyn < mask[indxn-1][0] ?
		       indyn : mask[indxn-1][0];
		    mask[indxn-1][1] = indyn > mask[indxn-1][1] ?
		       indyn : mask[indxn-1][1];
		} else {
		    indo = indxo < indxn ? indxo : indxn;
		    indn = (indxo > indxn ? indxo : indxn) - 1;
		    for (k = indo; k <= indn; ++k) {
			rintp = (double) k * (double) (indyo - indyn);
			rintp += (double) indxo * (double) indyn;
			rintp -= (double) indxn * (double) indyo;
			iintp = (int) (rintp /
					   (double) (indxo - indxn) + .5);
			mask[k-1][0] = iintp < mask[k-1][0] ?
			  iintp : mask[k-1][0];
			mask[k-1][1] = iintp > mask[k-1][1] ?
			  iintp : mask[k-1][1];
		    }
		    indxo = indxn;
		}
		indyo = indyn;
	    }
	}
/* END OF THE PICTURE . */
    }
    return;
}
