#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/setCommand.c,v 2.12 1997/06/24 11:00:03 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif

/*
 * $Log: setCommand.c,v $
 * Revision 2.12  1997/06/24 11:00:03  richter
 * *** empty log message ***
 *
 * Revision 2.12  1997/06/23 19:21:32  richter
 * property was disabled, not plot
 *
 * Revision 2.11  1997/03/31  15:26:40  richter
 * (DatasetObjectCmd): added a missing comma in the call to Tcl_AppendResult
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.2  1996/07/03  14:25:36  richter
 * moved getElement testElement to objManager.c
 * call setAutoConnect for grid files
 * slicing works again
 *
 * Revision 2.9.1.1  1996/05/19  15:58:18  richter
 * lots of changes, especially around slices
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:12  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "ctrProto.h"
#include "vieProto.h"
#include "heiProto.h"
#include "isoProto.h"
#include "molProto.h"
#include "pltProto.h"
#include "relProto.h"
#include "pthProto.h"
#include "setProto.h"
/* #include "slcProto.h"/* */
#include "styProto.h"
#include "strProto.h"
#include "xydProto.h"
#include "objProto.h"
#include "linProto.h"
#include "grdProto.h"

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static int configureDataset(Tcl_Interp *interp, DatasetP dPtr,
			    int argc, char **argv, int new);
static int showcfgDataset(Tcl_Interp *interp, DatasetP dPtr);
static int cfgsliceDataset(Tcl_Interp *interp, DatasetP dPtr,
                           int argc, char **argv, int new);
static int cfgloadDataset(Tcl_Interp *interp, DatasetP dPtr,
                          int argc, char **argv);
static int setConnections(Tcl_Interp *interp, DatasetP dPtr,
                          int argc, char **argv);
static void setAutoConnect(DatasetP dPtr);
static int setPostProcess(Tcl_Interp *interp, DatasetP dPtr);
static int setInfoProperty(Tcl_Interp *interp, DatasetP dPtr);
static int setInfo(Tcl_Interp *interp, DatasetP dPtr);
static int rotateSlice(Tcl_Interp *interp, DatasetP dPtr,
                       int argc, char **argv);
static int translateSlice(Tcl_Interp *interp, DatasetP dPtr,
                          int argc, char **argv);
static int interpolateSlice(Tcl_Interp *interp, DatasetP dPtr,
                          int argc, char **argv);
static void xform2cell(double vout[3], double vin[3],
		       double tmat[3][3], double orig[3]);
static void xform2slice(double vout[3], double vin[3],
			double tmat[3][3], double metr[3][3], double orig[3]);
/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */

int
DatasetCmd(ClientData dummy,		/* Not used. */
	   Tcl_Interp *interp,		/* Current interpreter. */
	   int argc,			/* Number of arguments. */
	   char **argv)			/* Argument strings. */
{
    DatasetP dPtr;

    if (argc < 3) {
	if ( argc == 2 ) {
            if (strEQ(argv[1], "configure", 4)) {
                return showcfgDataset(interp, NULL);
            } else {
                return setListByName(argv[1]);
            }
	} else {
	    return setListByName("*") ;
	}
    } else {
	if ( argv[1][0] != ':' ) {
	    Tcl_AppendResult(interp, "bad dataset name, must start with ':'",
			     (char *) NULL);
	    return TCL_ERROR;
	}

	dPtr = setAlloc(argv[1]);
	if (! dPtr) {
	    Tcl_AppendResult(interp, "dataset creation failed",
                             (char *) NULL);
	    return TCL_ERROR;
	}
	    
	if ( configureDataset(interp, dPtr, argc-2, argv+2, TRUE) != TCL_OK ) {
	    setFree(&dPtr);
	    return TCL_ERROR;
	}

	Tcl_CreateCommand(interp, dPtr->name, DatasetObjectCmd, (ClientData) dPtr,
			  (Tcl_CmdDeleteProc *) NULL );
        if (dPtr->type == DSET_SLICE) {
            interpolateSlice(interp, dPtr, argc-2, argv+2);
        }
        Tcl_SetVar2(interp, "active", "dataset", dPtr->name, TCL_GLOBAL_ONLY);
    }
    return TCL_OK;
}

int
DatasetObjectCmd(ClientData data,               /* dataset		*/
                 Tcl_Interp *interp,            /* Current interpreter. */
                 int argc,                      /* Number of arguments. */
                 char **argv)                   /* Argument strings.	*/
{
    DatasetP dPtr;
    int code;

    dPtr = (DatasetP) data;

    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " option ?args?\"", (char *) NULL);
        return TCL_ERROR;
    }

    if (strEQ(argv[1], "configure", 4)) {
        return configureDataset(interp, dPtr, argc-2, argv+2, FALSE);
    } else if (strEQ(NULL, "connections", 4)) {
        return setConnections(interp, dPtr, argc-2, argv+2);
    } else if (strEQ(NULL, "contour", 4)) {
	char *name = ckalloc( strlen(dPtr->name) + 3 );
	strcpy(name, dPtr->name);
	code = ctrListByName(strcat(name, ":*"));
	ckfree(name);
	return code;
    } else if (strEQ(NULL, "heightfield", 1)) {
	char *name = ckalloc( strlen(dPtr->name) + 3 );
	strcpy(name, dPtr->name);
	code = heiListByName(strcat(name, ":*"));
	ckfree(name);
	return code;
    } else if (strEQ(NULL, "info", 3)) {
	return setInfo(interp, dPtr);
    } else if (dPtr->type == DSET_SLICE && strEQ(NULL, "interpolate", 3)) {
        interpolateSlice(interp, dPtr, argc-2, argv+2);
    } else if (strEQ(NULL, "isosurface", 2)) {
	char *name = ckalloc( strlen(dPtr->name) + 3 );
	strcpy(name, dPtr->name);
	code = isoListByName(strcat(name, ":*"));
	ckfree(name);
	return code;
    } else if (strEQ(NULL, "molecule", 2)) {
	char *name = ckalloc( strlen(dPtr->name) + 3 );
	strcpy(name, dPtr->name);
	code = molListByName(strcat(name, ":*"));
	ckfree(name);
	return code;
    } else if (strEQ(NULL, "path", 2)) {
	char *name = ckalloc( strlen(dPtr->name) + 3 );
	strcpy(name, dPtr->name);
	code = pthListByName(strcat(name, ":*"));
	ckfree(name);
	return code;
    } else if (strEQ(NULL, "plot", 2)) {
	Tcl_AppendResult(interp, "plot is no longer valid for datasets",
           "\nsee molecule", NULL);
        return TCL_OK;
    } else if (strEQ(NULL, "property", 2)) {
        return setInfoProperty(interp, dPtr);
    } else if (strEQ(NULL, "relief", 3)) {
	char *name = ckalloc( strlen(dPtr->name) + 3 );
	strcpy(name, dPtr->name);
	code = relListByName(strcat(name, ":*"));
	ckfree(name);
	return code;
    } else if (strEQ(NULL, "remove", 3)) {
	if ( argc == 2 ) {
	    Tcl_DeleteCommand(interp, dPtr->name);
	    setFree(&dPtr);
	}
    } else if (dPtr->type == DSET_SLICE && strEQ(NULL, "rotate", 1)) {
        rotateSlice(interp, dPtr, argc-2, argv+2);
    } else if (dPtr->type == DSET_SLICE && strEQ(NULL, "translate", 1)) {
        translateSlice(interp, dPtr, argc-2, argv+2);
    } else if (strEQ(NULL, "xydiagram", 2)) {
	char *name = ckalloc( strlen(dPtr->name) + 3 );
	strcpy(name, dPtr->name);
	code = xydListByName(strcat(name, ":*"));
	ckfree(name);
	return code;
    } else {
	Tcl_AppendResult(interp, "unknown option \"", argv[1],
			 "\" must be configure, connection, contour,",
			 "heightfield, info, isosurface, molecule, path,",
                         "property, relief, remove or xydiagram", NULL);
	return TCL_ERROR;
    }

    return TCL_OK;
}

static int
configureDataset(Tcl_Interp *interp,
		 DatasetP dPtr,
		 int argc,
		 char **argv,
                 int new)
{
    if (argc <= 0) {
        return showcfgDataset(interp, dPtr);
    }

    if (strEQ(argv[0], "-slice", 2)) {
        return cfgsliceDataset(interp, dPtr, argc, argv, new);
    } else {
        return cfgloadDataset(interp, dPtr, argc, argv);
    }
}

static int
cfgsliceDataset(Tcl_Interp *interp,
                DatasetP dPtr,
                int argc,
                char **argv,
                int new)
{
    double value;
    double vx[3], vy[3], defpnt[3][3];	/* [point][xyz] */
    int length, ival;
    int i, l;
    DsetSlice *sPtr;
    DatasetP src;
    ObjectP oPtr;

    if (new) {
        dPtr->type = DSET_SLICE;
    } else {
        if (dPtr->type != DSET_SLICE) {
            Tcl_AppendResult(interp, "Not a slice\n", NIL);
            return TCL_ERROR;
        }
    }
    sPtr = dPtr->data.ds = dsetSliceAlloc();
    for (i=0; i<argc; ++i) {
	length = strlen(argv[i]);
	if (strncmp(argv[i],"-crystal",length) == 0) {
	    return TCL_OK;
	} else if (strncmp(argv[i],"-slice",length) == 0) {
	    ++i;
            if (argc-i < 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[i-1], " source\"", NIL);
		return TCL_ERROR;
	    }
            sPtr->source = argv[i];
            src = setFindByName(sPtr->source);
            if (src->type == DSET_GRID) {
                oPtr = &src->data.dg->objects;
            } else if (src->type == DSET_SLICE) {
                oPtr = &src->data.ds->objects;
            } else if (src->type == DSET_FOUR) {
                oPtr = &src->data.df->objects;
            } else {
                Tcl_AppendResult(interp, "slice only valid for GRID, SLICE or FOUR datasets\n", NIL);
		return TCL_ERROR;
	    }                
	} else if (strncmp(argv[i],"-3point",length) == 0) {
	    ++i;
	    if (argc-i < 3) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[i-1], " point-1 point-2 point-3\"", NIL);
		return TCL_ERROR;
	    }
	    for (l=0; l<3; ++l, ++i) {
		if (objGetCoord(interp, argv[i], oPtr, defpnt[l])
		    != TCL_OK) {
		    return TCL_ERROR;
		}
	    }
            --i;
	    for (l=0; l<3; ++l) {
		sPtr->orig[l] = defpnt[0][l];
		vx[l] = defpnt[1][l] - sPtr->orig[l];
		vy[l] = defpnt[2][l] - sPtr->orig[l];
	    }
	    linCoord(sPtr->axes, vx, vy);
	} else if ((strncmp(argv[i],"-xmin",length) == 0)
		   && (length >=4)) {
	    ++i;
	    if (argc-i < 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[i], " <value>", NIL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, argv[i], &value) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    sPtr->min[0] = value;
	} else if ((strncmp(argv[i],"-ymin",length) == 0)
		   && (length >=4)) {
	    ++i;
	    if (argc-i < 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[i], " <value>", NIL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, argv[i], &value) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    sPtr->min[1] = value;
	} else if ((strncmp(argv[i],"-zmin",length) == 0)
		   && (length >=4)) {
	    ++i;
	    if (argc-i < 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[i], " <value>", NIL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, argv[i], &value) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    sPtr->min[2] = value;
	} else if ((strncmp(argv[i],"-xmax",length) == 0)
		   && (length >=4)) {
	    ++i;
	    if (argc-i < 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[i], " <value>", NIL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, argv[i], &value) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    sPtr->max[0] = value;
	} else if ((strncmp(argv[i],"-ymax",length) == 0)
		   && (length >=4)) {
	    ++i;
	    if (argc-i < 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[i], " <value>", NIL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, argv[i], &value) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    sPtr->max[1] = value;
	} else if ((strncmp(argv[i],"-zmax",length) == 0)
		   && (length >=4)) {
	    ++i;
	    if (argc-i < 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[i], " <value>", NIL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, argv[i], &value) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    sPtr->max[2] = value;
	} else if (strcmp(argv[i],"-nx") == 0) {
	    ++i;
	    if (argc-i < 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[i], " <value>", NIL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetInt(interp, argv[i], &ival) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    sPtr->np[0] = ival;
	} else if (strcmp(argv[i],"-ny") == 0) {
	    ++i;
	    if (argc-i < 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[i], " <value>", NIL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetInt(interp, argv[i], &ival) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    sPtr->np[1] = ival;
	} else if (strcmp(argv[i],"-nz") == 0) {
	    ++i;
	    if (argc-i < 1) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[i], " <value>", NIL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetInt(interp, argv[i], &ival) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    sPtr->np[2] = ival;
	} else {
	    Tcl_AppendResult(interp, "unknown option \"",
			     argv[i], "\"", (char *) NULL);
	    return TCL_ERROR;
	}
    }

    return TCL_OK;

}

static int
cfgloadDataset(Tcl_Interp *interp,
		 DatasetP dPtr,
		 int argc,
		 char **argv)
{
    static char defaultFormat[] = "xdgrid";
    char *format;
    char *file;
    size_t length;
    int i, code;

    file   = NULL;
    format = defaultFormat;

    for (i=0; i<argc; ++i) {
	if (strEQ(argv[i], "-load", 2)) {
	    i++;
	    file = argv[i];
	} else if (strEQ(NULL, "-type", 2)) {
	    i++;
	    format = argv[i];
	} else if (strEQ(NULL, "-scale", 2)) {
            /* obsolete */
            ++i;
            if (argc-i < 1) {
                Tcl_AppendResult(interp, "wrong # args: should be \"-scale <value>\"",
                                 NIL);
                return TCL_ERROR;
            }
	} else {
	    Tcl_AppendResult(interp, "unknown option \"",
			     argv[i], "\"", (char *) NULL);
	    return TCL_ERROR;
	}
    }

    if (file) {
	setClean(dPtr, TRUE, FALSE);
        dPtr->type = DSET_UNKNOWN;
        length = strlen(format);
        if (strncmp(format,"xdfmap",length) == 0 ) {
            dPtr->type = DSET_FOUR;
	} else if ( strncmp(format,"xdpath",length ) == 0 ) {
            dPtr->type = DSET_PATH;
        } else if ( strncmp(format,"xdgrid",length ) == 0 ) {
            dPtr->type = DSET_GRID;
        } else if ( strncmp(format,"xddata",length ) == 0 ) {
            dPtr->type = DSET_DATA;
        }
        if (dPtr->type != DSET_UNKNOWN) {
	    if ( (code=setLoad(interp, dPtr, file)) ) {
		if (code == -1) {
		    Tcl_AppendResult(interp, "end of file during read\n",
                                     (char *) NULL);
                } else if (code < 0 || code >= 1000) {
		    char codbuf[100];
		    sprintf( codbuf, "\n\t error code: %d\n", code);
		    Tcl_AppendResult(interp, codbuf, (char *) NULL);
		} else {
		    Tcl_AppendResult(interp, strerror(code), (char *) NULL);
		}
		return TCL_ERROR;
	    }
        } else if (strncmp(format,"aim", length) == 0 ) {
            dPtr->type = DSET_GRID;
	    code = loadGrdaim(dPtr, file);
	    if (code) {
		Tcl_AppendResult(interp, strerror(code), (char *) NULL);
		return TCL_ERROR;
	    }
	} else {
	    Tcl_AppendResult(interp, "unknown grid file format \"", format,
			     "\": should be", " xdfmap, xdgrid or xdpath",
			     (char *) NULL);
	    return TCL_ERROR;
	}
    }
    setPostProcess(interp, dPtr);
    setAutoConnect(dPtr);
    return TCL_OK;
}

static int
showcfgDataset(Tcl_Interp *interp,
               DatasetP dPtr)
{
    Tcl_DString cfg;

    Tcl_DStringInit(&cfg);
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "load");
    Tcl_DStringAppendElement(&cfg, "file::create");
    Tcl_DStringAppendElement(&cfg, "Filename");
    Tcl_DStringAppendElement(&cfg, "");
    Tcl_DStringAppendElement(&cfg, "");
    Tcl_DStringEndSublist(&cfg);

    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "type");
    Tcl_DStringAppendElement(&cfg, "choice:1,xdgrid,xdfmap,xdpath,xddata,aim:create");
    Tcl_DStringAppendElement(&cfg, "Filetype");
    Tcl_DStringAppendElement(&cfg, "xdgrid");
    Tcl_DStringAppendElement(&cfg, "");
    Tcl_DStringEndSublist(&cfg);
    
    if (dPtr) {
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_compid");
        Tcl_DStringAppendElement(&cfg, dPtr->compid);
        Tcl_DStringEndSublist(&cfg);
        
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_title");
        Tcl_DStringAppendElement(&cfg, dPtr->title);
        Tcl_DStringEndSublist(&cfg);
        
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_prop");
        Tcl_DStringAppendElement(&cfg, dPtr->prop);
        Tcl_DStringEndSublist(&cfg);
        
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_flags");
        sprintf(buf, "%d", dPtr->flags);
        Tcl_DStringAppendElement(&cfg, buf);
        Tcl_DStringEndSublist(&cfg);
    }
        
    Tcl_DStringResult(interp, &cfg);
    return TCL_OK;
}

static int
setConnections(Tcl_Interp *interp,
               DatasetP dPtr,
               int argc,
               char **argv)
{
    if (strEQ(argv[0], "-auto", 2)) {
        setAutoConnect(dPtr);
    } else {
        Tcl_AppendResult(interp, "unknown option \"",
                         argv[0], "\"", (char *) NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}

static void
setAutoConnect(DatasetP dPtr)
{
    ObjectP oPtr;
    ConnP cPtr;
    int l, k, cmax;
    double d, dlimit, radl, radk;

    switch (dPtr->type) {
    case DSET_GRID: {
        oPtr = &(dPtr->data.dg->objects);
        cPtr = &(dPtr->data.dg->connex);
    }; break;
    case DSET_FOUR: {
        oPtr = &(dPtr->data.df->objects);
        cPtr = &(dPtr->data.df->connex);
    }; break;
    case DSET_PATH: {
        oPtr = &(dPtr->data.dp->objects);
        cPtr = &(dPtr->data.dp->connex);
    }; break;
    case DSET_SLICE: {
        oPtr = &(dPtr->data.ds->objects);
        cPtr = &(dPtr->data.ds->connex);
    }; break;
    default: return;
    }
    if (cPtr->nc != 0) {
        if (cPtr->obj1) ckfree((char *) cPtr->obj1);
        if (cPtr->obj2) ckfree((char *) cPtr->obj2);
        if (cPtr->style) ckfree((char *) cPtr->style);
        cPtr->nc = 0;
    }
    cmax = ((oPtr->na)*(oPtr->na-1))/2;
    if (cmax > 0) {
        cPtr->obj1 = (int *) ckalloc(cmax  * sizeof(int));
        cPtr->obj2 = (int *) ckalloc(cmax  * sizeof(int));
        cPtr->style = (StyleP *) ckalloc(cmax  * sizeof(StyleP));
        for (l=0; l<oPtr->na; ++l) {
            if (strcasecmp(oPtr->type[l], "ATOM") != 0) continue;
            radl = styGetElementRadius(getElement(oPtr->name[l]));
            for (k=l+1; k<oPtr->na; ++k) {
                if (strcasecmp(oPtr->type[k], "ATOM") != 0) continue;
                d = sqrt((oPtr->coord[l].x - oPtr->coord[k].x)*
                         (oPtr->coord[l].x - oPtr->coord[k].x) +
                         (oPtr->coord[l].y - oPtr->coord[k].y)*
                         (oPtr->coord[l].y - oPtr->coord[k].y) +
                         (oPtr->coord[l].z - oPtr->coord[k].z)*
                         (oPtr->coord[l].z - oPtr->coord[k].z));
                radk = styGetElementRadius(getElement(oPtr->name[k]));
                /*
                  printf("%s %s %.2f %.2f %.2f\n", oPtr->name[l],
                  oPtr->name[k], radl, radk, d);
                  */
                dlimit = 1.2*(radl + radk);
                if (d < dlimit) {
                    cPtr->obj1[cPtr->nc] = l;
                    cPtr->obj2[cPtr->nc] = k;
                    cPtr->style[cPtr->nc] = NULL;
                    ++cPtr->nc;
                }
            }
        }
    }
    pltMessage("inf", "auto generated %d connections\n", cPtr->nc);
}



static int
setPostProcess(Tcl_Interp *interp,
               DatasetP dPtr)
{
    GridP gPtr;
    
    switch (dPtr->type) {
    case DSET_GRID: {
        gPtr = &(dPtr->data.dg->grid);
    }; break;
    case DSET_FOUR: {
        gPtr = &(dPtr->data.df->grid);
    }; break;
    case DSET_SLICE: {
        gPtr = &(dPtr->data.ds->grid);
    }; break;
    default: return TCL_OK;
        }
    return grdSetRange(interp, gPtr);
}

static int
setInfoProperty(Tcl_Interp *interp,
	DatasetP dPtr)
{

    Tcl_SetResult(interp, dPtr->prop, TCL_STATIC);
    return TCL_OK;
}

static int
setInfo(Tcl_Interp *interp,
	DatasetP dPtr)
{
    Tcl_DString cmd;

    Tcl_DStringInit(&cmd);
    Tcl_DStringAppendElement(&cmd, "setinfo");
    Tcl_DStringAppendElement(&cmd, dPtr->name);
    if (Tcl_Eval(interp, Tcl_DStringValue(&cmd)) != TCL_OK) {
        return TCL_ERROR;
    }
    /*
    switch (dPtr->type) {
    default:
	fouInfo(interp, &dPtr->cell);
        grdInfo(interp, &dPtr->grid);
    objInfo(interp, &dPtr->objects);
    cnxInfo(interp, &dPtr->connex);
    }
    */
    return TCL_OK;
}

static int
rotateSlice(Tcl_Interp *interp,
            DatasetP dPtr,
	    int argc,
	    char **argv)
{
    int i,k , length;
    double value, angle[3];
    DsetSlice *sPtr;

    sPtr = dPtr->data.ds;
    
    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: must be \"",
			 " option values ...\"", (char *) NULL);
	return TCL_ERROR;
    }
    for (i=0; i<argc; ++i) {
	length = strlen(argv[i]) > (size_t) 2 ? strlen(argv[i]) : 2;
        if (strncmp(argv[i],"-eulerian",length) == 0) {
	    ++i;
	    if (i+3 < argc) {
		Tcl_AppendResult(interp, "not enough values", (char *) NULL);
		return TCL_ERROR;
	    }
	    for (k=0; k<3; ++k) {
		if ( Tcl_GetDouble(interp, argv[i+k], &value) != TCL_OK ) {
		    return TCL_ERROR;
		}
		angle[k] = value;
	    }
	    linRotate( sPtr->axes, angle[0], 2, 0, 1);
	    linRotate( sPtr->axes, angle[1], 1, 2, 0);
	    linRotate( sPtr->axes, angle[2], 2, 0, 1);
	} else if (strncmp(argv[i],"-x",length) == 0) {
	    ++i;
	    if (i+1 < argc) {
		Tcl_AppendResult(interp, "not enough values", (char *) NULL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, argv[i], &value) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    linRotate(sPtr->axes, value, 0, 1, 2);
	} else if (strncmp(argv[i],"-y",length) == 0) {
	    ++i;
	    if (i+1 < argc) {
		Tcl_AppendResult(interp, "not enough values", (char *) NULL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, argv[i], &value) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    linRotate(sPtr->axes, value, 1, 2, 0);
	} else if (strncmp(argv[i],"-z",length) == 0) {
	    ++i;
	    if (i+1 < argc) {
		Tcl_AppendResult(interp, "not enough values", (char *) NULL);
		return TCL_ERROR;
	    }
	    if ( Tcl_GetDouble(interp, argv[i], &value) != TCL_OK ) {
		return TCL_ERROR;
	    }
	    linRotate(sPtr->axes, value, 2, 0, 1);
	}
    }
    return TCL_OK;
}

static int
translateSlice(Tcl_Interp *interp,
               DatasetP dPtr,
	       int argc,
	       char **argv )
{
    int i;
    double value, shift[3];
    DsetSlice *sPtr;

    sPtr = dPtr->data.ds;
    
    if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args: must be \"",
			 " xtr ytr ztr\"", (char *) NULL);
	return TCL_ERROR;
    }
    for (i=0; i<3; ++i) {
	if ( Tcl_GetDouble(interp, argv[i], &value) != TCL_OK ) {
	    return TCL_ERROR;
	}
	shift[i] = value;
    }

    for (i=0; i<3; ++i) {
	sPtr->orig[i] = sPtr->orig[i] + shift[0]*sPtr->axes[0][i] +
            shift[1]*sPtr->axes[1][i] + shift[2]*sPtr->axes[2][i];
    }
    return TCL_OK;
}

static int
interpolateSlice(Tcl_Interp *interp,
		 DatasetP dPtr,
		 int argc,
		 char **argv)
{
    DatasetP src;
    double incr[3][3], p0[3], pp[3];
    double v;
    int i, j, k, l, n;
    int np[3], size;
    DsetSlice *sPtr;
    GridP gSrc;
    ObjectP oSrc;
    
    sPtr = dPtr->data.ds;
    src = setFindByName(sPtr->source);
    if (src->type == DSET_GRID) {
        gSrc = &src->data.dg->grid;
        oSrc = &src->data.dg->objects;
    } else if (src->type == DSET_SLICE) {
        gSrc = &src->data.ds->grid;
        oSrc = &src->data.dg->objects;
    } else if (src->type == DSET_FOUR) {
        gSrc = &src->data.df->grid;
        oSrc = &src->data.dg->objects;
    } else {
        Tcl_AppendResult(interp, "slice only valid for GRID, SLICE or FOUR datasets\n", NIL);
        return TCL_ERROR;
    }                

    /*
     * calculate increment vectors
     */
    for (i=0; i < 3; ++i) {
	if (sPtr->np[i] > 1) {
	    np[i] = sPtr->np[i];
	    for (k=0; k < 3; ++k) {
		incr[i][k] = sPtr->axes[i][k]
		  * (sPtr->max[i] - sPtr->min[i]) / (double)(np[i]-1);
	    }
	} else {
	    np[i] = 1;
	    for (k=0; k < 3; ++k) incr[i][k] = 0.;
	}
        /*
	printf("incr[%d]:", i);
	for (k=0; k < 3; ++k) printf(" %f", incr[i][k]);
	printf("\n"); /* */
    }

    /*
     * calculate starting point
     */
    xform2cell(p0, sPtr->min, sPtr->axes, sPtr->orig);
    /*    printf("p0: %f %f %f\n", p0[0], p0[1], p0[2]); /* */
    
    sPtr->grid.nx   = sPtr->np[0] > 1 ? sPtr->np[0] : 1;
    sPtr->grid.ny   = sPtr->np[1] > 1 ? sPtr->np[1] : 1;
    sPtr->grid.nz   = sPtr->np[2] > 1 ? sPtr->np[2] : 1;
    sPtr->grid.xran = sPtr->max[0] - sPtr->min[0];
    sPtr->grid.yran = sPtr->max[1] - sPtr->min[1];
    sPtr->grid.zran = sPtr->max[2] - sPtr->min[2];
    dPtr->xsize = sPtr->grid.xran;
    dPtr->ysize = sPtr->grid.yran;
    if (sPtr->grid.nx >1) {
	sPtr->grid.xinc = sPtr->grid.xran / (sPtr->grid.nx-1);
    } else {
	sPtr->grid.xinc = 0.;
    }
    if (sPtr->grid.ny >1) {
	sPtr->grid.yinc = sPtr->grid.yran / (sPtr->grid.ny-1);
    } else {
	sPtr->grid.yinc = 0.;
    }
    if (sPtr->grid.nz >1) {
	sPtr->grid.zinc = sPtr->grid.zran / (sPtr->grid.nz-1);
    } else {
	sPtr->grid.zinc = 0.;
    }
    sPtr->grid.ox = sPtr->min[0];
    sPtr->grid.oy = sPtr->min[1];
    sPtr->grid.oz = sPtr->min[2];
    size  = sPtr->grid.nx * sPtr->grid.ny * sPtr->grid.nz;
    sPtr->grid.va = (double *) ckalloc(size * sizeof(double));

    /*    printf("i k l   cell                            cell\n"); /* */
    for (l=0; l < np[2]; ++l) {
	for (k=0; k < np[1]; ++k) {
	    for (j=0; j<3; ++j) {
		pp[j] = p0[j] + k*incr[1][j] + l*incr[2][j];
	    }
	    for (i=0; i < np[0]; ++i) {
                /*
		printf("%d %d %d:  %f %f %f ",
		       i, k, l, pp[0], pp[1], pp[2]); /* */
		v = grdLinInterpol(gSrc, pp);
		n = i + k*sPtr->grid.nx + l*sPtr->grid.nx*sPtr->grid.ny;
                /*		printf("  v[%d]: %f\n", n, v); /* */
		sPtr->grid.va[n] = v;
		for (j=0; j<3; ++j) pp[j] += incr[0][j];
	    }
	}
    }

    /*
    printf("orig:   ");
    for (i=0; i<3; ++i) printf(" %f", sPtr->orig[i]);
    printf("\n");
    for (i=0; i<3; ++i) {
	printf("axes[%d]:", i);
	for (k=0; k < 3; ++k) printf(" %f", sPtr->axes[i][k]);
	printf("\n");
    } /* */
    if (oSrc->na > 0) {
        double dirmet[3][3] = {{1., 0., 0.}, {0., 1., 0.}, {0., 0., 1.}};
        ObjectP oPtr = &sPtr->objects;
        oPtr->na = oSrc->na;
        oPtr->coord = (c3dT *) ckalloc(oPtr->na * sizeof(c3dT));
        oPtr->name = (char **) ckalloc(oPtr->na * sizeof(char *));
        oPtr->type = (char **) ckalloc(oPtr->na * sizeof(char *));
        oPtr->style = (StyleP *) ckalloc( oPtr->na * sizeof(StyleP));
        for (n=0; n<oSrc->na; ++n) {
            double ct[3], co[3];

            ct[0] =  co[0] = oSrc->coord[n].x;
            ct[1] =  co[1] = oSrc->coord[n].y;
            ct[2] =  co[2] = oSrc->coord[n].z;
            
            xform2slice(ct, co, sPtr->axes, dirmet, sPtr->orig);

            oPtr->coord[n].x = ct[0];
            oPtr->coord[n].y = ct[1];
            oPtr->coord[n].z = ct[2];
            oPtr->name[n] = strDupl(oSrc->name[n]);
            oPtr->type[n] = strDupl(oSrc->type[n]);
            oPtr->style[n] = (StyleP) NULL;
            /*printf("%-8s {%f %f %f} {%f %f %f}\n", oPtr->name[n],
                   co[0], co[1], co[2], ct[0], ct[1], ct[2]);/* */
        }
    }
    return TCL_OK;
}

static void
xform2slice(double vout[3],
	    double vin[3],
	    double tmat[3][3],
	    double metr[3][3],
	    double orig[3])
{
    int i, k, l;
    double h[3];

    for (i=0; i<3; ++i) vout[i] = vin[i] - orig[i];
    for (i=0; i<3; ++i) {
	h[i] = 0.;
	for (k=0; k<3; ++k)
            for (l=0; l<3; ++l)
                h[i] += tmat[i][k] * metr[k][l] * vout[l];
    }
    for (i=0; i<3; ++i) vout[i] = h[i];

}

static void
xform2cell(double vout[3],
	   double vin[3],
	   double tmat[3][3],
	   double orig[3])
{
    int i, k;

    for (i=0; i<3; ++i) {
	vout[i] = orig[i];
	for (k=0; k<3; ++k)
	  vout[i] += tmat[k][i] * vin[k];
    }
}
