#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/setIO.c,v 2.12 1996/11/05 17:26:04 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif

/*
 * $Log: setIO.c,v $
 * Revision 2.12  1996/11/05 17:26:04  richter
 * (readData): changed some line length, fill nvar
 *
 * Revision 2.11  1996/11/02  16:36:47  richter
 * not more than 8 values per data point (readData)
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.4  1996/07/25  05:57:09  richter
 * readFmap takes a file handle as first argument
 *
 * Revision 2.9.1.3  1996/07/03  14:27:26  richter
 * use atof instead of sscanf for grid values
 *
 * Revision 2.9.1.2  1996/07/03  10:04:22  richter
 * some de-linting
 *
 * Revision 2.9.1.1  1996/05/19  16:01:42  richter
 * new file format (xddata)
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:13  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "setProto.h"
#include "linProto.h"
#include "pthProto.h"
#include "strProto.h"
#include "segProto.h"
#ifdef VMS
#include "strcasecmp.h"
#endif

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static int readHead(FILE *fd, DatasetP dPtr, GridP gPtr,
                    ObjectP oPtr, ConnP cPtr);
static int readGrid(FILE *fd, DatasetP dPtr, GridP gPtr);
static int readPath(FILE *fd, DatasetP dPtr);
static int readFmap(FILE *fd, DatasetP dPtr, GridP gPtr, CellP cell,
                    ObjectP oPtr, ConnP cPtr);
static int readData(FILE *fd, DatasetP dPtr, DsetData *dd);
static int fgetline(char *line, int n, FILE *stream);


/*
 *----------------------------------------------------------------------
 * variables with internal linkage
 *----------------------------------------------------------------------
 */
static char separ[] = " \n\t";
static char line[256];
static int err;
/*
 *----------------------------------------------------------------------
 * macros
 *----------------------------------------------------------------------
 */
#define READ_GRIDFILE if ( (err=fgetline(line, 80, grdfile)) ) return(err)

#define GFMT "%lf"

int
setLoad(Tcl_Interp *interp,
	DatasetP dPtr,
	char *file)
{
    FILE *fd;
    int code;
    GridP gPtr;
    ObjectP oPtr;
    ConnP cPtr;
    CellP cell;
    double dx, dy, dz;

    errno = 0;
    fd = fopen(file, "r");
    if (! fd) {
	Tcl_AppendResult(interp, "could not open dataset file", NULL);
	return errno;
    }

    switch (dPtr->type) {
    case DSET_GRID: {
        dPtr->data.dg = dsetGridAlloc();
        gPtr = &dPtr->data.dg->grid;
        oPtr = &dPtr->data.dg->objects;
        cPtr = &dPtr->data.dg->connex;
        if ((code=readHead(fd, dPtr, gPtr, oPtr, cPtr)) != 0) {
            Tcl_AppendResult(interp, "error reading XD grid file (head)", NULL);
            return code;
        }
        if ((code=readGrid(fd, dPtr, gPtr)) != 0) {
            Tcl_AppendResult(interp, "error reading XD grid file (values)", NULL);
            return code;
        }
        dPtr->xsize = gPtr->xran;
        dPtr->ysize = gPtr->yran;
    }; break;
    case DSET_PATH: {
        dPtr->data.dp = dsetPathAlloc();
        gPtr = &dPtr->data.dp->grid;
        oPtr = &dPtr->data.dp->objects;
        cPtr = &dPtr->data.dp->connex;
        if ((code=readHead(fd, dPtr, gPtr, oPtr, cPtr)) != 0) {
            Tcl_AppendResult(interp, "error reading XD path file", NULL);
            return code;
        }
	if ((code=readPath(fd, dPtr)) != 0) {
	    Tcl_AppendResult(interp, "error reading XD path file", NULL);
	    return code;
	}
        dx = dPtr->data.dp->xmax - dPtr->data.dp->xmin;
        dy = dPtr->data.dp->ymax - dPtr->data.dp->ymin;
        dz = dPtr->data.dp->zmax - dPtr->data.dp->zmin;
        dPtr->xsize = dPtr->ysize = sqrt(dx*dx + dy*dy + dz*dz);
    }; break;
    case DSET_FOUR: {
        double rad, pi;
        
        dPtr->data.df = dsetFourAlloc();
        gPtr = &dPtr->data.df->grid;
        oPtr = &dPtr->data.df->objects;
        cPtr = &dPtr->data.df->connex;
        cell = &dPtr->data.df->cell;
	if ( (code=readFmap(fd, dPtr, gPtr, cell, oPtr, cPtr)) != 0) {
	    Tcl_AppendResult(interp, "error reading XD fmap file", NULL);
	    return TCL_ERROR;
	}

        pi = 3.1415926535897932384626433;
        rad = pi/180.;

        cell->dirmet[0][0] = cell->xcell * cell->xcell;
        cell->dirmet[0][1] = cell->xcell * cell->ycell * cos(cell->gamma*rad);
        cell->dirmet[0][2] = cell->xcell * cell->zcell * cos(cell->beta*rad);
        cell->dirmet[1][0] = cell->dirmet[0][1];
        cell->dirmet[1][1] = cell->ycell * cell->ycell;
        cell->dirmet[1][2] = cell->ycell * cell->zcell * cos(cell->alpha*rad);
        cell->dirmet[2][0] = cell->dirmet[0][2];
        cell->dirmet[2][1] = cell->dirmet[1][2];
        cell->dirmet[2][2] = cell->zcell * cell->zcell;

        linInvert(cell->dirmet, cell->rcpmet);
    }; break;
    case DSET_DATA: {
        dPtr->data.dd = dsetDataAlloc();
        if ((code=readData(fd, dPtr, dPtr->data.dd)) != 0) {
            Tcl_AppendResult(interp, "error reading XD data file", NULL);
            return code;
        }
        dPtr->xsize = 10.;
        dPtr->ysize = 10.;
    }; break;
    }
 
    if (fclose(fd) == EOF) {
        return errno;
    } else {
	return 0;
    }
}

static int
readHead(FILE *fd,
	 DatasetP dPtr,
	 GridP gPtr,
         ObjectP oPtr,
         ConnP cPtr)
{
    char *tok;
    int i;

    int iFVers;

/*
 * 1. entry:
 * filetype and version
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    if (strncasecmp( line, "gridfile", 8 ) &&
        strncasecmp( line, "2dgrdfile", 8 ) &&
        strncasecmp( line, "3dgrdfile", 8 ) &&
	strncasecmp( line, "pathfile", 8 )) return(1012);
    sscanf( &line[8], "%d", &iFVers );
/*
 * 2. entry:
 * compound id and property
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    dPtr->compid = strDupl(strtok(line, separ));
    tok = strtok(NULL, separ);
    if ( tok ) {
	dPtr->prop = strDupl(tok);
    } else {
	if (strncmp(line, "        ", 8)) {
	    dPtr->prop = NULL;
	} else {
	    dPtr->prop   = dPtr->compid;
	    dPtr->compid = NULL;
	}
    }
/*
 * 3. entry:
 * title
 */
    if ( (err=fgetline(line, 250, fd)) ) return(err);
    if ( strlen(line) > (size_t) 0 ) {
	tok = &line[strlen(line)-1];
	if ( tok[0] == '\n' ) *tok-- = '\0';
	for (; tok > line && tok[0] == ' '; tok--) {}
	tok[1] = '\0';
	dPtr->title = strDupl(line);
    } else {
	dPtr->title = NULL;
    }
/*
 * 4. entry:
 * gridpoints
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    tok=strtok(line, separ);
    if ( tok ) {
	sscanf( tok, "%d", &gPtr->nx );
    } else {
	return(1013);
    }
    if ( (tok=strtok(NULL, separ)) ) {
	sscanf( tok, "%d", &gPtr->ny );
	if ( (tok=strtok(NULL, separ)) ) {
	    sscanf( tok, "%d", &gPtr->nz );
	} else {
	    gPtr->nz = 0;
	}
    } else {
	gPtr->ny = 0;
	gPtr->nz = 0;
    }
/*
 * 5. entry:
 * origin
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    tok=strtok(line, separ);
    if (tok) {
	sscanf( tok, GFMT, &gPtr->ox );
    } else {
	return(1013);
    }
    if ( gPtr->ny > 0 && (tok=strtok(NULL, separ)) ) {
	sscanf( tok, GFMT, &gPtr->oy );
	if ( gPtr->nz > 0 && (tok=strtok(NULL, separ)) ) {
	    sscanf( tok, GFMT, &gPtr->oz );
	} else {
	    gPtr->oz = 0.;
	}
    } else {
	gPtr->oy = 0.;
	gPtr->oz = 0.;
    }

/*
 * 6. entry:
 * physical dimensions
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    tok=strtok(line, separ);
    if ( tok ) {
	sscanf( tok, GFMT, &gPtr->xran );
    } else {
	return(1013);
    }
    gPtr->xinc = gPtr->xran / (double) (gPtr->nx-1);
    gPtr->yran = 0.;
    gPtr->zran = 0.;
    gPtr->yinc = 1.;
    gPtr->zinc = 1.;
    if ( (tok=strtok(NULL, separ)) ) {
	sscanf( tok, GFMT, &gPtr->yran );
	if (gPtr->ny > 1) {
	    gPtr->yinc = gPtr->yran / (double) (gPtr->ny-1);
	}
	if ( (tok=strtok(NULL, separ)) ) {
	    sscanf( tok, GFMT, &gPtr->zran );
	    if(gPtr->nz > 1) {
		gPtr->zinc = gPtr->zran / (double) (gPtr->nz-1);
	    }
	}
    }

/*
 * 7. entry:
 * objects
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    sscanf( line, "%d", &oPtr->na );
    if ( oPtr->na > 0 ) {
	oPtr->coord = (c3dT *) ckalloc( oPtr->na * sizeof(c3dT) );
	if ( ! oPtr->coord ) {
	    return(1014);
	}
	oPtr->name = (char **) ckalloc( oPtr->na * sizeof(char *));
	if ( ! oPtr->name ) {
	    return(1014);
	}
	oPtr->type = (char **) ckalloc( oPtr->na * sizeof(char *));
	if ( ! oPtr->type ) {
	    return(1014);
	}
	oPtr->style = (StyleP *) ckalloc( oPtr->na * sizeof(StyleP));
	if ( ! oPtr->style ) {
	    return(1014);
	}
	for (i=0; i < oPtr->na; i++) {
	    oPtr->style[i] = (StyleP) NULL;
	    if ( (err=fgetline(line, 80, fd)) ) return(err);
	    tok=strtok(line, separ);
	    if (strlen(tok) > (size_t) 8) tok[8] = '\0';
	    oPtr->name[i]=strDupl(tok);
	    if ( ! (tok=strtok(NULL, separ)) ) return(1015);
	    sscanf(tok, GFMT, &oPtr->coord[i].x);
	    if ( ! (tok=strtok(NULL, separ)) ) return(1015);
	    sscanf(tok, GFMT, &oPtr->coord[i].y);
	    if ( ! (tok=strtok(NULL, separ)) ) return(1015);
	    sscanf(tok, GFMT, &oPtr->coord[i].z);
	    if ( ! (tok=strtok(NULL, separ)) ) return(1015);
	    if (strlen(tok) > (size_t) 4) tok[4] = '\0';
	    oPtr->type[i]=strDupl(tok);
	}
    } else {
	oPtr->coord = NULL;
    }
/*
 * 8. entry:
 * connections
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    sscanf( line, "%d", &cPtr->nc );
    if ( cPtr->nc > 0 ) {
	cPtr->obj1 = (int *) ckalloc( cPtr->nc * sizeof(int));
	if ( ! cPtr->obj1 ) {
	    return(1014);
	}
	cPtr->obj2 = (int *) ckalloc( cPtr->nc * sizeof(int));
	if ( ! cPtr->obj2 ) {
	    return(1014);
	}
	for (i=0; i < cPtr->nc; i++) {
	    if ( (err=fgetline(line, 80, fd)) ) return(err);
	    sscanf(line, "%d %d", &cPtr->obj1[i], &cPtr->obj2[i]) ;
	}
    }

    return 0;
}

static int
readGrid(FILE *fd,
	 DatasetP dPtr,
         GridP gPtr)
{
    char *tok;
    int i, size;

/*
 * 9. entry:
 * grid values
 */
    size = gPtr->nx;
    if ( gPtr->ny > 1 ) size *= gPtr->ny;
    if ( gPtr->nz > 1 ) size *= gPtr->nz;

    gPtr->va = (double *) ckalloc( size * sizeof(double) );
    if ( ! gPtr->va ) return(2001);
    for (i=0; i<size; ) {
	if ( (err=fgetline(line, 80, fd)) ) return(err);
	for (tok=strtok(line, separ); tok&&i<size; tok=strtok(NULL, separ)) {
            /*	    sscanf(tok, GFMT, &gPtr->va[i++]); */
            gPtr->va[i++] = atof(tok);
	}
    }

    return 0;
}

static int
readPath(FILE *fd,
	 DatasetP dPtr)
{
    char *tok;
    int i, k, ncurves;
    int oid, npoints;
    double x, y, z;
    double xmin, xmax, ymin, ymax, zmin, zmax;
    PathdataP *p;
    SegmentP s;
/*
 * 9. entry:
 * curves
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    sscanf( line, "%d", &ncurves );
    p = &dPtr->data.dp->path;
    for (i=0; i<ncurves; ++i) {
	if ( (err=fgetline(line, 80, fd)) ) return(err);
	if ( ! (tok=strtok(line, separ)) ) return(1015);
	sscanf(tok, "%d", &oid);
	if ( ! (tok=strtok(NULL, separ)) ) return(1015);
	sscanf(tok, "%d", &npoints);
	if ( ! (tok=strtok(NULL, separ)) ) return(1015);
	*p = pthdataAlloc(oid, tok, npoints);
	s = (*p)->seg;
	for (k=0; k<npoints; ++k) {
	    if ( (err=fgetline(line, 80, fd)) ) return(err);
	    sscanf(line, "%lf %lf %lf", &x, &y, &z);
	    segAddSegment(&s, x, y, z);
            if (i==0 && k==0) {
                xmin = xmax = x;
                ymin = ymax = y;
                zmin = zmax = z;
            } else {
                if (x < xmin) {
                    xmin = x;
                } else if (x > xmax) {
                    xmax = x;
                }
                if (y < ymin) {
                    ymin = y;
                } else if (y > ymax) {
                    ymax = y;
                }
                if (z < zmin) {
                    zmin = z;
                } else if (z > zmax) {
                    zmax = z;
                }
            }
	}

	p = &(*p)->next;
    }

    dPtr->data.dp->xmin = xmin;
    dPtr->data.dp->xmax = xmax;
    dPtr->data.dp->ymin = ymin;
    dPtr->data.dp->ymax = ymax;
    dPtr->data.dp->zmin = zmin;
    dPtr->data.dp->zmax = zmax;
    return 0;
}

static int
readFmap(FILE *fd,
         DatasetP dPtr,
	 GridP gPtr,
	 CellP cell,
         ObjectP oPtr,
         ConnP cPtr)
{
    char *tok;
    int i, size;

    int iFVers;

/*
 * 1. entry:
 * filetype and version
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    if ( strncasecmp( line, "Fourfile", 8 ) ) return(1012);
    sscanf( &line[8], "%d", &iFVers );
    printf( "gridfile version: %d\n", iFVers );
/*
 * 2. entry:
 * compound id and property
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    dPtr->compid = strDupl(strtok(line, separ));
    tok = strtok(NULL, separ);
    if ( tok ) {
	dPtr->prop = strDupl(tok);
    } else {
	if (strncmp(line, "        ", 8)) {
	    dPtr->prop = NULL;
	} else {
	    dPtr->prop  = dPtr->compid;
	    dPtr->compid = NULL;
	}
    }
    printf( "compound id, property: \"%s\", \"%s\"\n", dPtr->compid, dPtr->prop);
/*
 * 3. entry:
 * title
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    if ( strlen(line) > (size_t) 0) {
	tok = &line[strlen(line)-1];
	if ( tok[0] == '\n' ) *tok-- = '\0';
	for (; tok > line && tok[0] == ' '; tok--) {}
	tok[1] = '\0';
	dPtr->title = strDupl(line);
    } else {
	dPtr->title = NULL;
    }
    printf( "title: \"%s\"\n", dPtr->title);    
/*
 * 4. entry:
 * gridpoints
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    tok=strtok(line, separ);
    if ( tok ) {
	sscanf( tok, "%d", &gPtr->nx );
    } else {
	return(1013);
    }
    gPtr->ny = 0;
    gPtr->nz = 0;
    if ( (tok=strtok(NULL, separ) )) {
	sscanf( tok, "%d", &gPtr->ny );
	if ( (tok=strtok(NULL, separ)) ) {
	    sscanf( tok, "%d", &gPtr->nz );
	}
    }
    printf("nx, ny, nz: %d, %d, %d\n",
	   gPtr->nx, gPtr->ny, gPtr->nz);
/*
 * 5. entry:
 * origin
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    tok=strtok(line, separ);
    if (tok) {
	sscanf( tok, GFMT, &gPtr->ox );
    } else {
	return(1013);
    }
    gPtr->oy = 0.;
    gPtr->oz = 0.;
    if ( gPtr->ny > 0 && (tok=strtok(NULL, separ)) ) {
	sscanf( tok, GFMT, &gPtr->oy );
	if ( gPtr->nz > 0 && (tok=strtok(NULL, separ)) ) {
	    sscanf( tok, GFMT, &gPtr->oz );
	}
    }
    printf("ox, oy, oz: %g, %g, %g\n",
	   gPtr->ox, gPtr->oy, gPtr->oz );
/*
 * 6. entry:
 * physical dimensions
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    tok=strtok(line, separ);
    if ( tok ) {
	sscanf( tok, GFMT, &gPtr->xran );
    } else {
	return(1013);
    }
    gPtr->yran = 0.;
    gPtr->zran = 0.;
    gPtr->yinc = 0.;
    gPtr->zinc = 0.;
    if (gPtr->nx >1) {
	gPtr->xinc = gPtr->xran / (double) (gPtr->nx-1);
    } else {
	gPtr->xinc = 0.;

    }

    if ( (tok=strtok(NULL, separ)) ) {
	sscanf( tok, GFMT, &gPtr->yran );
	if (gPtr->ny>1) gPtr->yinc = gPtr->yran / (double) (gPtr->ny-1);
	if ( (tok=strtok(NULL, separ)) ) {
	    sscanf( tok, GFMT, &gPtr->zran );
	    if (gPtr->nz>1) gPtr->zinc = gPtr->zran / (double) (gPtr->nz-1);
	}
    }
    printf("xran, yran, zran: %g, %g, %g\n",
	   gPtr->xran, gPtr->yran, gPtr->zran);
    printf("xinc, yinc, zinc: %g, %g, %g\n",
	   gPtr->xinc, gPtr->yinc, gPtr->zinc);
/*
 * 7. cell
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    tok=strtok(line, separ);
    if ( tok ) {
	sscanf( tok, GFMT, &cell->xcell );
    } else {
	return(1013);
    }
    cell->ycell = cell->xcell;
    cell->zcell = cell->xcell;
    cell->alpha = 90.;
    cell->beta  = 90.;
    cell->gamma = 90.;
    if ( (tok=strtok(NULL, separ)) ) {
	sscanf( tok, GFMT, &cell->ycell );
	cell->zcell = cell->ycell;
	if ( (tok=strtok(NULL, separ)) ) {
	    sscanf( tok, GFMT, &cell->zcell );
	    if ( (tok=strtok(NULL, separ)) ) {
		sscanf(tok, GFMT, &cell->alpha);
		if ( (tok=strtok(NULL, separ)) ) {
		    sscanf(tok, GFMT, &cell->beta);
		    if ( (tok=strtok(NULL, separ)) ) {
			sscanf(tok, GFMT, &cell->gamma);
		    }
		}
	    }
	}
    }
    printf("xcell, ycell, zcell: %g, %g, %g\n",
	   cell->xcell, cell->ycell, cell->zcell);
    printf("alpha, beta, gamma: %g, %g, %g\n",
	   cell->alpha, cell->beta, cell->gamma);
/*
 * 8. entry:
 * objects
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    sscanf( line, "%d", &oPtr->na );
    if ( oPtr->na > 0 ) {
	oPtr->coord = (c3dT *) ckalloc( oPtr->na * sizeof(c3dT) );
	if ( ! oPtr->coord ) {
	    return(1014);
	}
	oPtr->name = (char **) ckalloc( oPtr->na * sizeof(char *));
	if ( ! oPtr->name ) {
	    return(1014);
	}
	oPtr->type = (char **) ckalloc( oPtr->na * sizeof(char *));
	if ( ! oPtr->type ) {
	    return(1014);
	}
	oPtr->style = (StyleP *) ckalloc( oPtr->na * sizeof(StyleP));
	if ( ! oPtr->style ) {
	    return(1014);
	}
	for (i=0; i < oPtr->na; i++) {
	    oPtr->style[i] = (StyleP) NULL;
	    if ( (err=fgetline(line, 80, fd)) ) return(err);
	    tok=strtok(line, separ);
	    if (strlen(tok) > (size_t) 8) tok[8] = '\0';
	    oPtr->name[i]=strDupl(tok);
	    if ( ! (tok=strtok(NULL, separ)) ) return(1015);
	    sscanf(tok, GFMT, &oPtr->coord[i].x);
	    if ( ! (tok=strtok(NULL, separ)) ) return(1015);
	    sscanf(tok, GFMT, &oPtr->coord[i].y);
	    if ( ! (tok=strtok(NULL, separ)) ) return(1015);
	    sscanf(tok, GFMT, &oPtr->coord[i].z);
	    if ( ! (tok=strtok(NULL, separ)) ) return(1015);
	    if (strlen(tok) > (size_t) 4) tok[4] = '\0';
	    oPtr->type[i]=strDupl(tok);
	    printf("object %d: \"%s\" %g, %g, %g, \"%s\"\n", i, oPtr->type[i], oPtr->coord[i].x, oPtr->coord[i].y, oPtr->coord[i].z, oPtr->name[i] );
	}
    } else {
	oPtr->coord = NULL;
	printf("no objects\n");
    }
/*
 * 8. entry:
 * connections
 */
    if ( (err=fgetline(line, 80, fd)) ) return(err);
    sscanf( line, "%d", &cPtr->nc );
    if ( cPtr->nc > 0 ) {
	cPtr->obj1 = (int *) ckalloc( cPtr->nc * sizeof(int));
	if ( ! cPtr->obj1 ) {
	    return(1014);
	}
	cPtr->obj2 = (int *) ckalloc( cPtr->nc * sizeof(int));
	if ( ! cPtr->obj2 ) {
	    return(1014);
	}
	for (i=0; i < cPtr->nc; i++) {
	    if ( (err=fgetline(line, 80, fd)) ) return(err);
	    sscanf(line, "%d %d", &cPtr->obj1[i], &cPtr->obj2[i]) ;
	    printf("connection %d: %d  --  %d\n", i, cPtr->obj1[i], cPtr->obj2[i]);
	}
    } else {
	printf("no connections\n");
    }
/*
 * 10. entry:
 * grid values
 */
    size = gPtr->nx;
    if ( gPtr->ny > 1 ) size *= gPtr->ny;
    if ( gPtr->nz > 1 ) size *= gPtr->nz;

    if (size > 0) {
	gPtr->va = (double *) ckalloc( size * sizeof(double) );
	if ( ! gPtr->va ) return(2001);
	for (i=0; i<size; ) {
	    if ( (err=fgetline(line, 80, fd)) ) return(err);
	    for (tok=strtok(line, separ); tok; tok=strtok(NULL, separ)) {
		sscanf(tok, GFMT, &gPtr->va[i++]);
		printf( "%d: %g\n", i-1, gPtr->va[i-1] );
	    }
	}
    } else {
	gPtr->va = NULL;
    }

    return errno;
}

int
loadGrdaim(DatasetP dPtr,
	   char *file)
{
    FILE *grdfile;
    GridP gPtr;
    ObjectP oPtr;
    
    int ifun;
    int i, j, nread;
    double cr[2];

    grdfile = fopen( file, "r" );
    if ( ! grdfile ) {
	return errno;
    }
    
    gPtr = &dPtr->data.dg->grid;
    oPtr = &dPtr->data.dg->objects;
    
    READ_GRIDFILE;		/* wave function */
    READ_GRIDFILE;		/* title */
    READ_GRIDFILE;		/* grid size */
    sscanf( line, " %d %d %lf %lf",
	   &gPtr->nx, &gPtr->ny, &gPtr->xran, &gPtr->xinc );
    gPtr->nz = 0;
    gPtr->yran = gPtr->xran;
    gPtr->yinc = gPtr->xinc;
    cr[0] = gPtr->xran / 2.;
    cr[1] = gPtr->yran / 2.;
    if ( gPtr->nx*gPtr->ny ) {
	gPtr->va = (double *) ckalloc( gPtr->nx*gPtr->ny * sizeof(double) );
	if ( ! gPtr->va ) return(2001);
    } else {
	gPtr->va = (double *) NULL;
    }

    READ_GRIDFILE;		/* function type, number of atoms */
    sscanf( line, " %d %d", &ifun, &oPtr->na );
    if ( oPtr->na ) {
	oPtr->coord = (c3dT *) ckalloc( oPtr->na * sizeof(c3dT) );
	if ( ! oPtr->coord ) return(2001);
    } else {
	oPtr->coord = (c3dT *) NULL;
    }

    READ_GRIDFILE;		/* ??? */
    for ( i=0; i<oPtr->na; i++ ) {
	READ_GRIDFILE;
	sscanf( line, " %lf %lf %lf", &oPtr->coord[i].x,
	       &oPtr->coord[i].y,
	       &oPtr->coord[i].z);
	oPtr->coord[i].x += cr[0];
	oPtr->coord[i].y += cr[1];
    }

    for (i = 0; i < gPtr->ny; ++i) {
	for (j = 0; j < gPtr->nx; ++j) {
	    nread =  fscanf( grdfile, " %lf", &gPtr->va[i*gPtr->nx+j] );
	    if ( nread != 1 ) return(errno);
	}
    }
    fclose( grdfile );

    return 0;
}

static int
fgetline(char *line, int n, FILE *stream)
/*
 * Read one line.
 *
 * Return 0 on success, 'errno' on failure and -1 on EOF
 */
{
    char *errmsg;

    for (;;) {
	errmsg = fgets( line, 255, stream );
	if ( ! errmsg ) {
	    if ( errno == 0 ) return -1;
	    return errno;
	}
	if ( line[0] != '!' && line[0] != '#' ) {
	    return 0;
	}
    }
}

static int
readData(FILE *fd,
	 DatasetP dPtr,
         DsetData *dd)
{
    char *tok;
    int i, n, nvar;
    double value;
    int iFVers;

/*
 * 1. entry:
 * filetype and version
 */
    if ( (err=fgetline(line, 81, fd)) ) return(err);
    if ( strncasecmp(line, "datafile", 8) ) return(1012);
    sscanf( &line[8], "%d", &iFVers );
/*
 * 2. entry:
 * compound id and property
 */
    if ( (err=fgetline(line, 250, fd)) ) return(err);
    dPtr->compid = strDupl(strtok(line, separ));
    tok = strtok(NULL, separ);
    if ( tok ) {
	dPtr->prop = strDupl(tok);
    } else {
	if (strncmp(line, "        ", 8)) {
	    dPtr->prop = NULL;
	} else {
	    dPtr->prop   = dPtr->compid;
	    dPtr->compid = NULL;
	}
    }
/*
 * 3. entry:
 * title
 */
    if ( (err=fgetline(line, 250, fd)) ) return(err);
    if ( strlen(line) > (size_t) 0) {
	tok = &line[strlen(line)-1];
	if ( tok[0] == '\n' ) *tok-- = '\0';
	for (; tok > line && tok[0] == ' '; tok--) {}
	tok[1] = '\0';
	dPtr->title = strDupl(line);
    } else {
	dPtr->title = NULL;
    }
    /*
     * 4. entry:
     * nval names ...
     */
    dd->np = 0;
    if ( (err=fgetline(line, 250, fd)) ) return(err);
    tok = strtok(line, separ);
    if (tok) sscanf(tok, "%d", &dd->np );
    if (dd->np <= 0) return 0;
    for (nvar=0; ((tok=strtok(NULL, separ)) && nvar<8); ++nvar) {
        dd->nam[nvar] = strDupl(tok);
        dd->val[nvar] = (double *) ckalloc(dd->np * sizeof(double));
    }
    dd->nvar = nvar;
    for (n=0; n<dd->np; ++n) {
	if ( (err=fgetline(line, 250, fd)) ) return(err);
	for (i=0, tok=strtok(line, separ); tok && i<nvar;
             tok=strtok(NULL, separ), ++i) {
	    sscanf(tok, GFMT, &value);
            if (n==0) {
                dd->min[i] = dd->max[i] = value;
            } else if (value > dd->max[i]) {
                dd->max[i] = value;
            } else if (value < dd->min[i]) {
                dd->min[i] = value;
            }
            dd->val[i][n] = value;
	}
    }

    printf("xmin, xmax %f %f\n", dd->min[0], dd->max[0]);
    printf("ymin, ymax %f %f\n", dd->min[1], dd->max[1]);
    return 0;
}
