#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/setManager.c,v 2.11 1996/09/06 11:52:46 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 * $Log: setManager.c,v $
 * Revision 2.11  1996/09/06 11:52:46  richter
 * call objClean/cnxClean with 'all' parameter
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.4  1996/07/25  05:55:43  richter
 * use Tcl_SetVar2 to set active(dataset)
 * handle plot subcommand (print error message, but continue)
 * print correct message for end-of-file cases
 *
 * Revision 2.9.1.3  1996/07/03  14:28:03  richter
 * clean slices correctly
 *
 * Revision 2.9.1.2  1996/07/03  10:01:24  richter
 * move testElement and getElement to objManager.c
 * calll setAutoConnect after reading gridfile
 * configure output
 *
 * Revision 2.9.1.1  1996/05/19  16:02:26  richter
 * introduced set types (not the same as file types)
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:14  richter
 * First semi-public beta release
 *
 *
 */
/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "setProto.h"
#include "namProto.h"
#include "cnxProto.h"
#include "grdProto.h"
#include "objProto.h"
#include "strProto.h"
#include "fouProto.h"
#include "pthProto.h"

DatasetP
setAlloc(char *name)
{
    DatasetP dPtr;

    dPtr = (DatasetP) ckalloc( sizeof(DatasetT) );
    if (! dPtr) {
	Tcl_AppendResult(interp, "out of memory", (char *) NULL);
	return NULL;
    }

    setClean(dPtr, FALSE, TRUE);
    dPtr->name = strDupl(name);
    if (namAlloc("SET", name, (void *)dPtr) != TCL_OK) {
        setFree(&dPtr);
        dPtr = NULL;
    }

    return dPtr;
}

void
setClean(DatasetP dPtr,
	 int old,		/* was in use ==> free malloc'ed objects  */
	 int all)		/* clear everything, otherwise only data  */
				/* loaded from file is cleaned/freed	  */
{
    int i,k;
    
    if (!dPtr) return;

    if (old) {
        switch (dPtr->type) {
        case DSET_GRID: {
            if (dPtr->data.dg) {
                grdClean(&dPtr->data.dg->grid, old);
                objClean(&dPtr->data.dg->objects, old);
                cnxClean(&dPtr->data.dg->connex, old);
            }
        }; break;
        case DSET_FOUR: {
            if (dPtr->data.df) {
                grdClean(&dPtr->data.df->grid, old);
                objClean(&dPtr->data.df->objects, old);
                cnxClean(&dPtr->data.df->connex, old);
                fouClean(&dPtr->data.df->cell, old);
            }
        }; break;
        case DSET_PATH: {
            if (dPtr->data.dp) {
                grdClean(&dPtr->data.dp->grid, old);
                objClean(&dPtr->data.dp->objects, old);
                cnxClean(&dPtr->data.dp->connex, old);
                if (dPtr->data.dp->path) pthdataFree(&dPtr->data.dp->path);
            }
        }; break;
        case DSET_SLICE: {
            if (dPtr->data.ds) {
                grdClean(&dPtr->data.ds->grid, old);
                objClean(&dPtr->data.ds->objects, old);
                cnxClean(&dPtr->data.ds->connex, old);
                for (i=0; i<3; ++i) {
                    for (k=0; k<3; ++k) {
                        dPtr->data.ds->axes[i][k] = 0.;
                    }
                    dPtr->data.ds->axes[i][i] = 1.;
                    dPtr->data.ds->orig[i]    = 0.;
                    dPtr->data.ds->np[i]      = 0;
                    dPtr->data.ds->min[i]     = 0.;
                    dPtr->data.ds->max[i]     = 0.;
                }
            }
        }; break;
        case DSET_DATA: {
        }; break;
        default: {
            dPtr->type = DSET_UNKNOWN;
            if (dPtr->data.nl) {
                ckfree((char *) dPtr->data.nl);
                dPtr->data.nl = NULL;
            }
        }; break;
        }
    } else {
        dPtr->type = DSET_UNKNOWN;
        dPtr->data.nl = NULL;
    }

    if (old) {
	if (all) {
	    if (dPtr->name) ckfree(dPtr->name);
	}
	if (dPtr->compid) ckfree(dPtr->compid);
	if (dPtr->prop)   ckfree(dPtr->prop);
	if (dPtr->title)  ckfree(dPtr->title);
    }

    if (all) {
	dPtr->name =  NULL;
	dPtr->flags = 0;
        dPtr->xsize = 1.;
        dPtr->ysize = 1.;
    }
    dPtr->compid = NULL;
    dPtr->prop =   NULL;
    dPtr->title =  NULL;
    return;
}

void
setFree(DatasetP *set)
{
    DatasetP dPtr;

    dPtr = *set;
    if (!dPtr) return;

    namFree("SET", dPtr->name);

    switch (dPtr->type) {
    case DSET_GRID: {
        if (dPtr->data.dg) {
            grdClean(&dPtr->data.dg->grid, TRUE);
            objClean(&dPtr->data.dg->objects, TRUE);
            cnxClean(&dPtr->data.dg->connex, TRUE);
        }
    }; break;
    case DSET_FOUR: {
        if (dPtr->data.df) {
            grdClean(&dPtr->data.df->grid, TRUE);
            objClean(&dPtr->data.df->objects, TRUE);
            cnxClean(&dPtr->data.df->connex, TRUE);
            fouClean(&dPtr->data.df->cell, TRUE);
        }
    }; break;
    case DSET_PATH: {
        if (dPtr->data.dp) {
            grdClean(&dPtr->data.dp->grid, TRUE);
            objClean(&dPtr->data.dp->objects, TRUE);
            cnxClean(&dPtr->data.dp->connex, TRUE);
            if (dPtr->data.dp->path) pthdataFree(&dPtr->data.dp->path);
        }
    }; break;
    case DSET_SLICE: {
        if (dPtr->data.ds) {
            grdClean(&dPtr->data.ds->grid, TRUE);
            objClean(&dPtr->data.ds->objects, TRUE);
            cnxClean(&dPtr->data.ds->connex, TRUE);
        }
    }; break;
    case DSET_DATA:
    default: {
    }; break;
    }
    if (dPtr->data.nl) {
        ckfree((char *) dPtr->data.nl);
        dPtr->data.nl = NULL;
    }
    if (dPtr->name)   ckfree((char *)dPtr->name);
    if (dPtr->compid) ckfree(dPtr->compid);
    if (dPtr->prop)   ckfree(dPtr->prop);
    if (dPtr->title)  ckfree(dPtr->title);

    ckfree((char *)dPtr);
    *set = (DatasetP) NULL;
    return;
}

int
setListByName(char *match)
{
    return namListByName("SET", match);
}

DatasetP
setFindByName(char *name)
{
    return (DatasetP) namFindByName("SET", name);
}

char *
setGetName(char *name)
{
    size_t len;
    static char *p = NULL;

    if ( p ) {
	ckfree(p);
	p = NULL;
    }


    len = strchr(name+1, ':') - name;
    p = ckalloc(len+1);
    if ( ! p ) {
	return NULL;
    }
    strncpy(p, name, len);
    p[len] = '\000';
    return p;
    
}

DsetGrid *
dsetGridAlloc(void)
{
    DsetGrid *d;
    d = (DsetGrid *) ckalloc( sizeof(DsetGrid) );
    if (! d) {
	Tcl_AppendResult(interp, "out of memory", (char *) NULL);
	return NULL;
    }
    grdClean(&d->grid, FALSE);
    objClean(&d->objects, FALSE);
    cnxClean(&d->connex, FALSE);
    return d;
}

DsetFour *
dsetFourAlloc(void)
{
    DsetFour *d;
    d = (DsetFour *) ckalloc( sizeof(DsetFour) );
    if (! d) {
	Tcl_AppendResult(interp, "out of memory", (char *) NULL);
	return NULL;
    }
    grdClean(&d->grid, FALSE);
    objClean(&d->objects, FALSE);
    cnxClean(&d->connex, FALSE);
    return d;
}

DsetPath *
dsetPathAlloc(void)
{
    DsetPath *d;
    d = (DsetPath *) ckalloc( sizeof(DsetPath) );
    if (! d) {
	Tcl_AppendResult(interp, "out of memory", (char *) NULL);
	return NULL;
    }
    grdClean(&d->grid, FALSE);
    objClean(&d->objects, FALSE);
    cnxClean(&d->connex, FALSE);
    d->path = NULL;
    return d;
}

DsetSlice *
dsetSliceAlloc(void)
{
    DsetSlice *d;
    d = (DsetSlice *) ckalloc( sizeof(DsetSlice) );
    if (! d) {
	Tcl_AppendResult(interp, "out of memory", (char *) NULL);
	return NULL;
    }
    grdClean(&d->grid, FALSE);
    objClean(&d->objects, FALSE);
    cnxClean(&d->connex, FALSE);    
    return d;

}

DsetData *
dsetDataAlloc(void)
{
    DsetData *d;
    d = (DsetData *) ckalloc( sizeof(DsetData) );
    if (! d) {
	Tcl_AppendResult(interp, "out of memory", (char *) NULL);
	return NULL;
    }
    dataClean(d, FALSE);
    return d;
}

void
dataClean(DsetData *d,
          int old)
{
    int n;
    
    for (n=0; n<8; ++n) {
        if (old && d->val[n]) ckfree((char *) d->val[n]);
        d->val[n] = NULL;
    }

    return;
}
