#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/styManager.c,v 2.13 1996/11/06 11:02:08 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif
/*
 *======================================================================
 *
 * Module: styManager
 *
 * $Log: styManager.c,v $
 * Revision 2.13  1996/11/06 11:02:08  richter
 * (styClean): set fg.model to UNKNOWNMODEL
 * (styParseArgs): recognize -fg as an abbreviation for -foreground
 *
 * Revision 2.12  1996/11/05  17:23:34  richter
 * (styLoadElements): handle broken lines and duplicate elements more gracefully
 *
 * Revision 2.11  1996/11/03  16:37:41  richter
 * added fill option
 * use style group masks
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.3  1996/07/25  05:58:11  richter
 * bug fix
 *
 * Revision 2.9.1.2  1996/07/03  10:05:38  richter
 * split into styManager, styApplyManager, styFillManager, styCvtManager
 *
 * Revision 2.9.1.1  1996/05/19  16:03:43  richter
 * new routine styParseArgs
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:16  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"               /* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "strProto.h"
#include "styProto.h"
#include "pltProto.h"
#ifdef VMS
#include "strcasecmp.h"
#endif

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */

/*
 *----------------------------------------------------------------------
 * variables with internal linkage
 *----------------------------------------------------------------------
 */
struct ElementS {
    int ian;
    double rad;
    char *fg;
};

static Tcl_HashTable elementTable;

/*
 *----------------------------------------------------------------------
 * functions with external and internal linkage
 *----------------------------------------------------------------------
 */

void
styClean(StyleP sPtr,		/* Style to be cleaned			 */
	 int old)		/* was in use ==> free malloc'ed objects */
{
    int i;

    if (!sPtr) return;

    if (old) {
        if (sPtr->fg.name)      ckfree(sPtr->fg.name);
        if (sPtr->line.pattern) ckfree((char *)sPtr->line.pattern);
        if (sPtr->line.verbal)  ckfree(sPtr->line.verbal);
        if (sPtr->font.name)    ckfree(sPtr->font.name);
        if (sPtr->font.anchor)  ckfree(sPtr->font.anchor);
    }
    sPtr->fg.name      = NULL;
    sPtr->fg.model     = UNKNOWNMODEL;
    for (i=0; i<3; ++i) sPtr->fg.val[i] = 0.;
    sPtr->line.width   = 0.;
    sPtr->line.plength = 0;
    sPtr->line.pattern = NULL;
    sPtr->line.verbal  = NULL;
    sPtr->font.size    = 0.;
    sPtr->font.name    = NULL;
    sPtr->font.anchor  = NULL;
    sPtr->pg.mode      = PG_FILL;
    return;
}

int
styGetConfig(Tcl_DString *o,
             StyleP sPtr,
             int mask,
             int has_2nd,
             int onoff_2nd,
             StyleP sPtr2)
{
    Tcl_DStringStartSublist(o);
    Tcl_DStringAppendElement(o, "_SECTION");
    Tcl_DStringAppendElement(o, "Style Options");
    Tcl_DStringEndSublist(o);

    if (mask&STY_FG_ALL) {
        Tcl_DStringStartSublist(o);
        Tcl_DStringAppendElement(o, "foreground");
        Tcl_DStringAppendElement(o, "color");
        Tcl_DStringAppendElement(o, "Foreground Color");
        Tcl_DStringAppendElement(o, "");
        if (sPtr) {
            styGetConfigForeground(o, sPtr);
        } else {
            Tcl_DStringAppendElement(o, "");
        }
        Tcl_DStringEndSublist(o);
    }

    if (mask&STY_LINE_ALL) {
        Tcl_DStringStartSublist(o);
        Tcl_DStringAppendElement(o, "style");
        Tcl_DStringAppendElement(o, "style");
        Tcl_DStringAppendElement(o, "Linestyle");
        Tcl_DStringAppendElement(o, "solid");
        if (sPtr) {
            Tcl_DStringAppendElement(o, sPtr->line.verbal);
        } else {
            Tcl_DStringAppendElement(o, "");
        }
        Tcl_DStringEndSublist(o);
    }

    if (mask&STY_FONT_ALL) {
    /* sPtr->font */
    }

    if (mask&STY_POLYGON_ALL) {
        Tcl_DStringStartSublist(o);
        Tcl_DStringAppendElement(o, "polygon");
        Tcl_DStringAppendElement(o, "choice:1,point,line,fill");
        Tcl_DStringAppendElement(o, "Polygon Mode");
        Tcl_DStringAppendElement(o, "fill");
        if (sPtr) {
            switch (sPtr->pg.mode) {
            case PG_POINT:
                Tcl_DStringAppendElement(o, "point");
                break;
            case PG_LINE:
                Tcl_DStringAppendElement(o, "line");
                break;
            default:
                Tcl_DStringAppendElement(o, "fill");
                break;
            }
        } else {
            Tcl_DStringAppendElement(o, "");
        }
        Tcl_DStringEndSublist(o);
    }
    
    if (mask&STY_FG_ALL && has_2nd) {
        Tcl_DStringStartSublist(o);
        Tcl_DStringAppendElement(o, "foreground2");
        Tcl_DStringAppendElement(o, "color::onoff");
        Tcl_DStringAppendElement(o, "2nd Foreground Color");
        Tcl_DStringAppendElement(o, "");
        if (onoff_2nd && sPtr2) {
            styGetConfigForeground(o, sPtr2);
        } else {
            Tcl_DStringAppendElement(o, "");
        }
        Tcl_DStringEndSublist(o);
    }

    return TCL_OK;
}

styGetConfigForeground(Tcl_DString *o,
                       StyleP sPtr)
{
    ColorP cPtr = &(sPtr->fg);

    if (cPtr) {
        switch (cPtr->model) {
        case RGBMODEL:
            if (cPtr->name) {
                Tcl_DStringAppendElement(o, cPtr->name);
            } else {
                Tcl_DStringStartSublist(o);
                sprintf(buf, " %f %f %f", cPtr->val[0], cPtr->val[1],
                        cPtr->val[2]);
                Tcl_DStringAppendElement(o, "RGB");
                Tcl_DStringAppend(o, buf, -1);
                Tcl_DStringEndSublist(o);
            }
        }
    } else {
        Tcl_DStringAppendElement(o, NULL);
    }
    return TCL_OK;
}


int
styParseArgs(int *argc,
             char **argv,
             StyleP style1,
             StyleP style2,
             int *has2styles)
{
    int skip, i, k;
    int elemc;
    char **elemv;

    *has2styles = 0;
    for (skip=i=0; i < *argc; ++i) {
        if (strEQ(argv[i],"-foreground", 4) || strEQ(NULL, "-fg", 3)) {
            i++;
            if (i >= *argc) {
                Tcl_AppendResult(interp, "value for \"-foreground\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            if (Tcl_SplitList(interp, argv[i], &elemc, &elemv) != TCL_OK) {
                return TCL_ERROR;
            }
            if (elemc == 4) {
                /* really only one color of {TYP x y z} type */
                elemc = 1;
                elemv[0] = argv[i];
            }
            /* silently ignore these errors
            else if (elemc > 2) {
                Tcl_AppendResult(interp, "too many values", (char *) NULL);
                goto Error;
            } else if (style2 == NULL && elemc == 2) {
                Tcl_AppendResult(interp, "too many values", (char *) NULL);
                goto Error;
            }
            */
            if (styFillFG(interp, style1, elemv[0]) != TCL_OK) {
                goto Error;
            }
            if (elemc >= 2 && style2) {
                *has2styles = 1;
                if (styFillFG(interp, style2, elemv[1])) {
                    goto Error;
                }
            }
            ckfree((char *) elemv);
	} else if (strEQ(NULL,"-style", 4)) {
	    i++;
	    if (i >= *argc) {
		Tcl_AppendResult(interp, "value for \"-style\" missing",
				 (char *) NULL);
		return TCL_ERROR;
	    }
            if (styFillPattern(interp, style1, argv[i]) != TCL_OK)
                return TCL_ERROR;
        } else if (strEQ(NULL, "-polygon", 4)) {
            i++;
	    if (i >= *argc) {
		Tcl_AppendResult(interp, "value for \"-polygon\" missing",
				 (char *) NULL);
		return TCL_ERROR;
	    }
            if (styFillPolygon(interp, style1, argv[i]) != TCL_OK)
                return TCL_ERROR;
        } else {
            argv[skip++] = argv[i];
        }
    }
    *argc = skip;
    return TCL_OK;
    
Error:
    ckfree((char *) elemv);
    elemv = NULL;
    return TCL_ERROR;
}

int
styLoadElements(Tcl_Interp *interp,
                char *file)
{
    FILE *f;
    extern int errno;
    char key[5], *colour, line[81];
    int i, n, new;
    struct ElementS *element;
    Tcl_HashEntry *hPtr;
    
    f = fopen(file, "r");
    if (!f) {
        Tcl_AppendResult(interp, "could not open element style file ",
                         file, NULL);
        Tcl_AppendResult(interp, "\n", strerror(errno), NULL);
        return TCL_ERROR;
    }

    Tcl_InitHashTable(&elementTable, TCL_STRING_KEYS);
    while (fgets(line, 80, f)) {
        element = (struct ElementS *) ckalloc(sizeof(struct ElementS));
        i =  sscanf(line, "%4s %d %lf %n", key, &element->ian, &element->rad,
                    &n);
        if (i == 3) {
            colour = &line[n];
            n = strlen(colour) - 1;
            colour[n] = '\000';
            element->fg = strDupl(colour);
            hPtr = Tcl_CreateHashEntry(&elementTable, key, &new);
            if (! new) {
                ckfree(element->fg);
                ckfree((char *) element);
            } else {
                Tcl_SetHashValue(hPtr, (ClientData) element);
                /*
                  printf("/%-4s/ %3d %4.2f /%s/\n", key,
                  element->ian, element->rad,element->fg);
                  */
            }
        }
    }

    fclose(f);
    return TCL_OK;
}

char *
styGetElementFG(char *key)
{
    Tcl_HashEntry *hPtr;
    struct ElementS *element;
    char *deffg = "gray";

    hPtr = Tcl_FindHashEntry(&elementTable, key);

    if (hPtr) {
        element = (struct ElementS *) Tcl_GetHashValue(hPtr);
	return element->fg;
    } else {
	return deffg;
    }
}

double
styGetElementRadius(char *key)
{
    Tcl_HashEntry *hPtr;
    struct ElementS *element;
    double defrad = 0.49;

    hPtr = Tcl_FindHashEntry(&elementTable, key);

    if (hPtr) {
        element = (struct ElementS *) Tcl_GetHashValue(hPtr);
	return element->rad;
    } else {
	return defrad;
    }
}
