#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/vieCommand.c,v 2.12 1999/06/03 12:47:12 richter Exp $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif

/*
 *======================================================================
 *
 * Module: vieCommand
 *
 * Abstract: Defines the 'relief' command and creates relief object
 * commands.  Called from the tcl interpreter.
 *
 * $Log: vieCommand.c,v $
 * Revision 2.12  1999/06/03 12:47:12  richter
 * first throw at removable views
 *
 * Revision 2.11  1996/11/03 16:39:49  richter
 * don't use styGetConfig
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.3  1996/07/25  06:13:16  richter
 * quick ddr hack
 * view names are all lower case by default
 * use Tcl_SetVar2 to set active(view)
 * handle 'view <view>' correctly
 *
 * Revision 2.9.1.2  1996/07/03  10:55:44  richter
 * many changes
 *
 * Revision 2.9.1.1  1996/05/20  11:21:01  richter
 * lots of changes
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:10  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "vieProto.h"
#include "pltProto.h"
#include "styProto.h"
#include "strProto.h"


/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
/*
  static int configureView(Tcl_Interp *interp, ViewP view,
                         int argc, char **argv);
                         */
static int showcfgView(Tcl_Interp *interp, ViewP view);
static void viewEvalScaling(ViewOptT *vOpt);
/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */

int
ViewCmd(ClientData dummy,		/* Not used			*/
        Tcl_Interp *interp,		/* Current interpreter		*/
        int argc,			/* Number of arguments		*/
        char **argv)			/* Argument strings		*/
{
    ViewP view;
    Tcl_CmdInfo cinfo;
    char *name;

    if (argc == 1) return viewListByName("*");
    
    if (argc == 2 && strEQ(argv[1], "configure", 1)) return showcfgView(interp, NULL);

    name = argv[1];
    if (Tcl_GetCommandInfo(interp, name, &cinfo)) {
	Tcl_AppendResult(interp, "command \"", name, "\" already exists",
			 (char *) NULL);
	return TCL_ERROR;
    }

    view = viewAlloc(name);
    if (! view) {
	return TCL_ERROR;
    }

    argc -= 2;
    argv += 2;
    if (viewParseArgs(NULL, &argc, argv, &view->opt) != TCL_OK) {
        goto Error;
    }
    if (!view->opt.doscale) {
        Tcl_AppendResult(interp, "at least one of -width/-height, -scale or -vrange must be given", (char *) NULL);
        goto Error;
    }
    viewEvalScaling(&view->opt);

    if (argc < 2 || !strEQ(argv[0], "-type", 3)) {
        Tcl_AppendResult(interp, "-type must be specified", (char *) NULL);
        goto Error;
    }
    if (strEQ(argv[1], "xydiag", 4)) {
        view->type = VIEW_XYDIAG;
    } else if (strEQ(NULL, "contour", 4)) {
        view->type = VIEW_CONTOUR;
    } else if (strEQ(NULL, "relief", 4)) {
        view->type = VIEW_RELIEF;
    } else if (strEQ(NULL, "path", 4)) {
        view->type = VIEW_PATH;
    } else if (strEQ(NULL, "height", 4)) {
        view->type = VIEW_HEIGHT;
    } else if (strEQ(NULL, "surface", 4)) {
        view->type = VIEW_SURFACE;
    } else {
        Tcl_AppendResult(interp, "unknow view type", (char *) NULL);
        goto Error;
    }
      
    Tcl_CreateCommand(interp, name, ViewObjectCmd, (ClientData) view,
			  (Tcl_CmdDeleteProc *) NULL );
    
    if (pltCreateView(interp, view) != TCL_OK) {
        return TCL_ERROR;
    }
    Tcl_SetVar2(interp, "active", "view", name, TCL_GLOBAL_ONLY);
    return TCL_OK;

Error:
    viewFree(view);
    return TCL_ERROR;
}

int
ViewObjectCmd(ClientData data,               /* dataset		*/
              Tcl_Interp *interp,            /* Current interpreter. */
              int argc,                      /* Number of arguments. */
              char **argv)                   /* Argument strings.	*/
{
    ObjlistP o;
    ViewP view;
    ViewOptT vOpt;
    Tcl_DString viewcmd;
    int i,k;
    
    view = (ViewP) data;

    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " option ?args?\"", (char *) NULL);
        return TCL_ERROR;
    } else if (strEQ(argv[1], "configure", 1)) {
        if (argc == 2) {
            return showcfgView(interp, view);
        }
        vOpt.doscale = 0;
        vOpt.doorig = 0;
        vOpt.dorot = 0;
        vOpt.dotrans = 0;
        vOpt.scale = view->opt.scale;
        vOpt.width = view->opt.width;
        vOpt.height = view->opt.height;
        vOpt.vrange[0] = view->opt.vrange[0];
        vOpt.vrange[1] = view->opt.vrange[1];
        argc -= 2;
        viewParseArgs(NULL, &argc, argv+2, &vOpt);
        view->opt.scale = vOpt.scale;
        view->opt.width = vOpt.width;
        view->opt.height = vOpt.height;
        view->opt.vrange[0] = vOpt.vrange[0];
        view->opt.vrange[1] = vOpt.vrange[1];
        viewEvalScaling(&view->opt);
        if (vOpt.doorig) {
            view->opt.doorig = vOpt.doorig;
            for (i=0; i<3; ++i) {
                view->opt.orig[i] = vOpt.orig[i];
            }
        }
        if (vOpt.dorot) {
            view->opt.dorot = vOpt.dorot;
            for (i=0; i<3; ++i) {
                for (k=0; k<3; ++k) {
                    view->opt.rotm[i][k] = vOpt.rotm[i][k];
                }
            }
        }
        if (vOpt.dotrans) {
            view->opt.dotrans = vOpt.dotrans;
            for (i=0; i<3; ++i) {
                view->opt.trans[i] = vOpt.trans[i];
            }
        }
    } else if (strEQ(NULL, "ddr", 3)) {
        return pltDDR(view);
    } else if (strEQ(NULL, "objects", 1)) {
        Tcl_DStringInit(&viewcmd);
        for (o=view->head; o; o=o->next) {
            Tcl_DStringStartSublist(&viewcmd);
            Tcl_DStringAppendElement(&viewcmd, o->name);
            Tcl_DStringAppendElement(&viewcmd, o->type);
            Tcl_DStringEndSublist(&viewcmd);
        }
        Tcl_SetResult(interp, Tcl_DStringValue(&viewcmd), TCL_VOLATILE);
    } else if (strEQ(NULL, "remove", 3)) {
        if (pltRemoveView(view) != TCL_OK) { /* destroy window, which in */
            return TCL_ERROR;                /* turn removes objects */
        }                                    /* hopefully we are re-entrant */
        Tcl_DeleteCommand(interp, view->name);
        viewFree(view);
    } else {
	Tcl_AppendResult(interp, "unknown option \"", argv[1],
			 "\" must be configure, ddr, objects or remove\n",
			 (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

int
vieAutoCreate(Tcl_Interp *interp,
              char *name,
              char *type,
              ViewOptT *vOpt)
{
    Tcl_DString viewcmd;
    int k, l;
    char v[20];

    Tcl_DStringInit(&viewcmd);
    Tcl_DStringAppendElement(&viewcmd, "view");
    Tcl_DStringAppendElement(&viewcmd, name);
    Tcl_DStringAppendElement(&viewcmd, "-type");
    Tcl_DStringAppendElement(&viewcmd, type);
    
    if (vOpt->doscale) {
        Tcl_DStringAppendElement(&viewcmd, "-width");
        sprintf(v, "%g", vOpt->width);
        Tcl_DStringAppendElement(&viewcmd, v);
        Tcl_DStringAppendElement(&viewcmd, "-height");
        sprintf(v, "%g", vOpt->height);
        Tcl_DStringAppendElement(&viewcmd, v);
        Tcl_DStringAppendElement(&viewcmd, "-scale");
        sprintf(v, "%g", vOpt->scale);
        Tcl_DStringAppendElement(&viewcmd, v);
        Tcl_DStringAppendElement(&viewcmd, "-vrange");
        Tcl_DStringStartSublist(&viewcmd);
        sprintf(v, "%g", vOpt->vrange[0]);
        Tcl_DStringAppendElement(&viewcmd, v);
        sprintf(v, "%g", vOpt->vrange[1]);
        Tcl_DStringAppendElement(&viewcmd, v);
        Tcl_DStringEndSublist(&viewcmd);
    }
    
    if (vOpt->dotrans) {
        Tcl_DStringAppendElement(&viewcmd, "-translate");
        Tcl_DStringStartSublist(&viewcmd);
        sprintf(v, "%g", vOpt->trans[0]);
        Tcl_DStringAppendElement(&viewcmd, v);
        sprintf(v, "%g", vOpt->trans[1]);
        Tcl_DStringAppendElement(&viewcmd, v);
        sprintf(v, "%g", vOpt->trans[2]);
        Tcl_DStringAppendElement(&viewcmd, v);
        Tcl_DStringEndSublist(&viewcmd);
    }
    
    if (vOpt->doorig) {
        Tcl_DStringAppendElement(&viewcmd, "-origin");
        Tcl_DStringStartSublist(&viewcmd);
        sprintf(v, "%g", vOpt->orig[0]);
        Tcl_DStringAppendElement(&viewcmd, v);
        sprintf(v, "%g", vOpt->orig[1]);
        Tcl_DStringAppendElement(&viewcmd, v);
        sprintf(v, "%g", vOpt->orig[2]);
        Tcl_DStringAppendElement(&viewcmd, v);
        Tcl_DStringEndSublist(&viewcmd);
    }

    if (vOpt->dorot) {
        Tcl_DStringAppendElement(&viewcmd, "-matrix");
        Tcl_DStringStartSublist(&viewcmd);
        for (k=0; k<3; ++k) {
            for (l=0; l<3; ++l) {
                sprintf(v, "%g", vOpt->rotm[k][l]);
                Tcl_DStringAppendElement(&viewcmd, v);
            }
        }
        Tcl_DStringEndSublist(&viewcmd);
    }
    pltMessage("inf", "auto creating %s\n", Tcl_DStringValue(&viewcmd));
    return Tcl_Eval(interp, Tcl_DStringValue(&viewcmd));
}

static int
showcfgView(Tcl_Interp *interp,
            ViewP view)
{
    Tcl_DString cfg;

    Tcl_DStringInit(&cfg);
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "type");
    Tcl_DStringAppendElement(&cfg, "choice:1,xydiag,contour,relief,path,height,surface:create");
    Tcl_DStringAppendElement(&cfg, "Viewtype");
    Tcl_DStringAppendElement(&cfg, "");
    if (view) {
        switch (view->type) {
        case VIEW_XYDIAG: {
            Tcl_DStringAppendElement(&cfg, "xydiag");
        }; break;
        case VIEW_CONTOUR: {
            Tcl_DStringAppendElement(&cfg, "contour");
        }; break;
        case VIEW_RELIEF: {
            Tcl_DStringAppendElement(&cfg, "relief");
        }; break;
        case VIEW_PATH: {
            Tcl_DStringAppendElement(&cfg, "path");
        }; break;
        case VIEW_HEIGHT: {
            Tcl_DStringAppendElement(&cfg, "height");
        }; break;
        case VIEW_SURFACE: {
        Tcl_DStringAppendElement(&cfg, "surface");
        }; break;
        default: {
            Tcl_DStringAppendElement(&cfg, "unknown");
        }; break;
        }
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);

    if (view) {
        viewGetConfig(&cfg, &view->opt, 3);
        /*        styGetConfig(&cfg, &view->style, STY_ALL, 0, 0, NULL); /* */
    } else {
        viewGetConfig(&cfg, NULL, 3);
        /*        styGetConfig(&cfg, NULL, STY_ALL, 0, 0, NULL); /* */
    }

    Tcl_DStringResult(interp, &cfg);
    return TCL_OK;
}

static void
viewEvalScaling(ViewOptT *vOpt)
{
    if (vOpt->scale <= 0.) {
        double w, h;
        w = vOpt->width < 0. ? 18. : vOpt->width;
        h = vOpt->height < 0. ? 18. : vOpt->height;
        vOpt->scale = w / vOpt->vrange[0];
        if (vOpt->scale*vOpt->vrange[1] > h) {
            vOpt->scale = h /vOpt->vrange[1] ;
        }
    }
    if (vOpt->width < 0.) {
        vOpt->width = vOpt->vrange[0] * vOpt->scale;
    }
    if (vOpt->height < 0.) {
        vOpt->height = vOpt->vrange[1] * vOpt->scale;
    }
}
