#ifndef lint
static char *rcsid = "$Header: /scratch/fock/cvsroot/xd/xdgraph/src/xydCommand.c,v 2.14 1997/03/29 18:11:55 richter Rel $";
static char *copyright = "Copyright (c) 1995 The XD Group, Free University of Berlin. All rights reserved.";
#endif

/*
 * $Log: xydCommand.c,v $
 * Revision 2.14  1997/03/29 18:11:55  richter
 * preset style
 *
 * Revision 2.13  1996/11/06  10:57:03  richter
 * larger default view
 * preset styles with colour and anchor
 * ticsize is no more easily changeable
 *
 * Revision 2.12  1996/11/05  17:21:52  richter
 * added options -x, -y and -type (major changes)
 *
 * Revision 2.11  1996/09/06  12:00:39  richter
 * many changes to use the new display list interface
 *
 * Revision 2.10  1996/07/25  09:14:52  richter
 * new numbers because CV (emacs package) does not unterstand branches
 *
 * Revision 2.9.1.2  1996/07/25  06:17:25  richter
 * use Tcl_SetVar2 to set active(xydiagramm)
 * handle 'xydiag <xydiag>' correctly
 *
 * Revision 2.9.1.1  1996/07/03  10:58:22  richter
 * view scaling'configure output
 *
 * Revision 2.9  1996/05/20  11:29:26  richter
 * *** empty log message ***
 *
 * Revision 2.9  1996/04/09  15:54:28  richter
 * first ci for xdgraph (formerly xd2d)
 *
 * Revision 2.0  1995/05/12  11:54:12  richter
 * First semi-public beta release
 *
 *
 */

/*
 *----------------------------------------------------------------------
 * include files
 *----------------------------------------------------------------------
 */
#include "conf.h"		/* global configuration                 */
#include "xdgraph.h"		/* type definitions and array sizes     */
#include "cmdProto.h"
#include "xydProto.h"
#include "vieProto.h"
#include "pltProto.h"
#include "setProto.h"
#include "styProto.h"
#include "strProto.h"
#include "segProto.h"
#include "dlProto.h"

/*
 *----------------------------------------------------------------------
 * prototypes with internal linkage
 *----------------------------------------------------------------------
 */
static int configureXydiag(Tcl_Interp *interp, XydiagP xPtr,
			    int argc, char **argv, ViewOptT *vOpt);
static int showcfgXydiag(Tcl_Interp *interp, XydiagP xPtr);
static int plotXydiag(Tcl_Interp *interp, ViewP view, XydiagP xPtr);
static int clearXydiag(Tcl_Interp *interp, ViewP view, XydiagP xPtr);
static int plotAxis(Tcl_Interp *interp, ViewP view, AxisP axis,
                    StyleP style, int which);
static int clearAxis(Tcl_Interp *interp, ViewP view, AxisP axis);
static void axis(double xmin, double xmax,
                 double *x0, double *x1, double *step);
static void xydEngine(XydiagP xPtr, DsetData *dd);
/*
 *----------------------------------------------------------------------
 * variables with external linkage
 *----------------------------------------------------------------------
 */

int
XydiagCmd(ClientData dummy,		/* Not used. */
	   Tcl_Interp *interp,		/* Current interpreter. */
	   int argc,			/* Number of arguments. */
	   char **argv)			/* Argument strings. */
{
    XydiagP xPtr;
    DatasetP dPtr;
    DsetData *dd;
    ViewOptT vOpt;
    char *objName;
    ViewE vieType = VIEW_XYDIAG;

    if (argc == 1 ||
        (argc == 2 && strEQ(argv[1], "configure", 4))) {
         return showcfgXydiag(interp, NULL);
    }

    if (cmdAnalyzeName(argv[1], &objName, &dPtr) != TCL_OK) {
        return TCL_ERROR;
    }
    switch (dPtr->type) {
    case DSET_DATA: {
    }; break;
    default: {
	Tcl_AppendResult(interp, "xydiagram is only valid for DATA datasets\n",
                         (char *) NULL);
	return TCL_ERROR;
    }
    }
    dd = dPtr->data.dd;

    xPtr = xydAlloc(argv[1]);
    if (!xPtr) {
        Tcl_AppendResult(interp, "Xydiag creation failed", (char *) NULL);
        return TCL_ERROR;
    }
    xPtr->object = strDupl(objName);
    if (dd->nvar == 1) {
        xPtr->xidx = 0;
        xPtr->yidx = 1;
    } else {
        xPtr->xidx = 1;
        xPtr->yidx = 2;
    }
    vOpt.doscale = 1;
    vOpt.doorig  = 1;
    vOpt.orig[0] = 0.5;
    vOpt.orig[1] = 0.5;
    vOpt.orig[2] = 0.5;
    vOpt.dorot   = 0;
    vOpt.dotrans = 0;
    vOpt.scale = 15.;
    vOpt.width = 15.;
    vOpt.height = 15.;
    styFillFG(interp, &xPtr->style, "black");
    styFillLinewidth(interp, &xPtr->style, 1.);
    styFillPattern(interp, &xPtr->style, "solid");
    
    styFillFG(interp, &xPtr->xax.style, "black");
    styFillLinewidth(interp, &xPtr->style, 1.);
    styFillPattern(interp, &xPtr->xax.style, "solid");
    styFillAnchor(interp, &xPtr->xax.style, "sw");

    styFillFG(interp, &xPtr->yax.style, "black");
    styFillLinewidth(interp, &xPtr->style, 1.);
    styFillPattern(interp, &xPtr->yax.style, "solid");
    styFillAnchor(interp, &xPtr->yax.style, "s");
    if (configureXydiag(interp, xPtr, argc-2, argv+2, &vOpt) != TCL_OK) {
        goto Error;
    }

    vOpt.vrange[0] = 1.;
    vOpt.vrange[1] = 1.;
    if (cmdSetView(interp, &xPtr->view, 1, &vieType, objName,
                   argv[1], "xydiagram", &vOpt) != TCL_OK) {
        Tcl_AppendResult(interp, "view could not be set\n", NULL);
        goto Error;
    }

    Tcl_CreateCommand(interp, xPtr->name, XydiagObjectCmd, (ClientData) xPtr,
			  (Tcl_CmdDeleteProc *) NULL );
    Tcl_SetVar2(interp, "active", "xydiagram", xPtr->name, TCL_GLOBAL_ONLY);

    xydEngine(xPtr, dd);
        
    return TCL_OK;

Error:
    xydFree(&xPtr);
    return TCL_ERROR;
}

static void
xydEngine(XydiagP xPtr,
          DsetData *dd)
{
    double ax, ay;
    double min, max;
    SegmentP p;
    int xidx, yidx, i;

    xidx = xPtr->xidx - 1;
    if (xPtr->xidx == 0) {
        min = 0.;
        max = (double) (dd->np-1);
    } else {
        min = dd->min[xidx];
        max = dd->max[xidx];
    }
    axis(min, max, &xPtr->xax.min, &xPtr->xax.max, &xPtr->xax.step);

    yidx = xPtr->yidx - 1;
    if (xPtr->yidx == 0) {
        min = 0.;
        max = (double) (dd->np-1);
    } else {
        min = dd->min[yidx];
        max = dd->max[yidx];
    }
    axis(min, max, &xPtr->yax.min, &xPtr->yax.max, &xPtr->yax.step);
        
    ax = xPtr->xax.max - xPtr->xax.min;
    ay = xPtr->yax.max - xPtr->yax.min;
    if (xPtr->points) segFree(&xPtr->points);
    if (xPtr->type == XYD_LINES) {
        p = xPtr->points = segAlloc(dd->np+2, 0);
    } else {
        p = xPtr->points = segAlloc(dd->np+2, POINTSET);
    }
    for (i=0; i<dd->np; ++i) {
        double x, y, xraw, yraw;

        xraw = xPtr->xidx ? dd->val[xidx][i] : (double) i;
        yraw = xPtr->yidx ? dd->val[yidx][i] : (double) i;
        x = (xraw - xPtr->xax.min) / ax;
        y = (yraw - xPtr->yax.min) / ay;
        if (xPtr->type == XYD_LINES) {
            segAddSegment(&p, x, y, 0.);
        } else {
            segAddPoint(&p, x, y, 0.);
        }
    }
}

int
XydiagObjectCmd(ClientData data,              /* Xydiag		*/
                  Tcl_Interp *interp,           /* Current interpreter. */
                  int argc,                     /* Number of arguments. */
                  char **argv)                  /* Argument strings.	*/
{
    XydiagP xPtr;
    ViewP view;
    ViewOptT vOpt;

    xPtr = (XydiagP) data;
    view = viewFindByName(xPtr->view, NULL);

    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " option ?args?\"", (char *) NULL);
        return TCL_ERROR;
    }

    if (strEQ(argv[1], "configure", 2)) {	
        if (argc <= 2) {
            return showcfgXydiag(interp, xPtr);
        } else {
            return configureXydiag(interp, xPtr, argc-2, argv+2, &vOpt);
        }
    } else if (strEQ(NULL, "clear", 2)) {
        clearXydiag(interp, view, xPtr);
    } else if (strEQ(NULL, "info", 1)) {
        return TCL_OK;
        /*	return infoXydiag(interp, xPtr); /* */
    } else if (strEQ(NULL, "plot", 2)) {
        clearXydiag(interp, view, xPtr);
	return plotXydiag(interp, view, xPtr);
    } else if (strEQ(NULL, "remove", 3)) {
        clearXydiag(interp, view, xPtr);
        viewUnregisterObject(view, xPtr->name);
        Tcl_DeleteCommand(interp, xPtr->name);
        xydFree(&xPtr);
    } else {
	Tcl_AppendResult(interp, "unknown option \"", argv[1],
			 "\" must be configure, clear, info, plot ",
			 " or remove\n", (char *) NULL);
	return TCL_ERROR;
    }

    return TCL_OK;
}

static int
configureXydiag(Tcl_Interp *interp,
                XydiagP xPtr,
                int argc,
                char **argv,
                ViewOptT *vOpt)
{
    DatasetP dPtr;
    DsetData *dd;
    int has2styles, code;
    int elemc;
    char **elemv;
    int i, idx;

    if (argc <= 0) {
        return showcfgXydiag(interp, xPtr);
    }
    dPtr = setFindByName(xPtr->object);
    dd = dPtr->data.dd;
    
    code = viewParseArgs(NULL, &argc, argv, vOpt);
    if (code != TCL_OK) return code;

    styParseArgs(&argc, argv, &xPtr->style, NULL, &has2styles);
    
    for (i=0; i<argc; ++i) {
        if (strEQ(argv[i], "-view", 3)) {
            if (i++ >= argc) {
                Tcl_AppendResult(interp, "value for \"-view\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            xPtr->view = strDupl(argv[i]);
        } else if (strEQ(NULL, "-type", 3)) {
            if (i++ >= argc) {
                Tcl_AppendResult(interp, "value for \"-type\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            if (strncmp(argv[i], "point", 5) == 0) {
                xPtr->type = XYD_POINTS;
            } else if (strncmp(argv[i], "line", 4) == 0) {
                xPtr->type = XYD_LINES;
            } else {
                xPtr->type = XYD_UNKNOWN;
            }
        } else if (strEQ(NULL, "-x", 2)) {
            if (i++ >= argc) {
                Tcl_AppendResult(interp, "value for \"-x\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            if (Tcl_GetInt(interp, argv[i], &idx) != TCL_OK) {
                int match = FALSE;
                for (idx=0; idx<dd->nvar; ++idx) {
                    if (strcmp(dd->nam[idx], argv[i]) == 0) {
                        xPtr->xidx = idx+1;
                        match = TRUE;
                        break;
                    }
                }
                if (!match) {
                    Tcl_AppendResult(interp, "invalid variable on -x option",
                                     (char *) NULL);
                    return TCL_ERROR;
                }
            } else {
                if (idx < 0  || idx > dd->nvar) {
                    Tcl_AppendResult(interp, "invalid index on -x option",
                                     (char *) NULL);
                    return TCL_ERROR;
                } else {
                    xPtr->xidx = idx;
                }
            }
        } else if (strEQ(NULL, "-y", 2)) {
            if (i++ >= argc) {
                Tcl_AppendResult(interp, "value for \"-y\" missing",
                                 (char *) NULL);
                return TCL_ERROR;
            }
            if (Tcl_GetInt(interp, argv[i], &idx) != TCL_OK) {
                int match = FALSE;
                for (idx=0; idx<dd->nvar; ++idx) {
                    if (strcmp(dd->nam[idx], argv[i]) == 0) {
                        xPtr->yidx = idx+1;
                        match = TRUE;
                        break;
                    }
                }
                if (!match) {
                    Tcl_AppendResult(interp, "invalid variable on -y option",
                                     (char *) NULL);
                    return TCL_ERROR;
                }
            } else {
                if (idx < 0  || idx > dd->nvar) {
                    Tcl_AppendResult(interp, "invalid index on -y option",
                                     (char *) NULL);
                    return TCL_ERROR;
                } else {
                    xPtr->yidx = idx;
                }
            }
        } else {
            Tcl_AppendResult(interp, "unknown option \"",
                             argv[i], "\"", (char *) NULL);
            return TCL_ERROR;
        }
    }
    return TCL_OK;
}


static int
showcfgXydiag(Tcl_Interp *interp,
              XydiagP xPtr)
{
    Tcl_DString cfg;

    Tcl_DStringInit(&cfg);
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "view");
    Tcl_DStringAppendElement(&cfg, "string::create");
    Tcl_DStringAppendElement(&cfg, "View");
    Tcl_DStringAppendElement(&cfg, "");
    if (xPtr) {
        Tcl_DStringAppendElement(&cfg, xPtr->view);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);
        
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "type");
    Tcl_DStringAppendElement(&cfg, "choice:1,points,lines:create");
    Tcl_DStringAppendElement(&cfg, "Diagram Type");
    Tcl_DStringAppendElement(&cfg, "points");
    if (xPtr) {
        if (xPtr->type == XYD_LINES) {
            Tcl_DStringAppendElement(&cfg, "lines");
        } else {
            Tcl_DStringAppendElement(&cfg, "points");
        }
    }
    Tcl_DStringEndSublist(&cfg);
    
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "x");
    Tcl_DStringAppendElement(&cfg, "string");
    Tcl_DStringAppendElement(&cfg, "X Variable");
    Tcl_DStringAppendElement(&cfg, "");
    if (xPtr) {
        sprintf(buf, "%d", xPtr->xidx);
        Tcl_DStringAppendElement(&cfg, buf);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);
    
    Tcl_DStringStartSublist(&cfg);
    Tcl_DStringAppendElement(&cfg, "y");
    Tcl_DStringAppendElement(&cfg, "string");
    Tcl_DStringAppendElement(&cfg, "Y Variable");
    Tcl_DStringAppendElement(&cfg, "");
    if (xPtr) {
        sprintf(buf, "%d", xPtr->yidx);
        Tcl_DStringAppendElement(&cfg, buf);
    } else {
        Tcl_DStringAppendElement(&cfg, "");
    }
    Tcl_DStringEndSublist(&cfg);
    
    if (xPtr) {
        ViewP view = viewFindByName(xPtr->view, NULL);
        styGetConfig(&cfg, &xPtr->style, STY_FG_ALL|STY_LINE_ALL, 0, 0, NULL);
        if (view) {
            viewGetConfig(&cfg, &view->opt, 0);
        } else {
            viewGetConfig(&cfg, NULL, 0);
        }
    } else {
        styGetConfig(&cfg, NULL, STY_FG_ALL|STY_LINE_ALL, 0, 0, NULL);
        viewGetConfig(&cfg, NULL, 0);
    }

    if (xPtr) {
        Tcl_DStringStartSublist(&cfg);
        Tcl_DStringAppendElement(&cfg, "_object");
        Tcl_DStringAppendElement(&cfg, xPtr->object);
        Tcl_DStringEndSublist(&cfg);
    }
        
    Tcl_DStringResult(interp, &cfg);
    return TCL_OK;
}

static int
plotXydiag(Tcl_Interp *interp,
           ViewP view,
           XydiagP xPtr)
{
    if (!view) return TCL_ERROR;

    plotAxis(interp, view, &xPtr->xax, &xPtr->xax.style, 0);
    plotAxis(interp, view, &xPtr->yax, &xPtr->yax.style, 1);

    xPtr->tag = dlAlloc(view, DL_SEGMENT, DL_MAIN_PLANE);
    dlMkSegment(xPtr->tag, xPtr->points, &xPtr->style);

    pltUpdateView(view);
    return TCL_OK;
}

static int
plotAxis(Tcl_Interp *interp,
         ViewP view,
         AxisP axis,
         StyleP style,
         int which)
{
    double ax, step;
    int i, ns;
    char ticlab[32];
    double ticsize = 0.02;

    ax = axis->max - axis->min;
    step = axis->step / ax;
    axis->ns = ns = (1./step) + 1;
    axis->seg = segAlloc(ns+1, LINESET);
    axis->ticlab = (char **)ckalloc(ns*sizeof(char *));
    axis->ticcrd = (c3dT *) ckalloc(ns*sizeof(c3dT));
    for (i=0; i<ns; ++i) {
        sprintf(ticlab, "%.2g", axis->min+i*axis->step);
        axis->ticlab[i] = strDupl(ticlab);
        if (which == 0) {
            segAddLine(&axis->seg, i*step, 0., 0., i*step, -ticsize, 0.);
            axis->ticcrd[i].x = i*step;
            axis->ticcrd[i].y = -ticsize;
        } else {
            segAddLine(&axis->seg, 0., i*step, 0., -ticsize, i*step, 0.);
            axis->ticcrd[i].x = -ticsize;
            axis->ticcrd[i].y = i*step;
        }
        axis->ticcrd[i].z = 0.;
    }

    axis->segtag = dlAlloc(view, DL_SEGMENT, DL_MAIN_PLANE);
    dlMkSegment(axis->segtag, axis->seg, style);
    axis->tictag = (int *) ckalloc(ns*sizeof(int));
    for (i=0; i<ns; ++i) {
        axis->tictag[i] = dlAlloc(view, DL_TEXT, DL_MAIN_PLANE);
        dlMkText(axis->tictag[i], &axis->ticcrd[i], axis->ticlab[i], style);
    }
    return TCL_OK;
}

static int
clearXydiag(Tcl_Interp *interp,
            ViewP view,
            XydiagP xPtr)
{

    if (!view) return TCL_ERROR;

    clearAxis(interp, view, &xPtr->xax);
    clearAxis(interp, view, &xPtr->yax);

    dlRemove(view, xPtr->tag);
    xPtr->tag = 0;
    return TCL_OK;
}

static int
clearAxis(Tcl_Interp *interp,
          ViewP view,
          AxisP axis)
{
    int i;
    int ns = axis->ns;

    segFree(&axis->seg);
    for (i=0; i<ns; ++i) {
        ckfree(axis->ticlab[i]);
    }
    if (axis->ticlab) ckfree((char *) axis->ticlab);
    if (axis->ticcrd) ckfree((char *) axis->ticcrd);

    dlRemove(view, axis->segtag);
    axis->segtag = 0;
    for (i=0; i<ns; ++i) {
        dlRemove(view, axis->tictag[i]);
    }
    if (axis->tictag) ckfree((char *) axis->tictag);
    axis->tictag = 0;
    axis->ns = 0;
    return TCL_OK;
}

static void
axis(double xmin, double xmax,
     double *x0, double *x1, double *step)
{
    double s, delta;
    int nstep, ld;

    delta = xmax - xmin;
    ld = ((int) log10(delta)) - 1;
    if (ld == 0) {
        s = 1.;
    } else {
        s = pow(10., (double) ld);
    }
    nstep = (int) ceil(delta / s);

    if (nstep < 3) {
        s /= 5.;
    } else if (nstep < 6) {
        s /= 2.;
    } else if (nstep >= 25) {
        s *= 5.;
    } else if (nstep >= 12) {
        s *= 2.;
    }

    *x0 = floor(xmin / s) * s;
    *x1 = ceil(xmax / s) * s;
    *step = s;
    return;
}
