#
# $Header: /scratch/fock/cvsroot/xd/xdgraph/tcl/guiConfigure.tcl,v 1.12 1999/06/03 13:44:22 richter Exp $
#
#
# $Log: guiConfigure.tcl,v $
# Revision 1.12  1999/06/03 13:44:22  richter
# The [list] around [generate] was too much
#
# Revision 1.11  1999/06/03 12:35:47  richter
# Tcl 8 doesn't like the [generate ...] hack, currently I don't see a way
# to postpone the evaluation of this construct, so we don not do it anymore.
#
# Revision 1.10  1997/04/02 13:53:23  richter
# no file selector on non-Unix machines
#
# Revision 1.9  1996/11/26  14:58:45  richter
# check for existence of foreground2 key before using it
#
# Revision 1.8  1996/11/14  13:14:31  richter
# handle 2nd foreground (_onoff_$key) correctly
#
# Revision 1.7  1996/11/07  17:12:59  richter
# no need to clear an object before plotting here,
# the plot command does it anyway
#
# Revision 1.6  1996/10/11  11:05:16  richter
# *** empty log message ***
#
# Revision 1.5  1996/09/23  12:05:47  richter
# handle cancels in CSBox
#
# Revision 1.4  1996/09/14  09:38:20  richter
# added "required" option to name field
# added aim and maim to linestyle chooser
# changed trace_uservalue to handle aim mode
#
# Revision 1.3  1996/09/11  11:28:00  richter
# fixed a nasty typo in auto_generate_name
#
# Revision 1.2  1996/09/06  14:51:43  richter
# provide a procedure to generate names for objects
#
# Revision 1.1  1996/09/06  12:43:38  richter
# Initial revision
#
# Revision 1.1  1996/05/23  07:42:20  richter
# Initial revision
#
#
proc guiConfigure {object_type {object_cmd ""}} {
    global xd_tk_major xd_unix active

    if {$object_cmd == ""} {
	set is_new 1
	set object_cmd $object_type
	set object_name _NEW_$object_type
	if {$object_type == "driver"} {
	    set is_new 0
	}
    } else {
	set is_new 0
	regsub -all {:} $object_cmd {_} object_name
    }

    set w .config_${object_type}${object_name}
    if {[winfo exists $w]} {
        wm deiconify $w
        raise $w
        return
    }

    toplevel $w -bd 5
    if {$is_new} {
	wm title $w "Create $object_type"
    } else {
	wm title $w "Configure $object_type"
    }
#
    set baseName $object_name
    set curArray cur$baseName
    set cfgArray cfg$baseName
    set defArray def$baseName
    upvar #0 $curArray cur $cfgArray cfg $defArray def

    if {$object_type == "driver"} {
	set cfg_list [$object_cmd configure]
    } else {
	if {$is_new} {
	    set cfg_list {{name string::create,required Name "" ""}}
	} else {
	    set cfg_list [list [list name string::create,required Name "" $object_cmd]]
	}
	set cfg_list [concat $cfg_list [$object_cmd configure]]
    }
    set frame [make_section $w {{} {Specific Options}} $object_name 1]
    foreach c $cfg_list {
	unpack_cfg $c key label type size options choices
	if {[string index $key 0] == "_"} {
	    if {[string compare $key "_SECTION"] == 0} {
		set section_type [string tolower [lindex [lindex $c 1] 0]]
		set xpnd [expr [string compare $section_type $object_type]==0]
		set frame [make_section $w $c $object_name $xpnd]
	    }
	    continue
	}
	if {$size > 1} {
	    set allkeys {}
	    for {set i 0} {$i < $size} {incr i} {
		if {$i > 0} {
		    set enum ${key}_$i
		} else {
		    set enum $key
		}
		lappend allkeys $enum
		set def($enum) [lindex [lindex $c 3] $i]
		if {$is_new} {
		    set cfg($enum) ""
		} else {
		    set cfg($enum) [lindex [lindex $c 4] $i]
		}
		set cur($enum) $cfg($enum)
	    }
	} else {
	    set allkeys "$key"
	    set def($key) [lindex $c 3]
	    if {$is_new} {
		set cfg($key) ""
	    } else {
		set cfg($key) [lindex $c 4]
	    }
	    set cur($key) $cfg($key)
	}
	set cmd [list container $frame.$key -label $label \
		     -var [list $allkeys $baseName] \
		     -help config/$object_type/[string trimright $key 2]]
	switch -exact -- $key {
	    view {
		foreach view [view] {
		    lappend vl "$view $view"
		}
		lappend vl "automatic {}"
		lappend cmd -choose $vl
		if {$def(view) == ""} {
		    set def(view) $active(view)
		}
	    }
	    name {
		switch -exact -- $object_type {
		    dataset {
			lappend cmd -defCmd \
			    [list auto_generate_name :set "" [dataset]]
		    }
		    view -
		    driver {}
		    default {
			if {$active(dataset) != ""} {
			    set prefix "$active(dataset):"
			    append prefix [string range $object_type 0 2]
			    lappend cmd -defCmd \
				[list auto_generate_name $prefix "" \
				     [$active(dataset) $object_type]]
			}
		    }
		}
	    }
	}
	switch -exact -- $type {
	    choice {
		lappend cmd -choose $choices
	    }
	    color {
		lappend cmd -select Select_fg
		if {[string first "onoff" $options] >= 0} {
		    set cfg(_onoff_${key}) [string compare $cfg(${key}) ""]
		    lappend cmd -onoff _onoff_${key}
		}
	    }
	    file {
		if {$xd_unix} {
		    lappend cmd -select Select_file
		}
	    }
	    flag {
		lappend cmd -choose {{on on} {off off}}
	    }
	    list {
		set cfg(_mode_${key}) "unknown"
		lappend cmd -choose {{linear linear} {geom geom} {aim aim} {"- aim" maim} {user ""}}
	    }
	    style {
		lappend cmd -choose {{solid solid} {dot dot} {dash dash} {long long} {user ""}}
	    }
	}
	if {[string first "readonly" $options] >= 0 ||
	    ([string first "create" $options] >= 0 && ! $is_new)} {
	    lappend cmd -readonly 1
	}
	if {[string first "required" $options] >= 0 } {
	    lappend cmd -required 1
	}
##	puts "$cmd"
	eval $cmd
	switch -exact -- $type {
	    list {
		set vr $frame.$key.container
		label $vr.stepsl -text "# steps:" -anchor sw
		entry $vr.stepse -textvariable ${cfgArray}(_steps_$key) -width 4
		label $vr.startl -text "  start:"
		entry $vr.starte -textvariable ${cfgArray}(_start_$key) -width 4
		label $vr.incrl -text "  increment:"
		entry $vr.incre -textvariable ${cfgArray}(_incr_$key) -width 4
		frame $vr.space -height 3p -width 18p
		pack $vr.stepsl $vr.stepse $vr.startl $vr.starte\
		    $vr.incrl $vr.incre $vr.space \
		    -side left -expand yes -fill both
		trace variable cfg($key) w "trace_uservalues $frame $cfgArray"
		trace_uservalues $frame $cfgArray cfg $key w
		bind $frame.$key <Destroy> \
		    [list trace vdelete ${cfgArray}($key) w [list trace_uservalues $frame $cfgArray]]
	    }
	}
	pack $frame.$key -pady 5 -side top -expand yes -fill both
    }
    #
    set okcmd [list Do_Configure $object_cmd $is_new $cfg_list $cfgArray $curArray]
    switch -exact -- $object_type {
	"dataset" -
	"driver" {
	    object_buttons $w [list [list Ok "$okcmd 0"] [list Cancel destroy]]
	}
	default {
	    object_buttons $w [list [list Ok "$okcmd 0"]\
				   [list Plot "$okcmd 1"]\
				   [list Cancel destroy]]
	}
    }
#
    pack $w.buttons -pady 5 -side top -expand yes -fill both
}

proc Do_Configure {object_cmd is_new cfg_list cfgArray curArray plot w} {
    global active
    upvar #0 $cfgArray cfg $curArray cur

    if {$is_new} {
	if {$object_cmd != "dataset"} {
	    set name [complete_objectname $cfg(name)]
	    set cfg(name) $name
	}
	if {$cfg(name) == ""} {
	    error "No name specified"
	}
	set in [list $object_cmd $cfg(name)]
	set cmd ""
    } else {
	set in [list $object_cmd configure]
	set cmd $in
    }

    foreach c $cfg_list {
	unpack_cfg $c key label type size options choices
	if {[string index $key 0] == "_" ||
	    [string compare $key "name"] == 0 || 
	    [string compare $key "foreground2"] == 0} {
	    continue
	}
	if {! $is_new && ([string first "create" $options] > -1)} {
	    continue
	}
	if {$cfg($key) == ""} {
	    if {$is_new && ([string first "required" $options] > -1)} {
		error "option -$key is required"
	    }
	} else {
	    switch -exact -- $type {
		color {
		    if {[exists cfg _onoff_${key}2] &&
			$cfg(_onoff_${key}2)} {
			set fgn  "[list $cfg($key) $cfg(${key}2)]"
		    } else {
			set fgn $cfg($key)
		    }
		    if {[exists cur ${key}2] &&
			[string compare $cur(${key}2) ""] != 0} {
			set fgc "[list $cur($key) $cur(${key}2)]"
		    } else {
			set fgc $cur($key)
		    }
		    if {[string compare $fgn $fgc] != 0} {
			lappend in -$key $fgn
		    }
		}
		list {
		    switch -glob -- $cfg($key) {
			lin* -
			geo* {
			    lappend in -values [generate $cfg($key) $cfg(_steps_$key) $cfg(_start_$key) $cfg(_incr_$key)]
			}
			aim -
			maim {
#			    append in " -values \[generate $cfg($key) $cfg(_steps_$key) $cfg(_start_$key) \]"
			    lappend in -values [generate $cfg($key) $cfg(_steps_$key) $cfg(_start_$key)]
			}
			default {
			    if {[string compare $cfg($key) $cur($key)] != 0} {
				lappend in -$key $cfg($key)
			    }
			}
		    }
		}
		default {
		    set current ""
		    set config ""
		    if {$size > 1} {
			for {set i 0} {$i < $size} {incr i} {
			    if {$i > 0} {
				set enum ${key}_$i
			    } else {
				set enum $key
			    }
			    lappend current $cur($enum)
			    lappend config $cfg($enum)
			}
		    } else {
			set current $cur($key)
			set config $cfg($key)
		    }
		    if {[string compare $config $current] != 0} {
			lappend in -$key $config
		    }
		}
	    }
	}
    }

##    puts $in
    if {[string compare $in $cmd] != 0} {
	if {[catch {doCommand $in}] == 0} {
	    destroy $w
	    if $plot {
		update idletasks
#		if {!$is_new} {
#		    doCommand "$cfg(name) clear"
#		}
		doCommand "$cfg(name) plot"
	    } else {
		update idletasks
	    }
	}
    } else {
	destroy $w
    }
}

proc unpack_cfg {cfg keyName labName typName sizName optName choName} {
    upvar $keyName key $labName label $typName type
    upvar $sizName size $optName options $choName choices

    set key [lindex $cfg 0]
    if {[string index $key 0] == "_"} {
	return
    }

    set raw_type [split [lindex $cfg 1] ":"]
    set type [lindex $raw_type 0]
    set size [lindex $raw_type 1]
    set options [lindex $raw_type 2]
    set sublabel [split [lindex $raw_type 3] ","]
    if {$type == "choice"} {
	set cho [split $size ","]
	set size [lindex $cho 0]
	set choices {}
	foreach c  [lrange $cho 1 end] {
	    lappend choices "$c $c"
	}
    }
    if {$size == ""} {
	set size 1
    }

    set label [list [lindex $cfg 2] $sublabel]
}

proc handle_section {w varName} {
    upvar #0 $varName var

    if $var {
	pack $w -expand yes -fill both
    } else {
	pack forget $w
    }
}

proc make_section {w c object_name start} {

    set ftext [lindex $c 1]
    set fname [string tolower [lindex $ftext 0]]
    pack [frame $w.$fname] -expand yes -fill both -pady 5
    set frame [frame $w.$fname.frame]
    pack [checkbutton $w.$fname.button -text $ftext \
	      -variable ${object_name}_$fname \
	      -command "handle_section $frame ${object_name}_$fname"] \
	-anchor sw
    global ${object_name}_$fname
    set ${object_name}_$fname [expr ! $start]
    $w.$fname.button invoke
    return $frame
}

proc auto_generate_name {{prefix ""} {postfix ""} {except ""} args} {

    set name ""
    for {set i 1} {$i < 1000} {incr i} {
	set n [format "%s%3.3d%s" $prefix $i $postfix]
	if {[lsearch -exact $except $n] == -1 } {
	    set name $n
	    break
	}
    }
    return $name
}

proc Select_fg {w varName key} {
    upvar #0 $varName cfg

    if {[string compare $cfg($key) ""]} {
	set color [CSBox $w -modal yes -preset $cfg($key)]
    } else {
	set color [CSBox $w -modal yes]
    }
    if {[string compare $color ""]} {
	set cfg($key) $color
    }
}

proc Select_file {w varName key} {
    global xd_pwd
    upvar #0 $varName cfg

    if {[set file [FSBox $w -modal yes -pwd $xd_pwd]] != ""} {
	set cfg($key) $file
    }
}

proc trace_uservalues {w cfgName varName key access} {
    upvar #0 $cfgName cfg

    switch -glob -- $cfg($key) {
	lin* -
	geo* {
	    set mode auto
	}
	aim -
	maim {
	    set mode aim
	}
	default {
	    set mode user
	}
    }
    if {$cfg(_mode_${key}) == $mode} {
	return
    }
    set cfg(_mode_${key}) $mode
    set vr $w.values.container
    switch -exact -- $mode {
	"auto" {
	    $vr.stepsl config -fg black
	    $vr.stepse config -state normal -bg \#b98e8e
	    $vr.startl config -fg black
	    $vr.starte config -state normal -bg \#b98e8e
	    $vr.incrl config -fg black
	    $vr.incre config -state normal -bg \#b98e8e
	}
	"aim" {
	    $vr.stepsl config -fg black
	    $vr.stepse config -state normal -bg \#b98e8e
	    $vr.startl config -fg black
	    $vr.starte config -state normal -bg \#b98e8e
	    $vr.incrl config -fg \#b0b0b0
	    $vr.incre config -state disabled -bg \#b0b0b0
	}
	default {
	    $vr.stepsl config -fg \#b0b0b0
	    $vr.stepse config -state disabled -bg \#b0b0b0
	    $vr.startl config -fg \#b0b0b0
	    $vr.starte config -state disabled -bg \#b0b0b0
	    $vr.incrl config -fg \#b0b0b0
	    $vr.incre config -state disabled -bg \#b0b0b0
	}
    }
}
