#
# $Header: /scratch/fock/cvsroot/xd/xdgraph/tcl/guiHelpMenu_VMS.tcl,v 1.1 1997/04/02 13:44:22 richter Rel $
#
#
# $Log: guiHelpMenu_VMS.tcl,v $
# Revision 1.1  1997/04/02 13:44:22  richter
# Initial revision
#
# Revision 1.5  1996/11/07  10:48:28  richter
# make guiHelpMenu work with Tcl 7.3
#
# Revision 1.4  1996/11/01  17:50:53  richter
# sort help menu entries
#
# Revision 1.3  1996/09/30  11:40:41  richter
# catch error if no help directories exist
#
# Revision 1.2  1996/09/06  12:49:58  richter
# many changes
#
# Revision 1.1  1996/05/23  07:42:20  richter
# Initial revision
#
#

source $xd_library/podBox.tcl

proc guiHelpMenu {w} {
    global xd_tk_major xd_datadir env hlp_directory hlp_file

    set hlp_directory $xd_datadir/help/C
    set hlp_file ""

    menubutton $w -text Help -underline 0 -menu $w.menu
    set w [menu $w.menu]

    guiHelpDir $w .

    if {[catch "glob [lindex $hlp_directory 0]/*.dir" dirList] == 0} {
	foreach file [lsort $dirList] {
	    if {[file isdirectory $file]} {
		regexp {\]([^.]+)} $file match dir
		set m [menu $w.$dir]
		if {$xd_tk_major >= 4} {
		    $m configure -tearoff off
		}
		$w add cascade -label "$dir" -menu $m
		guiHelpDir $m $dir
	    }
	}
    }
}

proc guiHelpDir {w dir} {
    global hlp_directory
    if {[catch "glob [lindex $hlp_directory 0]/$dir/*.pod" podList] == 0} {
	foreach pod [lsort $podList] {
	    set label [getfirsthead1 $pod]
	    regexp {\]([^.]+)} $pod match file
	    if {$label == ""} {
		set label $file
	    }
	    $w add command -label $label -command "guiShowHelp $dir $file"
	}
    }
}
proc getfirsthead1 {file} {
    set F [open $file]
    while {[gets $F line] > -1} {
	if {[regexp -- {^=head1 (.*)} $line match label]} {
	    return $label
	}
    }
    close $F
    return ""
}

proc xdhelp {w} {
    global xd_tk_major 

    eval "guiShowHelp [split $w /]"
}

proc guiShowHelp {dir file {start ""}} {
    global xd_tk_major hlp_directory hlp_file hlp_locArray hlp_textWidget

    set oldfile $hlp_file

    set hlp_file [lindex $hlp_directory 0]/$dir/${file}.pod
    if {![file exists $hlp_file] && [llength $hlp_directory] > 1} {
	set hlp_file [lindex $hlp_directory 1]/$dir/${file}.pod
    }

    if {![file exists $hlp_file]} {
	error "Help file does not exist: $hlp_file"
    }

    set w .help
    if {[winfo exists $w]} {
	if {[string compare $oldfile $hlp_file] == 0} {
	    guiHelpJump $start
	    return
	} else {
	    $hlp_textWidget delete 1.0 end
	    set f $w.bt
	    $f.goto.m delete 0 end
	}
    } else {
	toplevel $w
	set f [frame $w.st]
	set hlp_textWidget [text $f.text -relief sunken -bd 2 \
		   -yscrollcommand "$f.scroll set" \
		   -width 80 -height 24]
	scrollbar $f.scroll -relief sunken -command "$f.text yview"
	pack $f.text $f.scroll -side left -expand yes -fill both
	set f [frame $w.bt -bd 2]
	menubutton $f.goto -text "Goto" -menu $f.goto.m -relief raised
	menu $f.goto.m
	button $f.ok -text "Done" -command "destroy $w"
	pack $f.goto $f.ok -side left -expand yes -ipadx 2 -ipady 2
	pack $w.bt -side bottom -fill x
	pack $w.st -expand yes -fill both
    }
    wm title $w "Help on $dir, $file"
#    puts "$hlp_file"
    podBox $hlp_textWidget $hlp_file topics hlp_locArray

    foreach key $topics {
	$f.goto.m add command -label "$key" \
	    -command "$hlp_textWidget yview -pickplace $hlp_locArray($key)"
    }

    guiHelpJump $start
}

proc guiHelpJump {where} {
    global hlp_textWidget hlp_locArray

    set loc ""
    if {$where == ""} {
	set loc 1.0
    } else {
	if {[exists hlp_locArray $where]} {
	    set loc $hlp_locArray($where)
	} else {
	    foreach topic [array names hlp_locArray] {
		if {[regexp -- "^-${where}.*" $topic]} {
		    set loc $hlp_locArray($topic)
		    break
		}
	    }
	}
    }
    if {$loc != ""} {
	$hlp_textWidget  yview -pickplace $loc
    }
}
